#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""FAL command for a BOX cloud service."""

import os

import libfal.falcmd as falcmd
from iobase.cmdexec import (CmdExec, PIPE, NULL)
from ionas.request import (Entry, RequestData, ResponseData)
from libfal.util import get_proxy, NoProxy
from fal_info import ReqTypes
from fal_info import Results
import pal_info

from pkgaddon.box_config import (REST_BASE_URL, LIB_PATH_FOR_BOX)

RESULT_NAME = "result"
FAILED_RESULT = "Failed"
CHECK_FAILED_MSG = "Check Failed!!"

LD_LIBRARY_PATH_NAME = "LD_LIBRARY_PATH"


class FalBox(falcmd.FalCmd):
    """FAL Box command implementation class."""

    def __download_handler(self, entry):
        """Handle download request and set entry to result."""

        result_entry = self.send_entry(entry, pal_info.PAL_BOX_PATH)
        return result_entry

    def __get_proxy(self, url):

        try:
            return get_proxy(url)
        except NoProxy:
            pass

    def __send_entry_with_fork(self, entry, target): 
        """Send entry to target it is a forked process."""

        new_env = os.environ
        new_env[LD_LIBRARY_PATH_NAME] = LIB_PATH_FOR_BOX 
        proc = str(RequestData(entry)) | \
                    CmdExec(target, stdout=PIPE, stderr=NULL,
                            close_fds=True, env=new_env)
        response_str = proc.stdout.read()
        if proc.wait() or not response_str:
            return None

        return ResponseData().load(response_str).get_entry()

    def __check_handler(self, entry):
        """Handle check request and set entry to result."""

        try:
            box_info = entry.get().get_clone()
            proxy_info = self.__get_proxy(REST_BASE_URL)

            if proxy_info:
                box_info.append_child(proxy_info)

            result_entry = self.__send_entry_with_fork(
                                Entry(ReqTypes.CHECK, box_info),
                                 pal_info.PAL_BOX_PATH)

            entry.set_result(result_entry.get_result())
            entry.get().append_child(
                             RESULT_NAME,
                             result_entry.get().get_value(
                                     RESULT_NAME, FAILED_RESULT))

        except:
            self.exception(CHECK_FAILED_MSG)
            entry.set_result(Results.CHECK_ERROR)
            entry.get().append_child(RESULT_NAME, FAILED_RESULT)

        return entry

    def __init__(self, option=None):
        """Constructor."""

        falcmd.FalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.DOWNLOAD, self.__download_handler)
        self.set_handler(ReqTypes.CHECK, self.__check_handler)

#
# Main
#

if __name__ == "__main__":
    FalBox(option=falcmd.FalCmdOption()).handle_request()
