#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

from iobase.exception import NotFoundError


def _is_dropbox(node):

    try:
        return node.get_node("shareservice/dropbox").is_enabled()
    except NotFoundError:
        return False


def _is_amazons3(node):

    try:
        return node.get_node("shareservice/amazons3").is_set_to_true("amazons3")
    except NotFoundError:
        return False


def _is_amazons3_accesskey(node):

    if node.get_value("shareservice/amazons3/accesskey"):
        return True
    return False


def _is_amazons3_secretkey(node):

    if node.get_value("shareservice/amazons3/secretkey"):
        return True
    return False


def _is_amazons3_bucket(node):

    if node.get_value("shareservice/amazons3/bucket"):
        return True
    return False


def _is_amazons3_httpmethod(node):

    if node.get_value("shareservice/amazons3/httpmethod"):
        return True
    return False


def _is_amazons3_storageclass(node):

    if node.get_value("shareservice/amazons3/storageclass"):
        return True
    return False


def _is_amazons3_downloadinterval(node):

    if node.get_value("shareservice/amazons3/downloadinterval"):
        return True
    return False


def _is_amazons3_uploadmaxfilesize(node):

    if node.get_value("shareservice/amazons3/uploadmaxfilesize"):
        return True
    return False


def _is_amazons3_mpupload(node):

    if node.get_node("shareservice/amazons3/mpupload").get_attr_value("enable"):
        return True
    return False


def _is_cloudn(node):

    try:
        return node.get_node("shareservice/cloudn.tmp").is_set_to_true("cloudn")
    except NotFoundError:
        return False


def _is_cloudn_accesskey(node):

    if node.get_value("shareservice/cloudn.tmp/accesskey"):
        return True
    return False


def _is_cloudn_secretkey(node):

    if node.get_value("shareservice/cloudn.tmp/secretkey"):
        return True
    return False


def _is_cloudn_bucket(node):

    if node.get_value("shareservice/cloudn.tmp/bucket"):
        return True
    return False


def _is_cloudn_httpmethod(node):

    if node.get_value("shareservice/cloudn.tmp/httpmethod"):
        return True
    return False


def _is_cloudn_storageclass(node):

    if node.get_value("shareservice/cloudn.tmp/storageclass"):
        return True
    return False


def _is_cloudn_downloadinterval(node):

    if node.get_value("shareservice/cloudn.tmp/downloadinterval"):
        return True
    return False


def _is_cloudn_uploadmaxfilesize(node):

    if node.get_value("shareservice/cloudn.tmp/uploadmaxfilesize"):
        return True
    return False


def _is_cloudn_mpupload(node):

    if node.get_node("shareservice/cloudn.tmp/mpupload").get_attr_value("enable"):
        return True
    return False






def _is_azure(node):

    try:
        return node.get_node("shareservice/azure").is_enabled()
    except NotFoundError:
        return False


def _is_azure_accountname(node):

    if node.get_value("shareservice/azure/accountname"):
        return True
    return False


def _is_azure_accesskey(node):

    if node.get_value("shareservice/azure/accesskey"):
        return True
    return False


def _is_azure_container(node):

    if node.get_value("shareservice/azure/container"):
        return True
    return False


def _is_azure_httpmethod(node):

    if node.get_value("shareservice/azure/httpmethod"):
        return True
    return False


def _is_azure_downloadinterval(node):

    if node.get_value("shareservice/azure/downloadinterval"):
        return True
    return False


def _s3_enable(node, enable=True):

    if enable:
        node.get_node("shareservice/amazons3").set_attrs({"enable": "true"})
    else:
        node.get_node("shareservice/amazons3").set_attrs({"enable": "false"})


def _s3_move(node, src, dst, attr):

    src_node = node.get_node(src)
    dst_node = node.get_node(dst)

    def move(path):
        dst_node.delete_child(path)
        dst_node.append_child(src_node.get_node(path))

    def delete(path):
        path_list = path.split("/")
        path_depth = 0
        cur_node = node
        for cur_path in path_list:
            if path_depth == len(path_list):
                cur_node.delete_child(cur_path)
                break
            cur_node = cur_node.get_node(cur_path)
            path_depth += 1

    move("accesskey")
    move("secretkey")
    move("bucket")
    move("httpmethod")
    move("storageclass")
    move("downloadinterval")
    move("endpoint")
    move("port")
    move("uploadmaxfilesize")
    move("mpupload")

    delete(src)

    dst_node.set_attrs({attr: "true"})
    

def rule_cloudsync(node):

    if _is_amazons3(node) and _is_cloudn(node) and _is_dropbox(node) and _is_azure(node):
        return False

    _s3_enable(node, False)

    if _is_amazons3(node):
        if not _is_amazons3_accesskey(node):
            return False
        if not _is_amazons3_secretkey(node):
            return False
        if not _is_amazons3_bucket(node):
            return False
        if not _is_amazons3_httpmethod(node):
            return False
        if not _is_amazons3_storageclass(node):
            return False
        if not _is_amazons3_downloadinterval(node):
            return False
        if not _is_amazons3_uploadmaxfilesize(node):
            return False
        if not _is_amazons3_mpupload(node):
            return False

        _s3_enable(node)

    if _is_cloudn(node):
        if not _is_cloudn_accesskey(node):
            return False
        if not _is_cloudn_secretkey(node):
            return False
        if not _is_cloudn_bucket(node):
            return False
        if not _is_cloudn_httpmethod(node):
            return False
        if not _is_cloudn_storageclass(node):
            return False
        if not _is_cloudn_downloadinterval(node):
            return False
        if not _is_cloudn_uploadmaxfilesize(node):
            return False
        if not _is_cloudn_mpupload(node):
            return False

        _s3_move(node, "shareservice/cloudn.tmp", "shareservice/amazons3", "cloudn")
        _s3_enable(node)

    if _is_azure(node):
        if not _is_azure_accountname(node):
            return False
        if not _is_azure_accesskey(node):
            return False
        if not _is_azure_container(node):
            return False
        if not _is_azure_httpmethod(node):
            return False
        if not _is_azure_downloadinterval(node):
            return False

    return True
