#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for a BOX cloud service."""

from __future__ import with_statement

try:
    import json
except:
    import simplejson as json

import os
import threading
import shutil
from urlparse import urlunparse

from iobase.cmdexec import (CmdExec, NULL)
from iobase.exception import ExecError
import libpal.palcmd as palcmd
try:
    from cloudsync.box.client import (FileClient,)
except:
    from ionas.restapi.client import (FileClient,)
from pal_info import (ReqTypes, Results)
try:
    from cloudsync.box import (client,)
except:
    from ionas.restapi import (client,)

import pal_config as config
from cloudsync.util import (get_proxies_dict,)
from cloudsync.boxinfo import (LoginError, BoxTokenMgr, BoxInfo,
                                ACCESS_TOKEN_NAME, REFRESH_TOKEN_NAME)
from pkgaddon.box_config import (CLIENT_ID, CLIENT_SECRET, OAUTH_BASE_URL,
                                    REST_BASE_URL, UPLOAD_BASE_URL,
                                    REDIRECT_URL, DOWNLOAD_TARGET_CONF_PATH,)


SHARENAME_NAME = "sharename"

RESULT_NAME = "result"
MAILADDRESS_NAME = "mailaddress"
PASSWD_NAME = "passwd"
PROXY_SERVER_PATH = "proxy/serverlist/server"
PROXY_HOST_PATH = PROXY_SERVER_PATH + "/host"
PROXY_PORT_PATH = PROXY_SERVER_PATH + "/port"

SEND_DOWNLOAD_SIGNAL_ARGS = [config.NASDSYNC_INITD, "reload"]

PROTOCOL_LIST = ("http", "https")

SUCCESS_RESULT = "Success"
FAILED_RESULT = "Failed"
TIMEOUT_RESULT = "TimeOut"
LOGIN_FAILED_RESULT = "LoginFailed"

CHECK_TIMEOUT_SEC = 20
RESULT_KEY = "result"
BOX_ROOT_PATH = "/"

CHECK_TOKEN_ID = "_check_token_id"


class BoxCheckTokenMgr(BoxTokenMgr):
    """A delivered class for a check to connect to server."""

    def __init__(self, base_url, client_id, client_secret,
            redirect_url, proxies=None):
        """Constructor."""

        self.access_token = ""
        self.refresh_token = ""
        BoxTokenMgr.__init__(self, base_url, client_id, client_secret,
                                redirect_url, proxies)

    def __del__(self):
        """Destructor."""

        try:
            if self.refresh_token:
                self._revoke_token(self.refresh_token)
        except:
            # Ignore error
            pass

    def get_access_token(self, token_id, mailaddress, password):
        """Return a access token for the content API of BOX server.
           Note: This token is not saved.
        """

        auth_code = self._get_auth_code(mailaddress, password)

        param = {
            "grant_type": "authorization_code",
            "code": auth_code,
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "redirect_url": self.redirect_url,
        }

        response = self._request(self.base_url + "/token", param)
        response_data = json.loads(response.read())
        self.access_token = response_data[ACCESS_TOKEN_NAME]
        self.refresh_token = response_data[REFRESH_TOKEN_NAME]
        return self.access_token

    def refresh_access_token(self, token_id):
        """Refresh a access token and return it."""

        # Force a login to be error.
        raise LoginError(token_id)


def _exec_check(client, result_dict):
    """Thread main for checking to access to the BOX server."""

    try:
        client.stat_directory(BOX_ROOT_PATH)
        result_dict[RESULT_KEY] = SUCCESS_RESULT

    except LoginError:
        result_dict[RESULT_KEY] = LOGIN_FAILED_RESULT

    except:
        result_dict[RESULT_KEY] = FAILED_RESULT


class PalBox(palcmd.PalCmd):
    """PAL Box command implementation class."""

    def __make_download_target_conf(self, share_name):
        """Make a conf file written target share name to."""

        temp_path = DOWNLOAD_TARGET_CONF_PATH + "." + str(os.getpid())
        try:
            with open(temp_path, "wb") as file:
                file.write(share_name)

            shutil.move(temp_path, DOWNLOAD_TARGET_CONF_PATH)
        except:
            if os.path.exists(temp_path):
                os.remove(temp_path)
            raise

    def __download_handler(self, entry):
        """Handle download request."""

        try:
            share_name = entry.get().get_node(SHARENAME_NAME).get_text()
            self.__make_download_target_conf(share_name)

            CmdExec(SEND_DOWNLOAD_SIGNAL_ARGS).execute(
                stdout=NULL, stderr=NULL)
            entry.set_result(Results.OK)

        except ExecError:
            self.exception("Download Failed!!")
            entry.set_result(Results.DOWNLOAD_ERROR)

        return entry

    def __check_handler(self, entry):
        """Handle check request."""

        result = SUCCESS_RESULT
        try:
            body = entry.get()
            mailaddress = body.get_value(MAILADDRESS_NAME)
            passwd = body.get_value(PASSWD_NAME)
            proxy_host = body.get_value(PROXY_HOST_PATH, None)
            proxy_port = body.get_value(PROXY_PORT_PATH, None)

            proxies_dict = get_proxies_dict(proxy_host, proxy_port,
                                            PROTOCOL_LIST)
            client = FileClient(BoxInfo(REST_BASE_URL, UPLOAD_BASE_URL,
                                        CHECK_TOKEN_ID,
                                        mailaddress, passwd,
                                        BoxCheckTokenMgr(OAUTH_BASE_URL,
                                                    CLIENT_ID,
                                                    CLIENT_SECRET,
                                                    REDIRECT_URL,
                                                    proxies_dict)),
                                proxies=proxies_dict)

            result_dict = {}
            check_thread = threading.Thread(target=_exec_check,
                                            args=(client, result_dict))
            check_thread.daemon = True
            check_thread.start()
            check_thread.join(CHECK_TIMEOUT_SEC)

            if check_thread.is_alive():
                result = TIMEOUT_RESULT
            else:
                result = result_dict[RESULT_KEY]
        except:
            self.exception("Check Failed!!")
            result = FAILED_RESULT

        if result == SUCCESS_RESULT:
            entry.set_result(Results.OK)
        else:
            entry.set_result(Results.CHECK_ERROR)

        entry.get().append_child(RESULT_NAME, result)
        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.DOWNLOAD, self.__download_handler)
        self.set_handler(ReqTypes.CHECK, self.__check_handler)

#
# Main
#

if __name__ == "__main__":
    PalBox(option=palcmd.PalCmdOption()).handle_request()
