#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for a OneDrive cloud service."""

from __future__ import with_statement

import os
import sys
import shutil
import pickle

import requests

from iobase.cmdexec import (CmdExec, NULL)
from iobase.exception import ExecError
import libpal.palcmd as palcmd
from pal_info import (ReqTypes, Results)
from onedrive.onedriveaccess import (OneDriveAccess,)

import pal_config as config
from cloudsync.util import (get_proxies_dict,)
from pkgaddon.onedrive_config import (APP_ID, CLIENT_SECRET, REDIRECT_URL,
                                        BUSINESS_APP_ID,
                                        BUSINESS_CLIENT_SECRET,
                                        BUSINESS_REDIRECT_URL,
                                        BUSINESS_AUTH_SERVER_URL,
                                        BUSINESS_AUTH_TOKEN_URL,
                                        BUSINESS_SERVICE_NAME,
                                        SCOPES,
                                        ONEDRIVE_CONF_DIR,
                                        ONEDRIVE_BUSINESS_CONF_DIR,
                                        DOWNLOAD_TARGET_CONF_PATH,)


SEND_DOWNLOAD_SIGNAL_ARGS = [config.NASDSYNC_INITD, "reload"]

PROTOCOL_LIST = ("http", "https")

PROXY_SERVER_PATH = "proxy/serverlist/server"
PROXY_HOST_PATH = PROXY_SERVER_PATH + "/host"
PROXY_PORT_PATH = PROXY_SERVER_PATH + "/port"

SHARE_NAME_PATH = "share/name"
SHARENAME_NAME = "sharename"
CODE_NAME = "code"

SESSION_NAME = "session"
SESSION_SUFFIX = ".session"

FORBUSINESS_ATTR = "forbusiness"
TRUE_TEXT = "true"
URL_NAME = "url"
TOKEN_NAME = "token"
ACCESS_TOKEN_KEY = "access_token"

ERROR_INVALID_GRANT = "invalid_grant"

ONEDRIVE_INFO_DIR = os.path.join(ONEDRIVE_CONF_DIR, "info")
ONEDRIVE_BUSINESS_INFO_DIR = os.path.join(ONEDRIVE_BUSINESS_CONF_DIR, "info")


class PalOneDrive(palcmd.PalCmd):
    """PAL OneDrive command implementation class."""

    def __for_business(self, options):
        """Check forbusiness attribute is set to "true"."""

        if FORBUSINESS_ATTR not in options:
            return False

        if options[FORBUSINESS_ATTR].lower() == TRUE_TEXT:
            return True

        return False

    def __delete_share_info_dir(self, share_name, for_business):
        """Delete iofo directory for "share_name"."""

        info_dir = ONEDRIVE_INFO_DIR
        if for_business:
            info_dir = ONEDRIVE_BUSINESS_INFO_DIR

        try:
            shutil.rmtree(os.path.join(info_dir, share_name))
        except:
            pass

    def __make_session_path(self, share_name, for_business):
        """Make a session file path."""

        conf_base_dir = ONEDRIVE_CONF_DIR
        if for_business:
            conf_base_dir = ONEDRIVE_BUSINESS_CONF_DIR

        return os.path.join(conf_base_dir, SESSION_NAME,
                                share_name + SESSION_SUFFIX)

    def __make_download_target_conf(self, share_name):
        """Make a conf file written target share name to."""

        temp_path = DOWNLOAD_TARGET_CONF_PATH + "." + str(os.getpid())
        try:
            with open(temp_path, "wb") as file:
                file.write(share_name)

            shutil.move(temp_path, DOWNLOAD_TARGET_CONF_PATH)
        except:
            if os.path.exists(temp_path):
                os.remove(temp_path)
            raise

    def __get_token(self, session_path):
        """Get token in "ssession_path" fileh."""

        if not os.path.exists(session_path):
            return ""

        with open(session_path, "rb") as session_file:
            session_config = pickle.load(session_file)

        if hasattr(session_config, ACCESS_TOKEN_KEY):
            return getattr(session_config, ACCESS_TOKEN_KEY)
        return ""

    def __get_handler(self, entry):
        """Handle get request."""

        try:
            if self.__for_business(entry.get_options()):
                onedrive_access = OneDriveAccess(BUSINESS_APP_ID,
                                    BUSINESS_CLIENT_SECRET,
                                    scopes=SCOPES,
                                    redirect_url=BUSINESS_REDIRECT_URL,
                                    session_path="",
                                    auth_server_url=BUSINESS_AUTH_SERVER_URL,
                                    auth_token_url=BUSINESS_AUTH_TOKEN_URL,
                                    service_name=BUSINESS_SERVICE_NAME)
            else:
                onedrive_access = OneDriveAccess(APP_ID, CLIENT_SECRET,
                                        scopes=SCOPES,
                                        redirect_url=REDIRECT_URL,
                                        session_path="")

            entry.get().append_child(URL_NAME, onedrive_access.auth_url)
            entry.set_result(Results.OK)
        except:
            self.exception("get failed!!")
            entry.set_result(Results.GET_ERROR)

        return entry

    def __apply_handler(self, entry):
        """Handle apply request."""

        try:
            for_business = self.__for_business(entry.get_options())
            share_name = entry.get().get_node(SHARE_NAME_PATH).get_text()
            session_path = self.__make_session_path(share_name, for_business)
            new_code = entry.get().get_node(CODE_NAME).get_text()
            if not new_code:
                # Delete session file
                os.unlink(session_path)
                self.__delete_share_info_dir(share_name, for_business)
                entry.set_result(Results.OK)
                return entry

            proxy_host = entry.get().get_value(PROXY_HOST_PATH, None)
            proxy_port = entry.get().get_value(PROXY_PORT_PATH, None)

            proxies_dict = get_proxies_dict(proxy_host, proxy_port,
                                            PROTOCOL_LIST)
            if for_business:
                onedrive_access = OneDriveAccess(BUSINESS_APP_ID,
                                    BUSINESS_CLIENT_SECRET,
                                    scopes=SCOPES,
                                    redirect_url=BUSINESS_REDIRECT_URL,
                                    session_path=session_path,
                                    auth_server_url=BUSINESS_AUTH_SERVER_URL,
                                    auth_token_url=BUSINESS_AUTH_TOKEN_URL,
                                    service_name=BUSINESS_SERVICE_NAME,
                                    proxies=proxies_dict)
            else:
                onedrive_access = OneDriveAccess(APP_ID, CLIENT_SECRET,
                                        scopes=SCOPES,
                                        redirect_url=REDIRECT_URL,
                                        session_path=session_path,
                                        proxies=proxies_dict)

            onedrive_access.authenticate({CODE_NAME: new_code})
            self.__delete_share_info_dir(share_name, for_business)
            entry.set_result(Results.OK)

        except requests.exceptions.ConnectionError:
            self.exception("apply connection failed!!")
            entry.set_result(Results.CONNECT_ERROR)

        except Exception, exc_ins:
            if str(exc_ins) == ERROR_INVALID_GRANT:
                self.exception("apply auth failed!!")
                entry.set_result(Results.AUTH_ERROR)
            else:
                self.exception("apply failed!!")
                entry.set_result(Results.APPLY_ERROR)
        except:
            self.exception("apply failed!!")
            entry.set_result(Results.APPLY_ERROR)

        return entry

    def __query_handler(self, entry):
        """Handle query request."""

        try:
            for_business = self.__for_business(entry.get_options())
            share_name = entry.get().get_node(SHARE_NAME_PATH).get_text()
            session_path = self.__make_session_path(share_name, for_business)
            entry.get().append_child(TOKEN_NAME,
                                        self.__get_token(session_path))
            entry.set_result(Results.OK)
        except:
            self.exception("query failed!!")
            entry.set_result(Results.QUERY_ERROR)

        return entry

    def __download_handler(self, entry):
        """Handle download request."""

        try:
            share_name = entry.get().get_node(SHARENAME_NAME).get_text()
            self.__make_download_target_conf(share_name)

            CmdExec(SEND_DOWNLOAD_SIGNAL_ARGS).execute(
                stdout=NULL, stderr=NULL)
            entry.set_result(Results.OK)

        except ExecError:
            self.exception("Download Failed!!")
            entry.set_result(Results.DOWNLOAD_ERROR)
        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.GET, self.__get_handler)
        self.set_handler(ReqTypes.APPLY, self.__apply_handler)
        self.set_handler(ReqTypes.QUERY, self.__query_handler)
        self.set_handler(ReqTypes.DOWNLOAD, self.__download_handler)

#
# Main
#

if __name__ == "__main__":
    PalOneDrive(option=palcmd.PalCmdOption()).handle_request()
