#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""
Copyright (C) I-O DATA DEVICE, INC.  All rights reserved.
"""
from __future__ import with_statement

import os
import logging
import socket
import errno
import pwd
import grp

from boto.s3.connection import S3Connection, S3ResponseError

from product_conf import (CONF_DIR, TEMP_DIR, LOG_DIR,
                          STOR_IF_LIST, STOR_IF_TO_VOL)
from product_conf.devinfo import VOL_INFO
from iobase.exception import InvalidArgError, NotFoundError
from ionas.nascmd import NasCmd
from ionas.filesync import (LocalFileTree, AmazonS3FileTree)

from fal_info import (ReqTypes, Results, FAL_SHARES_PATH)

from fal_msginfo import Msg, Cat

from nasdsync_util import (get_proxy,
                           send_request,
                           _init_check,
                           _log_event,
                           _clear_announce,
                           get_tmp_path,
                           reset_proxy_info)

from ionas.filesync.sync import EconomySyncInfo, EconomySync, SyncMgr, debug

from cloudsync.util import (lf2crlf,)


NAME_NAME = "name"
PATH_NAME = "path"
SHARESERVICE_NAME = "shareservice"
AMAZONS3_PATH = SHARESERVICE_NAME + "/amazons3"
BUCKET_NAME = "bucket"
ACCESSKEY_NAME = "accesskey"
SECRETKEY_NAME = "secretkey"
HTTPMETHOD_NAME = "httpmethod"
STORAGECLASS_NAME = "storageclass"
DOWNLOADINTERVAL_NAME = "downloadinterval"
ENDPOINT_NAME = "endpoint"
S3_PORT_NAME = "port"
UPLOADMAXFILESIZE_NAME = "uploadmaxfilesize"
MPUPLOAD_NAME = "mpupload"

AMAZONS3_SYNC_INFO_DIR = os.path.join(CONF_DIR, "amazons3/info")

UID_NAME = "uid"
GID_NAME = "gid"

STATUS_GATEWAY_TIMEOUT = 504

ACCESSKEY_INFO = "accesskey"
SECRETKEY_INFO = "secretkey"
PROXY_HOST_INFO = "proxy_host"
PROXY_PORT_INFO = "proxy_port"
PORT_INFO = "port"
IS_SECURE_INFO = "is_secure"
BUCKET_INFO = "bucket"
ENDPOINT_INFO = "endpoint"
DOWNLOAD_TIME_INFO = "download_time"
UPLOADING_INFO = "uploading"

INTERNAL_NAME = STOR_IF_TO_VOL[STOR_IF_LIST[0]][1]
UPLOADED_PATH = os.path.join(VOL_INFO[INTERNAL_NAME]["mnt_point"],
                             "amazons3_uploaded")

MAX_FSIZE_UNIT = 1024 * 1024


# for logging ################################################################
from ionas.fileaccess import LocalAccess
from cloudsync.logger import Logger
import traceback as tb
import datetime as dt


class S3Sync(EconomySync):

    SUCCESS = "success"
    FAILED = "failed"
    #STOPPED = "stopped"

    def __init__(self, *args, **kwargs):
        super(S3Sync, self).__init__(*args, **kwargs)
        self._logger = Logger(self.info.share_name,
                              self.__logfilename(self.info.share_name),
                              LocalAccess(),
                              AmazonS3FileTree)

    def try_sync(self, *args, **kwargs):
        result = self.SUCCESS
        try:
            self.__init_sync_log()
            super(S3Sync, self).try_sync(*args, **kwargs)
        except:
            try:
                LOG_ABORT_UPLOAD = "///// Abort Upload ///// \r\n"
                self._logger.write(LOG_ABORT_UPLOAD + lf2crlf(tb.format_exc()),
                                   is_need_save=True)
            except:
                pass
            result = self.FAILED
            raise
        finally:
            #self.__end_handling()
            self.__end_log(result)

    def log_callback(self, log):
        super(S3Sync, self).log_callback(log)
        self._logger.write_log(log)
        self._logger._flush_and_renew_log_buf()

        # NOTE: これがないとログのファイル転送行より後がフラッシュされない
        self._logger.write('', is_need_save=True)

    def _try_upload(self, *args, **kwargs):
        LOG_START_UPLOAD = \
            "\r\n///// Start uploading files from local ==> cloud /////\r\n"
        LOG_END_UPLOAD = \
            "///// End uploading /////\r\n"

        self._logger.write(LOG_START_UPLOAD)
        super(S3Sync, self)._try_upload(*args, **kwargs)
        self._logger.write(LOG_END_UPLOAD)

    def _try_download(self, *args, **kwargs):
        LOG_START_DOWNLOAD = \
            "\r\n///// Start downloading files from cloud ==> local /////\r\n"
        LOG_END_DOWNLOAD = \
            "///// End downloading /////\r\n"

        if self.is_force_download:
            self._logger._renew_log_buf()
            self.__init_download_log()
        self._logger.write(LOG_START_DOWNLOAD)
        super(S3Sync, self)._try_download(*args, **kwargs)
        self._logger.write(LOG_END_DOWNLOAD)

    def __init_sync_log(self):
        now = dt.datetime.now()
        #log_file_path = os.path.join(SYNC_LOG_DIR,
        #                             self.info.share_name,
        #                             self.__get_sync_start_daytime_str(now))
        #self._logger.init_log_param(log_file_path)
        self._logger.init_log_param(self.__logfilename(self.info.share_name))
        self._logger.set_sync_start_time(now)
        self.__write_log_header(now)

    def __init_download_log(self):
        now = dt.datetime.now()
        #log_file_path = os.path.join(SYNC_LOG_DIR,
        #                             self.info.share_name,
        #                             self.__get_sync_start_daytime_str(now))
        #self._logger.init_log_param(log_file_path)
        self._logger.init_log_param(self.__logfilename(self.info.share_name))
        self._logger.set_sync_start_time(now)
        self.__write_log_header(now)

    def __write_log_header(self, daytime):
        LOG_TITLE =             ("****************************************\r\n"
                                 "   Sync with Amazon S3 Log              \r\n"
                                 "****************************************\r\n")
        LOG_LOCAL_SHARE_TMPL =   "Local Share    : '%s'\r\n"
        LOG_SERVER_FOLDER_TMPL = "Bucket         : '%s'\r\n"
        LOG_START_DAYTIME_TMPL = "Start Day Time : %s\r\n"

        self._logger.write(LOG_TITLE)
        self._logger.write(LOG_LOCAL_SHARE_TMPL % self.info.share_name)
        self._logger.write(LOG_SERVER_FOLDER_TMPL % self.info.bucket)
        self._logger.write(LOG_START_DAYTIME_TMPL % daytime.strftime("%Y/%m/%d %H:%M:%S"))

    def __end_log(self, result):
        try:
            self._logger.set_sync_end_time(dt.datetime.today())
            self._logger.end_log(result)
        except:
            pass

    def __logfilename(self, sharename):
        LAN_DISK_Log = 'LAN DISK Log'
        try:
            LAN_DISK_Log = Request(FAL_SHARES_PATH).request(Entry(ReqTypes.LOAD, """<shares/>""")).get_entry().get().get_value("sharelist/share[name='%s']/path" % LAN_DISK_Log)
        except:
            LAN_DISK_Log = "/mnt/sataraid1/share/%s" % LAN_DISK_Log

        service = 'Amazon S3'
        if self.info.endpoint == 'str.cloudn-service.com':  # FIXME
            service = 'Cloudn'

        now = dt.datetime.now()
        return os.path.join(LAN_DISK_Log,
                            service,
                            sharename,
                            "%04d%02d%02d_%02d%02d%02d" % (now.year,
                                                               now.month,
                                                               now.day,
                                                               now.hour,
                                                               now.minute,
                                                               now.second))

##############################################################################


def _check_last_target(self):
     # COMMENT 今回の対象bucket,対象endpointを記録しておき、
     # 次回起動時に前回起動時から
     # それらのいずれかが変わっていれば
     # 初回同期の動きをするようにする。前回起動時の
     # 対象bucketと対象endpointから不変なら何もしない

    info = self.info

    bucket_name = self._get_info(BUCKET_INFO)
    endpoint_name = self._get_info(ENDPOINT_INFO)

    if (bucket_name != info.bucket or
        endpoint_name != info.endpoint):
        self._delete_info_dir(info.share_name)

    self._set_info(BUCKET_INFO, info.bucket)
    self._set_info(ENDPOINT_INFO, info.endpoint)
    self._set_info(ACCESSKEY_INFO, info.accesskey)
    self._set_info(SECRETKEY_INFO, info.secretkey)
    self._set_info(PROXY_HOST_INFO, str(info.proxy_host))
    self._set_info(PROXY_PORT_INFO, str(info.proxy_port))
    self._set_info(PORT_INFO, str(info.port))
    self._set_info(IS_SECURE_INFO, str(info.is_secure))


def _post_success_upload(self):
    uploaded_path = os.path.join(UPLOADED_PATH, self.info.share_name)
    if os.path.exists(uploaded_path):
        os.unlink(uploaded_path)


def _set_trees(self):
    info = self.info

    uploading_path = os.path.join(AMAZONS3_SYNC_INFO_DIR,
                                  info.share_name,
                                  UPLOADING_INFO)
    uploaded_path = os.path.join(UPLOADED_PATH, info.share_name)

    self.local_tree = LocalFileTree(info.share_abs_path,
                                    tmp_path=get_tmp_path(info.share_abs_path),
                                    log_callback=self.log_callback)
    self.server_tree = AmazonS3FileTree(info.accesskey,
                                        info.secretkey,
                                        info.bucket,
                                        is_secure=info.is_secure,
                                        rrs=info.rrs,
                                        proxy_host=info.proxy_host,
                                        proxy_port=info.proxy_port,
                                        host=info.endpoint,
                                        port=info.port,
                                        max_file_size=info.up_max_fsize,
                                        mpupload=info.mpupload,
                                        log_callback=self.log_callback,
                                        uploading_path=uploading_path,
                                        uploaded_path=uploaded_path)


def _is_timeout_err(self, err):
    try:
        if isinstance(err, socket.gaierror):
            timeout_args_list = (
                socket.EAI_NONAME,
                socket.EAI_NODATA)

            if err.args[0] in timeout_args_list:
                return True

        elif isinstance(err, socket.timeout):
            return True

        elif isinstance(err, socket.sslerror):
            if err.args[0] == "The read operation timed out":
                return True

        elif isinstance(err, socket.error):
            timeout_args_list = (
                errno.ETIMEDOUT,
                errno.EHOSTUNREACH)

            if err.args[0] in timeout_args_list:
                return True

            try:
                if err.args[1].endswith("(Gateway Time-out)"):
                    return True
            except:
                pass

        elif isinstance(err, IOError):
            if isinstance(err.args[1], socket.timeout):
                return True

        elif isinstance(err, S3ResponseError):
            if err.status == STATUS_GATEWAY_TIMEOUT:
                return True

    except:
        return False
    return False


class AmazonS3SyncMgr(SyncMgr):

    INFO_DIR = AMAZONS3_SYNC_INFO_DIR

    SERVICE_ERR_LIST = [S3ResponseError]

    SHARE_CONF_SERVICE_ENABLE_PATH = \
        "sharelist/share/shareservice/amazons3[@enable='true']/../.."

    ANNOUNCE_CAT = Cat.AMAZONS3

    NASCMD_FOR_LOGGING = NasCmd(Results, logger_name="AMAZONS3",
        log=LOG_DIR + "/amazons3.log", log_level=logging.DEBUG)

    #SYNC_CLASS = EconomySync
    SYNC_CLASS = S3Sync

    class AmazonS3Info(EconomySyncInfo):

        def __init__(self, share_name, share_abs_path,
                     uid, gid, download_interval,
                     bucket, accesskey, secretkey, is_secure,
                     rrs, endpoint, proxy_host, proxy_port,
                     port, up_max_fsize, mpupload):
            super(self.__class__, self).__init__(share_name,
                                                 share_abs_path,
                                                 uid,
                                                 gid,
                                                 download_interval)
            self.bucket = bucket
            self.accesskey = accesskey
            self.secretkey = secretkey
            self.is_secure = is_secure
            self.rrs = rrs
            self.endpoint = endpoint
            self.proxy_host = proxy_host
            self.proxy_port = proxy_port
            self.port = port
            self.up_max_fsize = up_max_fsize
            self.mpupload = mpupload

    def _load_share_list(self):
        share_config = send_request(FAL_SHARES_PATH, ReqTypes.LOAD)
        try:
            return share_config.get_node_list(
                self.SHARE_CONF_SERVICE_ENABLE_PATH)
        except NotFoundError:
            return []

    def _get_share_user_group_id(self, share):

        nobody_uid = pwd.getpwnam("nobody")[2]
        nogroup_gid = grp.getgrnam("nogroup")[2]

        uid = share.get_value(UID_NAME, nobody_uid)
        gid = share.get_value(GID_NAME, nogroup_gid)

        return (int(uid), int(gid))

    def _get_sync_info(self, share):
        amazons3_node = share.get_node(AMAZONS3_PATH)

        if amazons3_node.get_node(HTTPMETHOD_NAME).get_text() == "http":
            is_secure = False
        elif amazons3_node.get_node(HTTPMETHOD_NAME).get_text() == "https":
            is_secure = True
        else:
            raise InvalidArgError(HTTPMETHOD_NAME)

        if amazons3_node.get_node(STORAGECLASS_NAME).get_text() == "STANDARD":
            rrs = False
        elif amazons3_node.get_node(STORAGECLASS_NAME).get_text() == "RRS":
            rrs = True
        else:
            raise InvalidArgError(STORAGECLASS_NAME)

        endpoint = amazons3_node.get_node(ENDPOINT_NAME).get_text()
        if not endpoint:
            endpoint = S3Connection.DefaultHost

        uid, gid = self._get_share_user_group_id(share)

        proxy_host, proxy_port = self._get_proxy(is_secure, endpoint)

        port = amazons3_node.get_value(S3_PORT_NAME, "")
        if port == "":
            port = None
        else:
            port = int(port)

        up_max_fsize = amazons3_node.get_value(UPLOADMAXFILESIZE_NAME, "")
        if up_max_fsize == "":
            up_max_fsize = None
        else:
            up_max_fsize = int(up_max_fsize) * MAX_FSIZE_UNIT

        try:
            mpupload = amazons3_node.get_node(MPUPLOAD_NAME).is_enabled()
        except NotFoundError:
            mpupload = True

        return self.AmazonS3Info(
            share.get_node(NAME_NAME).get_text(),
            share.get_node(PATH_NAME).get_text(),
            uid,
            gid,
            int(amazons3_node.get_node(DOWNLOADINTERVAL_NAME).get_text()),
            amazons3_node.get_node(BUCKET_NAME).get_text(do_strip=False),
            amazons3_node.get_node(ACCESSKEY_NAME).get_text(do_strip=False),
            amazons3_node.get_node(SECRETKEY_NAME).get_text(do_strip=False),
            is_secure,
            rrs,
            endpoint,
            proxy_host,
            proxy_port,
            port,
            up_max_fsize,
            mpupload)

    def _make_sync_dict(self, share):
        info = self._get_sync_info(share)

        return {info.share_name: self.SYNC_CLASS(
             info,
             service_err_list=self.SERVICE_ERR_LIST,
             info_dir=self.INFO_DIR,
             no_root_entry_err_log=Msg.ERR_AMAZONS3_BUCKET_NOT_EXIST,
             no_space_on_local_err_log=Msg.ERR_AMAZONS3_NO_SPACE_ON_LOCAL,
             no_space_on_server_err_log="",
             invalid_identifier_err_log=Msg.ERR_AMAZONS3_INVALID_ACCESSKEY,
             invalid_password_err_log=Msg.ERR_AMAZONS3_INVALID_SECRETKEY,
             invalid_fs_err_log=Msg.ERR_AMAZONS3_INVALID_FORMAT,
             file_too_large_err_log=Msg.ERR_AMAZONS3_FILE_TOO_LARGE,
             skip_err_log=Msg.ERR_AMAZONS3_SYNC_SKIP,
             timeout_err_log=Msg.ERR_AMAZONS3_TIMEOUT,
             misc_err_log=Msg.ERR_AMAZONS3_SYNC,
             nascmd_for_logging=self.NASCMD_FOR_LOGGING,
             _set_trees=_set_trees, _init_check=_init_check,
             _is_timeout_err=_is_timeout_err, _log_event=_log_event,
             _clear_announce=_clear_announce,
             reset_proxy_info=reset_proxy_info,
             _check_last_target=_check_last_target,
             _post_success_upload=_post_success_upload,
             download_target_fname=os.path.join(TEMP_DIR,
                                                "amazons3_download_target"))}

    def _get_sync_dict(self):
        sync_dict = {}
        for share in self._load_share_list():
            try:
                sync_dict.update(self._make_sync_dict(share))
            except:
                debug(do_tb=True)
                continue
        self._delete_disable_info_dir(sync_dict)
        return sync_dict

    def _get_proxy(self, is_secure, endpoint):
        if is_secure:
            url = "https://" + endpoint + "/"
        else:
            url = "http://" + endpoint + "/"

        proxy_info = get_proxy(url)

        if proxy_info:
            return tuple(proxy_info.split(":"))
        else:
            return (None, None)

    def _reset_proxy_info(self):
        for share in self.sync_dict:
            sync = self.sync_dict[share]
            host, port = self._get_proxy(sync.info.is_secure,
                                         sync.info.endpoint)
            sync.reset_proxy_info(host, port)

    def _get_info(self, share_name, info):
        try:
            with open(os.path.join(self.INFO_DIR,
                                   share_name,
                                   info),
                      "rb") as file:
                return file.read()
            return False
        except:
            return False

    def _pre_delete_disable_info_dir(self, share_name):
        bucket = self._get_info(share_name, BUCKET_INFO)
        endpoint = self._get_info(share_name, ENDPOINT_INFO)
        accesskey = self._get_info(share_name, ACCESSKEY_INFO)
        secretkey = self._get_info(share_name, SECRETKEY_INFO)
        proxy_host = self._get_info(share_name, PROXY_HOST_INFO)
        proxy_port = self._get_info(share_name, PROXY_PORT_INFO)
        port = self._get_info(share_name, PORT_INFO)
        is_secure = self._get_info(share_name, IS_SECURE_INFO)

        if proxy_host == "None":
            proxy_host = None

        if proxy_port == "None":
            proxy_port = None

        if port == "None":
            port = None
        else:
            port = int(port)

        if is_secure == "False":
            is_secure = False
        else:
            is_secure = True

        uploading_path = os.path.join(AMAZONS3_SYNC_INFO_DIR,
                                      share_name,
                                      UPLOADING_INFO)

        amazons3_tree = AmazonS3FileTree(accesskey, secretkey, bucket,
            is_secure=is_secure, proxy_host=proxy_host,
            proxy_port=proxy_port, host=endpoint, port=port,
            uploading_path=uploading_path)

        amazons3_tree.delete_all_mps()

    def _log_event(self, event, args={}, err_info=""):
        _log_event(self, event, args={}, err_info="")

    def _clear_announce(self):
        _clear_announce(self)
