#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""
Copyright (C) I-O DATA DEVICE, INC.  All rights reserved.
"""
from __future__ import with_statement

import os
import logging
import socket
import errno
import pwd
import grp

from azure import BLOB_SERVICE_HOST_BASE
from azure import WindowsAzureMissingResourceError
from azure.storage import WindowsAzureError
from azure.http import HTTPError

from product_conf import CONF_DIR, TEMP_DIR, LOG_DIR
from iobase.exception import NotFoundError
from ionas.nascmd import NasCmd
from ionas.filesync import (LocalFileTree, AzureFileTree)

from fal_info import (ReqTypes, Results, FAL_SHARES_PATH, FAL_NETWORK_PATH)

from fal_msginfo import Msg, Cat

from nasdsync_util import (get_proxy,
                           send_request,
                           _init_check,
                           _log_event,
                           _clear_announce,
                           get_tmp_path,
                           reset_proxy_info)



from ionas.filesync.sync import EconomySyncInfo, EconomySync, SyncMgr, debug

from cloudsync.util import (lf2crlf,)


NAME_NAME = "name"
PATH_NAME = "path"
SHARESERVICE_NAME = "shareservice"
AZURE_PATH = SHARESERVICE_NAME + "/azure"
CONTAINER_NAME = "container"
ACCOUNTNAME_NAME = "accountname"
ACCESSKEY_NAME = "accesskey"
HTTPMETHOD_NAME = "httpmethod"
STORAGECLASS_NAME = "storageclass"
DOWNLOADINTERVAL_NAME = "downloadinterval"
ENDPOINT_NAME = "endpoint"

AZURE_SYNC_INFO_DIR = os.path.join(CONF_DIR, "azure/info")

UID_NAME = "uid"
GID_NAME = "gid"

CONTAINER_INFO = "container"
ENDPOINT_INFO = "endpoint"

STATUS_GATEWAY_TIMEOUT = 504

ENDPOINT_URL_FMT = "%s://%s%s/"

MAC_ADDR_PATH = "netiflist/netif[@name='lan1']/mac"


# for logging ################################################################
from ionas.fileaccess import LocalAccess
from cloudsync.logger import Logger
import traceback as tb
import datetime as dt


class AzureSync(EconomySync):

    SUCCESS = "success"
    FAILED = "failed"
    #STOPPED = "stopped"

    def __init__(self, *args, **kwargs):
        super(AzureSync, self).__init__(*args, **kwargs)
        self._logger = Logger(self.info.share_name,
                              self.__logfilename(self.info.share_name),
                              LocalAccess(),
                              AzureFileTree)

    def try_sync(self, *args, **kwargs):
        result = self.SUCCESS
        try:
            self.__init_sync_log()
            super(AzureSync, self).try_sync(*args, **kwargs)
        except:
            try:
                LOG_ABORT_UPLOAD = "///// Abort Upload ///// \r\n"
                self._logger.write(LOG_ABORT_UPLOAD + lf2crlf(tb.format_exc()),
                                   is_need_save=True)
            except:
                pass
            result = self.FAILED
            raise
        finally:
            #self.__end_handling()
            self.__end_log(result)

    def log_callback(self, log):
        super(AzureSync, self).log_callback(log)
        self._logger.write_log(log)
        self._logger._flush_and_renew_log_buf()

        # NOTE: これがないとログのファイル転送行より後がフラッシュされない
        self._logger.write('', is_need_save=True)

    def _try_upload(self, *args, **kwargs):
        LOG_START_UPLOAD = \
            "\r\n///// Start uploading files from local ==> cloud /////\r\n"
        LOG_END_UPLOAD = \
            "///// End uploading /////\r\n"

        self._logger.write(LOG_START_UPLOAD)
        super(AzureSync, self)._try_upload(*args, **kwargs)
        self._logger.write(LOG_END_UPLOAD)

    def _try_download(self, *args, **kwargs):
        LOG_START_DOWNLOAD = \
            "\r\n///// Start downloading files from cloud ==> local /////\r\n"
        LOG_END_DOWNLOAD = \
            "///// End downloading /////\r\n"

        if self.is_force_download:
            self._logger._renew_log_buf()
            self.__init_download_log()
        self._logger.write(LOG_START_DOWNLOAD)
        super(AzureSync, self)._try_download(*args, **kwargs)
        self._logger.write(LOG_END_DOWNLOAD)

    def __init_sync_log(self):
        now = dt.datetime.now()
        #log_file_path = os.path.join(SYNC_LOG_DIR,
        #                             self.info.share_name,
        #                             self.__get_sync_start_daytime_str(now))
        #self._logger.init_log_param(log_file_path)
        self._logger.init_log_param(self.__logfilename(self.info.share_name))
        self._logger.set_sync_start_time(now)
        self.__write_log_header(now)

    def __init_download_log(self):
        now = dt.datetime.now()
        #log_file_path = os.path.join(SYNC_LOG_DIR,
        #                             self.info.share_name,
        #                             self.__get_sync_start_daytime_str(now))
        #self._logger.init_log_param(log_file_path)
        self._logger.init_log_param(self.__logfilename(self.info.share_name))
        self._logger.set_sync_start_time(now)
        self.__write_log_header(now)

    def __write_log_header(self, daytime):
        LOG_TITLE =             ("****************************************\r\n"
                                 "   Sync with Microsoft Azure Log        \r\n"
                                 "****************************************\r\n")
        LOG_LOCAL_SHARE_TMPL =   "Local Share    : '%s'\r\n"
        #LOG_SERVER_FOLDER_TMPL = "Server Folder  : '%s'\r\n"
        LOG_SERVER_FOLDER_TMPL = "Container      : '%s' (%s)\r\n"
        LOG_START_DAYTIME_TMPL = "Start Day Time : %s\r\n"

        self._logger.write(LOG_TITLE)
        self._logger.write(LOG_LOCAL_SHARE_TMPL % self.info.share_name)
        #self._logger.write(LOG_SERVER_FOLDER_TMPL % self.info.root)
        #self._logger.write(LOG_SERVER_FOLDER_TMPL % self.info.container)
        self._logger.write(LOG_SERVER_FOLDER_TMPL % (self.info.container,
                                                     self.info.accountname))
        self._logger.write(LOG_START_DAYTIME_TMPL % daytime.strftime("%Y/%m/%d %H:%M:%S"))

    def __end_log(self, result):
        try:
            self._logger.set_sync_end_time(dt.datetime.today())
            self._logger.end_log(result)
        except:
            pass

    def __logfilename(self, sharename):
        LAN_DISK_Log = 'LAN DISK Log'
        try:
            LAN_DISK_Log = Request(FAL_SHARES_PATH).request(Entry(ReqTypes.LOAD, """<shares/>""")).get_entry().get().get_value("sharelist/share[name='%s']/path" % LAN_DISK_Log)
        except:
            LAN_DISK_Log = "/mnt/sataraid1/share/%s" % LAN_DISK_Log
        now = dt.datetime.now()
        return os.path.join(LAN_DISK_Log,
                            'Microsoft Azure',
                            sharename,
                            "%04d%02d%02d_%02d%02d%02d" % (now.year,
                                                               now.month,
                                                               now.day,
                                                               now.hour,
                                                               now.minute,
                                                               now.second))

##############################################################################


def _check_last_target(self):
    """
    今回の対象container,対象endpointを記録しておき、
    次回起動時に前回起動時から
    それらのいずれかが変わっていれば
    初回同期の動きをするようにする。前回起動時の
    対象containerと対象endpointから不変なら何もしない
    """

    container_name = self._get_info(CONTAINER_INFO)
    if container_name != self.info.container:
        self._delete_info_dir(self.info.share_name)
        self._set_info(CONTAINER_INFO, self.info.container)


def _set_trees(self):
    info = self.info
    self.local_tree = LocalFileTree(info.share_abs_path,
                                    tmp_path=get_tmp_path(info.share_abs_path),
                                    log_callback=self.log_callback)

    self.server_tree = AzureFileTree(info.accountname,
                                    info.accesskey,
                                    info.container,
                                    protocol=info.protocol,
                                    proxy_host=info.proxy_host,
                                    proxy_port=info.proxy_port,
                                    unique_info=info.unique_info,
                                    log_callback=self.log_callback)


def _is_timeout_err(self, err):
    try:
        if isinstance(err, socket.gaierror):
            timeout_args_list = (
                socket.EAI_NONAME,
                socket.EAI_NODATA)

            if err.args[0] in timeout_args_list:
                return True

        elif isinstance(err, socket.timeout):
            return True

        elif isinstance(err, socket.sslerror):
            if err.args[0] == "The read operation timed out":
                return True

        elif isinstance(err, socket.error):
            timeout_args_list = (
                errno.ETIMEDOUT,
                errno.EHOSTUNREACH)

            if err.args[0] in timeout_args_list:
                return True

            try:
                if err.args[1].endswith("(Gateway Time-out)"):
                    return True
            except:
                pass

        elif isinstance(err, IOError):
            if isinstance(err.args[1], socket.timeout):
                return True

    except:
        return False
    return False


class AzureSyncMgr(SyncMgr):

    INFO_DIR = AZURE_SYNC_INFO_DIR

    SERVICE_ERR_LIST = [WindowsAzureMissingResourceError,
                        WindowsAzureError,
                        HTTPError]

    SHARE_CONF_SERVICE_ENABLE_PATH = \
        "sharelist/share/shareservice/azure[@enable='true']/../.."

    ANNOUNCE_CAT = Cat.AZURE

    NASCMD_FOR_LOGGING = NasCmd(Results, logger_name="AZURE",
        log=LOG_DIR + "/azure.log", log_level=logging.DEBUG)

    #SYNC_CLASS = EconomySync
    SYNC_CLASS = AzureSync

    class AzureInfo(EconomySyncInfo):

        def __init__(self, share_name, share_abs_path,
                     uid, gid, download_interval,
                     container, accountname, accesskey, protocol,
                     proxy_host, proxy_port, unique_info):
            super(self.__class__, self).__init__(share_name,
                                                 share_abs_path,
                                                 uid,
                                                 gid,
                                                 download_interval)
            self.container = container
            self.accountname = accountname
            self.accesskey = accesskey
            self.protocol = protocol
            self.proxy_host = proxy_host
            self.proxy_port = proxy_port
            self.unique_info = unique_info

    #コンストラクタ引数で渡すべき?
    #UPDATE_PROPERTIES = {"INFO_DIR": INFO_DIR,
    #                     "SERVICE_ERR_LIST": SERVICE_ERR_LIST,
    #                     "FILE_TOO_LARGE_ERR_LOG":
    #                          Msg.ERR_AZURE_FILE_TOO_LARGE,
    #                     "SKIP_ERR_LOG": Msg.ERR_AZURE_SYNC_SKIP,
    #                     "NASCMD_FOR_LOGGING": NASCMD_FOR_LOGGING,
    #                     "_check_last_target": _check_last_target,
    #                     "_set_trees": _set_trees,
    #                     "reset_proxy_info": reset_proxy_info}

    def _load_share_list(self):
        share_config = send_request(FAL_SHARES_PATH, ReqTypes.LOAD)
        try:
            return share_config.get_node_list(
                self.SHARE_CONF_SERVICE_ENABLE_PATH)
        except NotFoundError:
            return []

    def _get_share_user_group_id(self, share):

        nobody_uid = pwd.getpwnam("nobody")[2]
        nogroup_gid = grp.getgrnam("nogroup")[2]

        uid = share.get_value(UID_NAME, nobody_uid)
        gid = share.get_value(GID_NAME, nogroup_gid)

        return (int(uid), int(gid))

    def _get_sync_info(self, share):
        azure_node = share.get_node(AZURE_PATH)

        protocol = azure_node.get_node(HTTPMETHOD_NAME).get_text()
        container = azure_node.get_node(CONTAINER_NAME).get_text()

        uid, gid = self._get_share_user_group_id(share)
        proxy_host, proxy_port = self._get_proxy(protocol, container)
        unique_info = self._get_product_unique_info()

        return self.AzureInfo(
            share.get_node(NAME_NAME).get_text(),
            share.get_node(PATH_NAME).get_text(),
            uid,
            gid,
            int(azure_node.get_node(DOWNLOADINTERVAL_NAME).get_text()),
            container,
            azure_node.get_node(ACCOUNTNAME_NAME).get_text(do_strip=False),
            azure_node.get_node(ACCESSKEY_NAME).get_text(do_strip=False),
            protocol,
            proxy_host,
            proxy_port,
            unique_info)

    def _make_sync_dict(self, share):
        info = self._get_sync_info(share)
        return {info.share_name: self.SYNC_CLASS(
                 info,
                 service_err_list=self.SERVICE_ERR_LIST,
                 info_dir=self.INFO_DIR,
                 no_root_entry_err_log=Msg.ERR_AZURE_CONTAINER_NOT_EXIST,
                 no_space_on_local_err_log=Msg.ERR_AZURE_NO_SPACE_ON_LOCAL,
                 no_space_on_server_err_log="",
                 invalid_identifier_err_log="",
                 invalid_password_err_log=Msg.ERR_AZURE_INVALID_ACCESSKEY,
                 invalid_fs_err_log=Msg.ERR_AZURE_INVALID_FORMAT,
                 file_too_large_err_log=Msg.ERR_AZURE_FILE_TOO_LARGE,
                 skip_err_log=Msg.ERR_AZURE_SYNC_SKIP,
                 timeout_err_log=Msg.ERR_AZURE_TIMEOUT,
                 misc_err_log=Msg.ERR_AZURE_SYNC,
                 nascmd_for_logging=self.NASCMD_FOR_LOGGING,
                 _set_trees=_set_trees, _init_check=_init_check,
                 _is_timeout_err=_is_timeout_err, _log_event=_log_event,
                 _clear_announce=_clear_announce,
                 reset_proxy_info=reset_proxy_info,
                 _check_last_target=_check_last_target,
                 download_target_fname=os.path.join(TEMP_DIR,
                                                    "azure_download_target"))}

    def _get_sync_dict(self):
        sync_dict = {}
        for share in self._load_share_list():
            try:
                sync_dict.update(self._make_sync_dict(share))
            except:
                debug(do_tb=True)
                continue
        self._delete_disable_info_dir(sync_dict)
        return sync_dict

    def _get_proxy(self, protocol, container):
        url = ENDPOINT_URL_FMT % (protocol, container, BLOB_SERVICE_HOST_BASE)
        proxy_info = get_proxy(url)

        if proxy_info:
            return tuple(proxy_info.split(":"))
        else:
            return (None, None)

    def _reset_proxy_info(self):
        for share in self.sync_dict:
            sync = self.sync_dict[share]
            host, port = self._get_proxy(sync.info.protocol,
                                         sync.info.container)
            sync.reset_proxy_info(host, port)

    def _log_event(self, event, args={}, err_info=""):
        _log_event(self, event, args={}, err_info="")

    def _clear_announce(self):
        _clear_announce(self)

    def _get_product_unique_info(self):
        net_info = send_request(FAL_NETWORK_PATH, ReqTypes.QUERY, "<network/>")
        mac_addr = net_info.get_value(MAC_ADDR_PATH)
        return "".join(mac_addr.split(":"))
