#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""
Sync module for the BOX cloud service.

Copyright (C) I-O DATA DEVICE, INC.  All rights reserved.
"""
from __future__ import with_statement

try:
    import json
except:
    import simplejson as json

import os
import time
import shutil
import socket
import errno
import pwd
import grp
try:
    import cloudsync.box.urllib2 as urllib2
except:
    import urllib2
try:
    import cloudsync.box.httplib as httplib
except:
    import httplib
import logging
import traceback
from sys import (exc_info,)
from StringIO import StringIO
from datetime import (datetime,)

from product_conf import (CONF_DIR, TEMP_DIR, LOG_DIR,)
from iobase.exception import (NotFoundError,)
from ionas.exception import (BaseAccessException,
                             NoEntryError, NoRootEntryError,
                             InvalidFileSystemError, NoSpaceError,
                             ExistsError, SizeTooLargeError,
                             InternalError, NameTooLongError,
                             PermissionError)
from ionas.fileaccess import (LocalAccess,)
from ionas.filesync import (SKIP_ERROR_LOG_LV, FILE_SIZE_OVER_LOG)
from ionas.filesync.sync import (EconomySyncInfo, EconomySyncEx, SyncMgr,
                                    debug, DOWNLOAD_TIME_INFO)

from fal_msginfo import (Msg, Cat,)
from fal_info import (ReqTypes, Results, FAL_SHARES_PATH,)
from fal_util import (save_event_log, clear_announce,)
from cloudsync.util import (lf2crlf, recreate_http_error)
from cloudsync.logger import (Logger, InternalLog)
from cloudsync.boxinfo import (LoginError,)
from cloudsync.boxlocalfiletree import (BoxLocalFileTree, EQUIVALENT_SKIP_LOG)
from cloudsync.boxfiletree import (BoxFileTree, OperationLocked,
                                    ForbiddenNameError)
from nasdsync_util import (get_sync_log_dir_path, is_share_available,
                                get_tmp_path, get_proxy, send_request,)
from pkgaddon.box_config import (REST_BASE_URL, ROOT_PATH,
                                    DOWNLOAD_TARGET_CONF_PATH)

LOGGER_NAME = "BOX"
LOG_FILE_PATH = os.path.join(LOG_DIR, "box.log")

LOG_DIR_NAME = "Box"
SYNC_LOG_DIR = os.path.join(get_sync_log_dir_path(), LOG_DIR_NAME)

UPLOADING = "uploading"
DOWNLOADING = "downloading"

RETRY_PERIOD = 600
RECOVER_PERIOD = 10

NO_ROOT_ENTRY_ERR_LOG = 1
NO_SPACE_ON_LOCAL_ERR_LOG = 2
NO_SPACE_ON_SERVER_ERR_LOG = 3
INVALID_IDENTIFIER_ERR_LOG = 4
INVALID_PASSWORD_ERR_LOG = 5
INVALID_FS_ERR_LOG = 6
FILE_TOO_LARGE_ERR_LOG = 7
SKIP_ERR_LOG = 8
TIMEOUT_ERR_LOG = 9
MISC_ERR_LOG = 10
AUTH_ERR_LOG = 11
FORBIDDEN_FILE_NAME_ERR_LOG = 12
NO_PERMISSION_ERR_LOG = 13
LOCKED_ERR_LOG = 14
CONNECT_ERR_LOG = 15
SAME_FILE_NAME_EXISTS_ERR_LOG = 16
TOO_LONG_FILE_NAME_ERR_LOG = 17

ERR_LOG_MAP = {
        NO_ROOT_ENTRY_ERR_LOG: Msg.ERR_BOX_FOLDER_NOT_FOUND,
        NO_SPACE_ON_LOCAL_ERR_LOG: Msg.ERR_BOX_NO_SPACE_ON_LOCAL,
        NO_SPACE_ON_SERVER_ERR_LOG: Msg.ERR_BOX_NO_SPACE_ON_SERVER,
        INVALID_FS_ERR_LOG: Msg.ERR_BOX_INVALID_FORMAT,
        FILE_TOO_LARGE_ERR_LOG: Msg.ERR_BOX_TOO_LARGE_FILE,
        SKIP_ERR_LOG: Msg.ERR_BOX_SYNC_SKIP,
        TIMEOUT_ERR_LOG: Msg.ERR_BOX_TIMEOUT,
        MISC_ERR_LOG: Msg.ERR_BOX_SYNC,
        AUTH_ERR_LOG: Msg.ERR_BOX_AUTH,
        FORBIDDEN_FILE_NAME_ERR_LOG: Msg.ERR_BOX_FORBIDDEN_FILE_NAME,
        NO_PERMISSION_ERR_LOG: Msg.ERR_BOX_NO_PERMISSION,
        LOCKED_ERR_LOG: Msg.ERR_BOX_LOCKED,
        CONNECT_ERR_LOG: Msg.ERR_BOX_CONNECT,
        SAME_FILE_NAME_EXISTS_ERR_LOG: Msg.ERR_BOX_SAME_FILE_NAME_EXISTS,
        TOO_LONG_FILE_NAME_ERR_LOG: Msg.ERR_BOX_TOO_LONG_FILE_NAME,
    }


JSON_ERR_CODE_KEY = "code"

err_code_except_dict = {
    "item_name_in_use": ExistsError,
    "forbidden": PermissionError,
    "access_denied_insufficient_permissions": PermissionError,
    "not_found": NoEntryError,
    "access_denied_item_locked": OperationLocked,
    "storage_limit_exceeded": NoSpaceError,
    "file_size_limit_exceeded": SizeTooLargeError,
    "item_name_invalid": ForbiddenNameError,
    }

NOBODY_NAME = "nobody"
NOGROUP_NAME = "nogroup"

SYNC_DIR_FMT_STR = "%04d%02d%02d_%02d%02d%02d"

SHARE_CONF_SERVICE_ENABLE_SHARE_PATH = \
        "sharelist/share/shareservice/box[@enable='true']/../.."
SHARE_BOX_XPATH = "shareservice/box"

NAME_NAME = "name"
PATH_NAME = "path"
DOWNLOADINTERVAL_NAME = "downloadinterval"
UPLOADMAXFILESIZE_NAME = "uploadmaxfilesize"
MAILADDRESS_NAME = "mailaddress"
PASSWD_NAME = "passwd"

BOX_SYNC_INFO_DIR = os.path.join(CONF_DIR, "box/info")

UID_NAME = "uid"
GID_NAME = "gid"

TARGET_ROOT_INFO = "target_root"
MAILADDRESS_INFO = "mailaddress"

SUCCESS = "success"
FAILED = "failed"
STOPPED = "stopped"

LOG_TITLE = \
"****************************************\r\n" + \
"   BOX Sync Log\r\n" + \
"****************************************\r\n"

LOG_LOCAL_SHARE_TMPL = "Local Share    : '%s'\r\n"
LOG_MAILADDRESS_TMPL = "Mail address   : '%s'\r\n"
LOG_START_DAYTIME_TMPL = "Start Day Time : %s\r\n"

LOG_START_UPLOAD = \
    "\r\n///// Start uploading files from local ==> cloud /////\r\n"

LOG_END_UPLOAD = \
    "///// End uploading /////\r\n"

LOG_START_DOWNLOAD = \
    "\r\n///// Start downloading files from cloud ==> local /////\r\n"

LOG_END_DOWNLOAD = \
    "///// End downloading /////\r\n"

LOG_CANCEL_SYNC = "///// Cancel Sync /////\r\n"

LOG_EXACT_FIRST_UPLOAD = "///// Exact First Upload /////\r\n"
LOG_FIRST_UPLOAD = "///// First Upload /////\r\n"
LOG_NORMAL_UPLOAD = "///// Normal Upload /////\r\n"
LOG_ABORT_UPLOAD = "///// Abort Upload ///// \r\n"

LOG_EXACT_FIRST_DOWNLOAD = "///// Exact First Download ///// \r\n"
LOG_FIRST_DOWNLOAD = "///// First Download ///// \r\n"
LOG_NORMAL_DOWNLOAD = "///// Normal Download ///// \r\n"
LOG_ABORT_DOWNLOAD = "///// Abort Download ///// \r\n"

RESULT_DIR_NAME = "result"


class ServerCommunicationError(BaseAccessException):
    pass


class BoxCallBack():
    """Call back class for a BoxSync class."""

    def __init__(self, box_sync):
        """Constructor."""

        self._box_sync = box_sync

    def low_level_except_handler(self, err, target_path=None):
        """Low level exception handler."""

        if isinstance(err, urllib2.HTTPError):
            if httplib.UNAUTHORIZED == err.code:
                raise ServerCommunicationError(detail=err)

            if not hasattr(err, "read"):
                raise err

            err_data = err.read()
            err_code = self._box_sync.get_http_err_code(err_data)
            new_err = recreate_http_error(err, err_data)
            if err_code not in err_code_except_dict:
                raise new_err

            if target_path is None:
                raise err_code_except_dict[err_code](detail=new_err)

            # Handle if no root directory
            if self._box_sync.info.root == target_path:
                if err_code_except_dict[err_code] == NoEntryError:
                    raise NoRootEntryError(detail=new_err)

                if err_code_except_dict[err_code] == ForbiddenNameError:
                    raise NoRootEntryError(detail=new_err)

            raise err_code_except_dict[err_code](detail=new_err)

        elif (isinstance(err, UnicodeError) or
              isinstance(err, UnicodeDecodeError) or
              isinstance(err, UnicodeEncodeError)):
            raise ForbiddenNameError(detail=err)

        elif isinstance(err, urllib2.URLError):
            raise err

        raise err

    def log_callback(self, log):
        """Call back method for log a "log" Log class instance."""

        self._box_sync.write_log(log)


class BoxSync(EconomySyncEx):
    """Sync class for the BOX cloud service."""

    SERVICE_ERR_LIST = []

    def _get_sync_start_daytime_str(self, now):
        """Return sync start day and time string."""

        return SYNC_DIR_FMT_STR % (
            now.year, now.month, now.day, now.hour, now.minute, now.second)

    def _is_timeout_exception(self, exc):
        """Check if exception is for timeout."""

        try:
            if isinstance(exc, socket.gaierror):
                timeout_args_list = (
                    socket.EAI_NONAME,
                    socket.EAI_NODATA)

                if exc.args[0] in timeout_args_list:
                    return True

            elif isinstance(exc, socket.timeout):
                return True

            elif isinstance(exc, socket.sslerror):
                if exc.args[0] == "The read operation timed out":
                    return True

            elif isinstance(exc, socket.error):
                timeout_args_list = (
                    errno.ETIMEDOUT,
                    errno.EHOSTUNREACH)

                if exc.args[0] in timeout_args_list:
                    return True

                try:
                    if exc.args[1].endswith("(Gateway Time-out)"):
                        return True
                except:
                    pass

            elif isinstance(exc, IOError):
                if isinstance(exc.args[1], socket.timeout):
                    return True

        except:
            return False
        return False

    def _is_misc_recover_err(self, exc):
        """Check if "exc" is recover error."""

        # Assume a recover errro.
        return True

    def _get_ev_log(self, exc):
        """Return evnet log messege code."""

        if self._is_timeout_exception(exc):
            debug(["!!! timeout err !!!"])
            return ERR_LOG_MAP[TIMEOUT_ERR_LOG]

        if isinstance(exc, urllib2.HTTPError) or \
           isinstance(exc, urllib2.URLError):
            return ERR_LOG_MAP[CONNECT_ERR_LOG]

        if hasattr(exc, "detail"):
            if isinstance(exc.detail, urllib2.HTTPError) or \
               isinstance(exc.detail, urllib2.URLError):
                return ERR_LOG_MAP[CONNECT_ERR_LOG]

        if self._is_misc_recover_err(exc):
            debug(["!!! misc recover err !!!"])
            return ERR_LOG_MAP[MISC_ERR_LOG]

        return None

    def _log_event(self, event, args={}, err_info=""):
        """Log "event" event."""

        ROOT_UID = 0
        ROOT_GID = 0

        if not event:
            return

        if err_info:
            self.internal_log.error(str(err_info))

        try:
            save_event_log(event, args)
        except:
            try:
                pre_uid = os.geteuid()
                pre_gid = os.getegid()
                os.setegid(ROOT_GID)
                os.seteuid(ROOT_UID)
                save_event_log(event, args)
            finally:
                os.setegid(pre_gid)
                os.seteuid(pre_uid)

    def _init_check(self):
        """Check if target share is available."""

        is_share_available(self.info.share_name, self.info.share_abs_path)

    def _set_trees(self):
        """Set local and server tree class instance."""

        self._local_tree = \
            BoxLocalFileTree(self.info.share_abs_path,
                               tmp_path=get_tmp_path(self.info.share_abs_path),
                               log_callback=self._callback.log_callback)
        self._server_tree = \
            BoxFileTree(token_id=self.info.share_name,
                           mailaddress=self.info.mailaddress,
                           passwd=self.info.passwd,
                           root=self.info.root,
                           uploadmaxfilesize=self.info.uploadmaxfilesize,
                           proxy_host=self.info.proxy_host,
                           proxy_port=self.info.proxy_port,
                           misc_callback=self._callback)

    def _recover_err_handle(self, exc, do_recover):
        share_name = self.info.share_name

        ev_log = self._get_ev_log(exc)

        if not ev_log:
            return False

        if do_recover:
            debug(["!!! do recover !!!"])
            time.sleep(RECOVER_PERIOD)
            return True

        self._log_event(ev_log, {"name": share_name})

        self.is_need_retry_sleep = True

        raise exc, None, exc_info()[2]

    def __no_retry_err_handle(self, exc):
        share_name = self.info.share_name

        if isinstance(exc, InvalidFileSystemError):
            self._log_event(ERR_LOG_MAP[INVALID_FS_ERR_LOG],
                            {"name": share_name})
        elif isinstance(exc, NoRootEntryError):
            if exc.detail != None:
                return

            # Don't retry.  Maybe volume is not mount.
        else:
            return

        self._set_disable_share(share_name)

        debug(["!!! no retry err !!!"])

        raise exc, None, exc_info()[2]

    def _retry_err_handle(self, exc):
        share_name = self.info.share_name
        if isinstance(exc, NoSpaceError) and \
                 self.current_direction == DOWNLOADING:
            self._log_event(ERR_LOG_MAP[NO_SPACE_ON_LOCAL_ERR_LOG],
                     {"name": share_name})

        elif isinstance(exc, NoSpaceError) and \
               isinstance(exc.detail, urllib2.HTTPError):
            self._log_event(ERR_LOG_MAP[NO_SPACE_ON_SERVER_ERR_LOG],
                     {"name": self.info.mailaddress + ":" + self.info.root})

        elif (isinstance(exc, ServerCommunicationError) or
                isinstance(exc, LoginError)):

            self._log_event(ERR_LOG_MAP[AUTH_ERR_LOG],
                     {"name": share_name})

        elif isinstance(exc, NoRootEntryError):
            self._log_event(ERR_LOG_MAP[NO_ROOT_ENTRY_ERR_LOG],
                    {"name": self.info.mailaddress + ":" + self.info.root})
        else:
            return

        debug(["!!! retry err !!!"])
        self.is_need_retry_sleep = True
        raise exc, None, exc_info()[2]

    def _misc_err_handle(self, exc):
        self._log_event(ERR_LOG_MAP[MISC_ERR_LOG],
                        {"name": self.info.share_name},
                        traceback.format_exc())
        raise exc, None, exc_info()[2]

    def _init_except_handler(self, exc):

        self.__no_retry_err_handle(exc)
        self._misc_err_handle(exc)

    def _access_except_handler(self, exc, is_do_recover=False):

        self.__no_retry_err_handle(exc)
        self._retry_err_handle(exc)
        if self._recover_err_handle(exc, is_do_recover):
            return True
        self._misc_err_handle(exc)
        return False

    def _delete_info_dir(self, share_name):
        pass

    def _clear_info_dir(self, share_name):
        try:
            shutil.rmtree(
                os.path.join(self.info.info_dir, share_name))
        except:
            pass

    def _get_download_target_fname(self):

        return os.path.join(TEMP_DIR, "box_download_target")

    def _check_last_target(self):

        info = self.info

        tgt_root_name = self._get_info(TARGET_ROOT_INFO)
        mailaddress_name = self._get_info(MAILADDRESS_INFO)
        if not tgt_root_name:
            self._set_info(TARGET_ROOT_INFO, info.root)
        if not mailaddress_name:
            self._set_info(MAILADDRESS_INFO, info.mailaddress)
        if (tgt_root_name != info.root or
            mailaddress_name != info.mailaddress):
            self._clear_info_dir(info.share_name)
            self._set_info(TARGET_ROOT_INFO, info.root)
            self._set_info(MAILADDRESS_INFO, info.mailaddress)

    def _write_log_header(self, daytime):

        self._logger.write(LOG_TITLE)
        self._logger.write(LOG_LOCAL_SHARE_TMPL % self.info.share_name)
        self._logger.write(LOG_MAILADDRESS_TMPL % self.info.mailaddress)
        self._logger.write(LOG_START_DAYTIME_TMPL % \
                               daytime.strftime("%Y/%m/%d %H:%M:%S"))

    def __init_sync_log(self):
        """Initialize sync log."""

        now = datetime.now()
        log_file_path = os.path.join(SYNC_LOG_DIR,
                                     self.info.share_name,
                                     self._get_sync_start_daytime_str(now))
        self._logger.init_log_param(log_file_path)
        self._logger.set_sync_start_time(now)
        self._write_log_header(now)

    def __init_download_log(self):

        now = datetime.now()
        log_file_path = os.path.join(SYNC_LOG_DIR,
                                     self.info.share_name,
                                     self._get_sync_start_daytime_str(now))
        self._logger.init_log_param(log_file_path)
        self._logger.set_sync_start_time(now)
        self._write_log_header(now)

    def __end_log(self, result):

        try:
            self._logger.set_sync_end_time(datetime.today())
            self._logger.end_log(result)
        except:
            pass

    def __init__(self, info):
        """Constructor."""

        self.info = info
        self.current_direction = None
        self.is_need_retry_sleep = False
        self.internal_log = InternalLog(LOGGER_NAME, LOG_FILE_PATH,
                                        logging.DEBUG)
        now = datetime.now()
        log_file_path = os.path.join(SYNC_LOG_DIR,
                                     self.info.share_name,
                                     self._get_sync_start_daytime_str(now))
        self._logger = \
            Logger(self.info.share_name, log_file_path, LocalAccess(),
                    BoxFileTree)
        self._callback = BoxCallBack(self)

        super(BoxSync, self).__init__(info)

    def get_http_err_code(self, err_data):
        """Retrun error code in response JSON data."""

        try:
            return json.loads(err_data)[JSON_ERR_CODE_KEY]
        except:
            return None

    def _save_corresponding_event_log(self, log):
        """Arg log is Log class instance."""

        if isinstance(log.exc_arg, NoSpaceError):
            if isinstance(log.exc_arg.detail, urllib2.HTTPError):
                self._log_event(ERR_LOG_MAP[NO_SPACE_ON_SERVER_ERR_LOG],
                        {"name": self.info.mailaddress + ":" + self.info.root})

        # skip log
        if log.lv != SKIP_ERROR_LOG_LV:
            return

        if self.current_direction == UPLOADING:
            # upload時はコピー元
            log_arg_prefix = self.info.share_name + "/"
        else:
            log_arg_prefix = self.info.mailaddress + ":"

        if log.log == FILE_SIZE_OVER_LOG:
            self._log_event(ERR_LOG_MAP[FILE_TOO_LARGE_ERR_LOG],
                {"name": self.info.share_name + "/" + log.relpath})

        elif log.log == EQUIVALENT_SKIP_LOG:
            self._log_event(ERR_LOG_MAP[SAME_FILE_NAME_EXISTS_ERR_LOG],
                {"name": self.info.share_name + "/" + log.relpath})

        elif isinstance(log.exc_arg, ForbiddenNameError):
            self._log_event(ERR_LOG_MAP[FORBIDDEN_FILE_NAME_ERR_LOG],
                {"name": log_arg_prefix + log.relpath})

        elif isinstance(log.exc_arg, NameTooLongError):
            self._log_event(ERR_LOG_MAP[TOO_LONG_FILE_NAME_ERR_LOG],
                {"name": log_arg_prefix + log.relpath})

        elif isinstance(log.exc_arg, PermissionError):
            self._log_event(ERR_LOG_MAP[NO_PERMISSION_ERR_LOG],
                {"name": self.info.mailaddress + ":" + log.relpath})

        elif isinstance(log.exc_arg, OperationLocked):
            self._log_event(ERR_LOG_MAP[LOCKED_ERR_LOG],
                {"name": self.info.mailaddress + ":" + log.relpath})
        else:
            self._log_event(ERR_LOG_MAP[SKIP_ERR_LOG],
                {"name": log_arg_prefix + log.relpath})

    def write_log(self, log):
        """Arg log is Log class instance."""

        self._save_corresponding_event_log(log)
        self._logger.write_log(log)

    def _save_result(self, result):

        result_path = os.path.join(self.info.info_dir,
                                   self.info.share_name,
                                   RESULT_DIR_NAME)

        dir_name = os.path.dirname(result_path)
        if not os.path.isdir(dir_name):
            os.makedirs(dir_name)

        with open(result_path, "w") as fh:
            fh.write(str(result))

    def _server_session_open(self):

        self._server_tree.session_open()

    def _server_session_close(self):

        try:
            self._server_tree.session_close()
        except:
            pass

    def _set_force_download(self, enable=True):

        try:
            with open(DOWNLOAD_TARGET_CONF_PATH) as conf:
                if self.info.share_name == conf.read().strip():
                    self.is_force_download = True
                    os.unlink(os.path.join(DOWNLOAD_TARGET_CONF_PATH))
        except:
            pass

    def _is_enable_download(self):

        return self.is_force_download or self.info.download_interval

    def __end_handling(self, result):

        try:
            self._server_session_close()
        except:
            pass

        try:
            self._save_result(result)
            self.__end_log(result)
            self.current_direction = None
        except:
            pass

        if self.is_need_retry_sleep:
            time.sleep(RETRY_PERIOD)
            self.is_need_retry_sleep = False

    def _try_exact_first_upload(self):

        self._logger.write(LOG_EXACT_FIRST_UPLOAD)

        try:
            self._try_dst_tree_access_transfer(self._local_tree,
                                               self._server_tree,
                                               do_check_update_fix=True)
        except:
            raise
        else:
            self._post_exact_first_upload()

    def _try_exact_first_download(self):

        self._logger.write(LOG_EXACT_FIRST_DOWNLOAD)

        is_executed = False

        before_uid = os.geteuid()
        before_gid = os.getegid()
        os.setegid(self.info.gid)
        os.seteuid(self.info.uid)

        try:
            last_download_time = self._get_last_time(DOWNLOAD_TIME_INFO)
            if not self._is_exec_download(last_download_time):
                debug(["not exec download"])
                return

            is_executed = True

            self._try_dst_tree_access_transfer(self._server_tree,
                self._local_tree, do_check_update_fix=False)

        except:
            os.seteuid(before_uid)
            os.setegid(before_gid)
            raise
        else:
            os.seteuid(before_uid)
            os.setegid(before_gid)
            if is_executed:
                self._post_exact_first_download()

    def _try_first_download(self):

        self._logger.write(LOG_FIRST_DOWNLOAD)
        super(BoxSync, self)._try_first_download()

    def _try_normal_upload(self):

        try:
            if self.is_first_upload:
                self._logger.write(LOG_FIRST_UPLOAD)
            else:
                self._logger.write(LOG_NORMAL_UPLOAD)
        except:
            pass

        super(BoxSync, self)._try_normal_upload()

    def _try_normal_download(self):

        self._logger.write(LOG_NORMAL_DOWNLOAD)
        super(BoxSync, self)._try_normal_download()

    def try_sync(self):

        result = SUCCESS
        try:
            self.is_need_retry_sleep = False
            self.current_direction = UPLOADING

            self.__init_sync_log()

            # upload
            self._logger.write(LOG_START_UPLOAD)
            self._try_upload()
            self._logger.write(LOG_END_UPLOAD)

            # download
            self._set_force_download()
            if self._is_enable_download():
                self.current_direction = DOWNLOADING
                self._logger.write(LOG_START_DOWNLOAD)
                self._try_download()
                self._logger.write(LOG_END_DOWNLOAD)

        except:
            result = FAILED
            try:
                self._logger.write(LOG_ABORT_UPLOAD + \
                               lf2crlf(traceback.format_exc()),
                               is_need_save=True)
            except:
                pass
            raise
        finally:
            self.__end_handling(result)

    def try_download_immediate(self):

        result = SUCCESS
        try:
            self.is_need_retry_sleep = False
            self.current_direction = DOWNLOADING
            self.is_force_download = True
            self.__init_download_log()
            self._logger.write(LOG_START_DOWNLOAD)
            self._try_download()
            self._logger.write(LOG_END_DOWNLOAD)
        except:
            result = FAILED
            try:
                self._logger.write(LOG_ABORT_DOWNLOAD + \
                                   lf2crlf(traceback.format_exc()),
                                   is_need_save=True)
            except:
                pass
            raise
        finally:
            self.__end_handling(result)

    def cancel_sync(self):

        try:
            self._logger.write(LOG_CANCEL_SYNC, is_need_save=True)
            self.__end_log(STOPPED)
        except:
            pass
        finally:
            self._server_session_close()


class BoxSyncMgr(SyncMgr):
    """Sync Manager for the BOX cloud service."""

    INFO_DIR = BOX_SYNC_INFO_DIR
    ANNOUNCE_CAT = Cat.BOX

    class BoxSyncInfo(EconomySyncInfo):
        """Sync information calss for the BOX cloud service."""

        def __init__(self, share_name, share_abs_path,
                     uid, gid, download_interval,
                     mailaddress, passwd,
                     root, uploadmaxfilesize,
                     proxy_host, proxy_port,
                     info_dir):
            """Constructor."""

            super(self.__class__, self).__init__(share_name,
                                                 share_abs_path,
                                                 uid,
                                                 gid,
                                                 download_interval)
            self.mailaddress = mailaddress
            self.passwd = passwd
            self.root = root
            self.uploadmaxfilesize = uploadmaxfilesize * 1024 * 1024  # MB -> B
            self.proxy_host = proxy_host
            self.proxy_port = proxy_port
            self.info_dir = info_dir

    def _load_share_list(self):
        share_config = send_request(FAL_SHARES_PATH, ReqTypes.LOAD)
        try:
            return share_config.get_node_list(
                                        SHARE_CONF_SERVICE_ENABLE_SHARE_PATH)
        except NotFoundError:
            return []

    def _get_share_user_group_id(self, share):

        nobody_uid = pwd.getpwnam(NOBODY_NAME)[2]
        nogroup_gid = grp.getgrnam(NOGROUP_NAME)[2]

        uid = share.get_value(UID_NAME, nobody_uid)
        gid = share.get_value(GID_NAME, nogroup_gid)

        return (int(uid), int(gid))

    def _get_sync_info(self, share):

        box_node = share.get_node(SHARE_BOX_XPATH)

        uid, gid = self._get_share_user_group_id(share)

        proxy_host, proxy_port = self._get_proxy(REST_BASE_URL)

        return self.BoxSyncInfo(
            share.get_node(NAME_NAME).get_text(),
            share.get_node(PATH_NAME).get_text(),
            uid,
            gid,
            int(box_node.get_node(DOWNLOADINTERVAL_NAME).get_text()),
            box_node.get_node(MAILADDRESS_NAME).get_text(do_strip=False),
            box_node.get_node(PASSWD_NAME).get_text(do_strip=False),
            ROOT_PATH,
            int(box_node.get_node(
                    UPLOADMAXFILESIZE_NAME).get_text(do_strip=False)),
            proxy_host,
            proxy_port,
            info_dir=self.INFO_DIR)

    def _make_sync_dict(self, share):

        info = self._get_sync_info(share)
        return {info.share_name: BoxSync(info)}

    def _save_result(self, result, share_name):

        result_path = os.path.join(self.INFO_DIR, share_name, RESULT_DIR_NAME)
        dir_name = os.path.dirname(result_path)
        if not os.path.isdir(dir_name):
            os.makedirs(dir_name)
        with open(result_path, "w") as fh:
            fh.write(str(result))

    def _save_fail_result(self, fail_share_list):

        for share in fail_share_list:
            try:
                self._save_result(FAILED, share)
            except:
                pass

    def _get_sync_dict(self):

        fail_list = []
        sync_dict = {}
        enable_sync_dict = {}

        for share in self._load_share_list():
            try:
                enable_sync_dict[share.get_node(NAME_NAME).get_text()] = None
                sync_dict.update(self._make_sync_dict(share))
            except:
                debug(do_tb=True)
                try:
                    fail_list.append(share.get_node(NAME_NAME).get_text())
                except:
                    pass
                continue

        # Note: "sync_dict" don't include a "info" for share which is in
        #       a volume not mounted.
        self._delete_disable_info_dir(enable_sync_dict)
        self._save_fail_result(fail_list)
        return sync_dict

    def _get_proxy(self, url):
        """Return the tuple of a proxy host and port for "url"."""

        proxy_info = get_proxy(url)

        if proxy_info:
            return tuple(proxy_info.split(":"))
        else:
            return (None, None)

    def _reset_proxy_info(self):
        for share in self.sync_dict:
            sync = self.sync_dict[share]
            host, port = self._get_proxy(REST_BASE_URL)
            sync.reset_proxy_info(host, port)

    def _clear_announce(self):
        clear_announce("all", self.ANNOUNCE_CAT)

    def download_immediate(self, tgt_share_name_list):

        try:
            debug([self.__class__.__name__, datetime.now()],
                  False, fmt="/// %s CloudSyncStart /// %s\n")

            is_cur_err = False

            for share_name, sync in self.sync_dict.iteritems():

                if share_name not in tgt_share_name_list:
                    continue

                try:
                    sync.try_download_immediate()

                except Exception:
                    debug(do_tb=True)

                    if not sync.is_disable:
                        is_cur_err = True

                    self._reset_proxy_info()

                finally:
                    debug([share_name, datetime.now()],
                          False, fmt="+++ %s ShareSyncEnd +++ %s\n")

            if self.is_last_error and not is_cur_err:
                self._clear_announce()

            self.is_last_error = is_cur_err

        finally:
            debug([self.__class__.__name__, datetime.now()],
                  False, fmt="/// %s CloudSyncEnd /// %s\n")

    def sigterm_handler(self, signum, frame):

        for share, sync in self.sync_dict.iteritems():
            try:
                sync.cancel_sync()
            except:
                pass

        os._exit(0)
