#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""
Copyright (C) I-O DATA DEVICE, INC.  All rights reserved.
"""
from __future__ import with_statement

import os
import pwd
import grp
import errno
import fcntl
import logging
import socket

# from dropbox import rest, session
from dropbox2.client import DropboxClient
from dropbox2.client import (UnknownDropboxError,
                             InvalidAccessTokenDropboxError,
                             TooManyRequestsDropboxError,
                             ServerDropboxError,
                             EndpointSpecificDropboxError,
                             NotFoundDropboxError,
                             ConflictDropboxError,
                             MalformedPathDropboxError,
                             InvalidOauth1TokenInfoDropboxError,
                             AppIdMismatchDropboxError,
                             InsufficientSpaceDropboxError)

from product_conf import LOG_DIR
from ionas.nascmd import NasCmd
# from ionas.filesync import (LocalFileTree, DropboxFileTree)
from ionas.filesync import LocalFileTree
from cloudsync.dropboxfiletree import DropboxFileTree

from fal_info import (Results)

from fal_msginfo import Msg, Cat


from nasdsync_util import (MOUNT_ROOT,
                           EXTERNAL_SHARE_DICT,
                           get_proxy,
                           _init_check,
                           _log_event,
                           get_tmp_path,
                           _clear_announce)

# from ionas.filesync.sync import (SimpleSyncInfo, SimpleSync, SyncMgr, debug)
from cloudsync._basedropboxsync import (SimpleSyncInfo, SimpleSync, SyncMgr, debug)

from cloudsync.util import (lf2crlf,)


SHARE_CONFIG_FILE = "share_config"
DEFAULT_SHARE_ROOT = MOUNT_ROOT + "/sataraid1/share"
CONF_DIR = "/mnt/data/conf/dropbox"
LOCK_DIR = "/var/lock/landisk"
LOCK_SUFFIX = ".lock"

APP_KEY = "av49kj4lr28xk65"
APP_SECRET = "irh68ok5h6hqbar"
ACCESS_TYPE = 'dropbox'

# TARGET_URL = "https://" + session.DropboxSession.API_HOST + "/"
# PROXY_ENV_NAME = "DROPBOX_PROXY"


# for logging ################################################################
from ionas.fileaccess import LocalAccess
from cloudsync.logger import Logger
import traceback as tb
import datetime as dt

ROOT_UID = 0


class Proxy:
    """Class for handling proxy."""

    def __init__(self, get_proxy_func):
        self.get_proxy_func = get_proxy_func
        self.proxy_map = {}

    def __call__(self, url):
        """Try to get proxy and return it."""

        if url not in self.proxy_map:
            saved_euid = os.geteuid()
            try:
                os.seteuid(ROOT_UID)
                self.proxy_map[url] = self.get_proxy_func(url)
            except:
                pass
            finally:
                os.seteuid(saved_euid)

        return self.proxy_map[url]

    def clear(self):
        """Claer proxy_map."""

        self.proxy_map = {}


class DropboxSync(SimpleSync):

    SUCCESS = "success"
    FAILED = "failed"
    #STOPPED = "stopped"

    def __init__(self, *args, **kwargs):
        super(DropboxSync, self).__init__(*args, **kwargs)
        self._logger = Logger(self.info.share_name,
                              self.__logfilename(self.info.share_name),
                              LocalAccess(),
                              DropboxFileTree)

    def try_sync(self, *args, **kwargs):
        result = self.SUCCESS
        try:
            self.__init_sync_log()
            super(DropboxSync, self).try_sync(*args, **kwargs)
        except:
            try:
                LOG_ABORT_UPLOAD = "///// Abort Upload ///// \r\n"
                self._logger.write(LOG_ABORT_UPLOAD + lf2crlf(tb.format_exc()),
                                   is_need_save=True)
            except:
                pass
            result = self.FAILED
            raise
        finally:
            #self.__end_handling()
            self.__end_log(result)

    def log_callback(self, log):
        super(DropboxSync, self).log_callback(log)
        self._logger.write_log(log)
        self._logger._flush_and_renew_log_buf()

        # NOTE: これがないとログのファイル転送行より後がフラッシュされない
        self._logger.write('', is_need_save=True)

    def _try_upload(self, *args, **kwargs):
        LOG_START_UPLOAD = \
            "\r\n///// Start uploading files from local ==> cloud /////\r\n"
        LOG_END_UPLOAD = \
            "///// End uploading /////\r\n"

        self._logger.write(LOG_START_UPLOAD)
        super(DropboxSync, self)._try_upload(*args, **kwargs)
        self._logger.write(LOG_END_UPLOAD)

    def _try_download(self, *args, **kwargs):
        self._logger.write("\r\n")
        self._logger.write("///// Start downloading files from cloud ==> local /////\r\n")
        super(DropboxSync, self)._try_download(*args, **kwargs)
        self._logger.write("///// End downloading /////\r\n")

    def __init_sync_log(self):
        now = dt.datetime.now()
        #log_file_path = os.path.join(SYNC_LOG_DIR,
        #                             self.info.share_name,
        #                             self.__get_sync_start_daytime_str(now))
        #self._logger.init_log_param(log_file_path)
        self._logger.init_log_param(self.__logfilename(self.info.share_name))
        self._logger.set_sync_start_time(now)
        self.__write_log_header(now)

    #def __init_download_log(self):
    #    now = dt.datetime.now()
    #    #log_file_path = os.path.join(SYNC_LOG_DIR,
    #    #                             self.info.share_name,
    #    #                             self.__get_sync_start_daytime_str(now))
    #    #self._logger.init_log_param(log_file_path)
    #    self._logger.init_log_param(self.__logfilename(self.info.share_name))
    #    self._logger.set_sync_start_time(now)
    #    self.__write_log_header(now)

    def __write_log_header(self, daytime):
        LOG_TITLE =             ("****************************************\r\n"
                                 "   Sync with Dropbox Log                \r\n"
                                 "****************************************\r\n")
        LOG_LOCAL_SHARE_TMPL =   "Local Share    : '%s'\r\n"
        #LOG_SERVER_FOLDER_TMPL = "Server Folder  : '%s'\r\n"
        LOG_START_DAYTIME_TMPL = "Start Day Time : %s\r\n"

        self._logger.write(LOG_TITLE)
        self._logger.write(LOG_LOCAL_SHARE_TMPL % self.info.share_name)
        #self._logger.write(LOG_SERVER_FOLDER_TMPL % self.info.container)
        self._logger.write(LOG_START_DAYTIME_TMPL % daytime.strftime("%Y/%m/%d %H:%M:%S"))

    def __end_log(self, result):
        try:
            self._logger.set_sync_end_time(dt.datetime.today())
            self._logger.end_log(result)
        except:
            pass

    def __logfilename(self, sharename):
        LAN_DISK_Log = 'LAN DISK Log'
        try:
            LAN_DISK_Log = Request(FAL_SHARES_PATH).request(Entry(ReqTypes.LOAD, """<shares/>""")).get_entry().get().get_value("sharelist/share[name='%s']/path" % LAN_DISK_Log)
        except:
            LAN_DISK_Log = "/mnt/sataraid1/share/%s" % LAN_DISK_Log
        now = dt.datetime.now()
        return os.path.join(LAN_DISK_Log,
                            'Dropbox',
                            sharename,
                            "%04d%02d%02d_%02d%02d%02d" % (now.year,
                                                               now.month,
                                                               now.day,
                                                               now.hour,
                                                               now.minute,
                                                               now.second))

##############################################################################


def _set_trees(self):
    info = self.info
    self.local_tree = LocalFileTree(info.share_abs_path,
                                    tmp_path=get_tmp_path(info.share_abs_path),
                                    log_callback=self.log_callback,
                                    ctime=False)

    # self.server_tree = DropboxFileTree(info.sess,
    self.server_tree = DropboxFileTree(info.dropbox_client,
                                       log_callback=self.log_callback)


def _is_timeout_err(self, err):
    try:
        # if isinstance(err, rest.RESTSocketError):
        if isinstance(err, socket.error):
            # """
            # 'Error connecting to "api-content.dropbox.com"'の
            # "api-content.dropbox.com"が
            # "api.dropbox.com"だったりするので
            # 末尾文字列で判定
            # """

            # timeout_args_list = (
            #     'timed out',
            #     "(-5, 'No address associated with hostname')",
            #     "(101, 'Network is unreachable')",
            #     "(113, 'No route to host')",
            #     'Tunnel connection failed: 504 Gateway Time-out')

            # for arg in timeout_args_list:
            #     if err.args[0].endswith(arg):
            #         return True

            if err.args[0] in [
                    'timed out',
                    -5,   # No address associated with hostname
                    101,  # Network is unreachable
                    113,  # No route to host
                    504,  # 'Tunnel connection failed: 504 Gateway Time-out',
                    'Tunnel connection failed: 504 Gateway Time-out',
                ]:
                return True

        elif isinstance(err, socket.sslerror):
            timeout_args_list = ('The read operation timed out',)

            if err.args[0] in timeout_args_list:
                return True

    except:
        return False
    return False


class DropboxShareConfig:
    """Class for the share configuration file."""

    def __init__(self, conf_file):
        """Constructor."""

        self.conf_path = os.path.join(CONF_DIR, conf_file)
        self.lock_path = os.path.join(LOCK_DIR, conf_file) + LOCK_SUFFIX
        self.lock_file = None
        self.config = {}

        self.LINE_SPLIT = "|"
        self.KEY_IDX = 0
        self.SECRET_IDX = 1

    def lock(self):
        """Lock a lock file."""

        if not self.lock_file:
            try:
                self.lock_file = open(self.lock_path, "w")
            except (IOError, OSError), err:
                if err.errno == errno.ENOENT:
                    os.makedirs(LOCK_DIR)
                    self.lock_file = open(self.lock_path, "w")
                else:
                    raise

            fcntl.flock(self.lock_file.fileno(), fcntl.LOCK_EX)

    def unlock(self):
        """Unlock a lock file."""

        if self.lock_file:
            self.lock_file.close()
            self.lock_file = None

    def load(self):
        """Load configuration from a file."""

        self.config = {}
        try:
            conf_file = open(self.conf_path, "r")
            for line in conf_file:
                try:
                    line = line.strip()
                    (share, key, secret) = line.split(self.LINE_SPLIT)

                    self.config[share] = [key, secret]
                except:
                    pass

            conf_file.close()
        except:
            pass

    def save(self):
        """Save configuration to a file."""

        conf_file = open(self.conf_path, "w")
        try:
            for share in self.config:
                save_list = [share]
                save_list += self.config[share]

                conf_file.write(self.LINE_SPLIT.join(save_list) + "\n")

        except:
            conf_file.close()

    def get_target_list(self):
        """Return target share folder list. """

        if not len(self.config):
            self.load()

        return self.config.keys()

    def get_access_token(self, share):
        """Return the tuple key and secret pair for share."""

        if not len(self.config):
            self.load()

        return (self.config[share][self.KEY_IDX],
                    self.config[share][self.SECRET_IDX])

    def set_access_token(self, share, key, secret):
        """Set the key and secret for share."""

        if not len(self.config):
            self.load()

        self.config[share] = [key, secret]

    def clear_access_token(self, share):
        """Clear the key and secret for share."""

        if not len(self.config):
            self.load()

        del self.config[share]


class DropboxSyncMgr(SyncMgr):

    # SERVICE_ERR_LIST = [rest.RESTSocketError, rest.ErrorResponse]
    SERVICE_ERR_LIST = [UnknownDropboxError,
                        InvalidAccessTokenDropboxError,
                        TooManyRequestsDropboxError,
                        ServerDropboxError,
                        EndpointSpecificDropboxError,
                        NotFoundDropboxError,
                        ConflictDropboxError,
                        MalformedPathDropboxError,
                        InvalidOauth1TokenInfoDropboxError,
                        AppIdMismatchDropboxError,
                        InsufficientSpaceDropboxError]

    SHARE_CONF_SERVICE_ENABLE_PATH = \
        "sharelist/share/shareservice/dropbox[@enable='true']/../.."

    ANNOUNCE_CAT = Cat.DROPBOX

    NASCMD_FOR_LOGGING = NasCmd(Results, logger_name="DROPBOX",
        log=LOG_DIR + "/dropbox.log", log_level=logging.DEBUG)

    #SYNC_CLASS = SimpleSync
    SYNC_CLASS = DropboxSync

    class DropboxInfo(SimpleSyncInfo):
        # def __init__(self, share_name, share_abs_path, uid, gid, sess):
        def __init__(self, share_name, share_abs_path, uid, gid, dropbox_client):
            super(self.__class__, self).__init__(share_name,
                                                 share_abs_path,
                                                 uid,
                                                 gid)
            # self.sess = sess
            self.dropbox_client = dropbox_client

    def __init__(self):
        self.proxy_manager = Proxy(get_proxy)
        self._reset_proxy_info()
        super(self.__class__, self).__init__()
      
    # def _get_info(self, share_name, share_abs_path, sess, dropbox_client=None):
    def _get_info(self, share_name, share_abs_path, dropbox_client):
        nobody_uid = pwd.getpwnam("nobody")[2]
        nogroup_gid = grp.getgrnam("nogroup")[2]

        return self.DropboxInfo(share_name,
                                share_abs_path,
                                nobody_uid,
                                nogroup_gid,
                                # sess,
                                # dropbox_client=dropbox_client)
                                dropbox_client)

    def __load_share_config(self):
        try:
            from ionas.request import Request, Entry
            from fal_info import Results, ReqTypes
            from fal_info import FAL_SHARES_PATH
            response = Request(FAL_SHARES_PATH).request(Entry(ReqTypes.LOAD, """<shares/>"""))
            entry = response.get_entry()
            targets = {}
            for share_node in entry.get().get_node_list('sharelist/share[shareservice/dropbox/@enable="true"]'):
                try:
                    name = share_node.get_node('name').get_text()
                    path = share_node.get_node('path').get_text()
                    download_enabled = not share_node.get_node('shareservice/dropbox').is_set_to_true('uploadonly')
                    targets[name] = {
                        'path': path,
                        'download_enabled': download_enabled,
                    }
                except:
                    pass
            return targets
        except:
            return {}

    def __load_access_token(self, sharename):
        try:
            from subprocess import (Popen, PIPE)
            command = ['/usr/local/bin/dropbox_access_token_manager_for_nasdsync', '--load-access-token', sharename]
            process = Popen(command, stdout=PIPE)
            output, error = process.communicate()
            access_token = output.strip()
            return access_token
        except:
            return None

    def _get_sync_dict(self):
        debug()
        share_config = DropboxShareConfig(SHARE_CONFIG_FILE)
        share_config.lock()
        share_config.load()
        share_config.unlock()

        sync_dict = {}
        for k, v in self.__load_share_config().iteritems():
            share_name = k
            share_path = v['path']
            download_enabled = v['download_enabled']
            client_id, client_secret = APP_KEY, APP_SECRET
            dropbox_client = DropboxClient(client_id, client_secret)
            access_token = self.__load_access_token(share_name)
            if access_token:
                dropbox_client.access_token = access_token
            else:
                continue
            dropbox_client.get_proxy = self.proxy_manager

            try:
                # key, secret = share_config.get_access_token(share_name)
                # sess = session.DropboxSession(APP_KEY,
                #                               APP_SECRET,
                #                               ACCESS_TYPE)
                # sess.set_token(key, secret)
                info = self._get_info(share_name,
                                      share_path,
                                      # sess,
                                      dropbox_client)
                sync_dict[share_name] = self.SYNC_CLASS(
                   info,
                   service_err_list=self.SERVICE_ERR_LIST,
                   no_root_entry_err_log="",
                   no_space_on_local_err_log=Msg.ERR_DROPBOX_NO_SPACE_ON_LOCAL,
                   no_space_on_server_err_log=Msg.ERR_DROPBOX_NO_SPACE_ON_SERVER,
                   invalid_identifier_err_log="",
                   invalid_password_err_log="",
                   invalid_fs_err_log=Msg.ERR_DROPBOX_INVALID_FORMAT,
                   file_too_large_err_log=Msg.ERR_DROPBOX_FILE_TOO_LARGE,
                   skip_err_log=Msg.ERR_DROPBOX_SYNC_SKIP,
                   timeout_err_log=Msg.ERR_DROPBOX_TIMEOUT,
                   misc_err_log=Msg.ERR_DROPBOX_SYNC,
                   nascmd_for_logging=self.NASCMD_FOR_LOGGING,
                   _set_trees=_set_trees,
                   _init_check=_init_check,
                   _is_timeout_err=_is_timeout_err,
                   _log_event=_log_event,
                   _clear_announce=_clear_announce)
                sync_dict[share_name].download_enabled = download_enabled
            except:
                pass
#        for share in share_config.get_target_list():
#            share_base = os.path.basename(share)
#            if share_base in EXTERNAL_SHARE_DICT:
#                share_abs_path = EXTERNAL_SHARE_DICT[share_base]
#            else:
#                share_abs_path = os.path.join(DEFAULT_SHARE_ROOT, share)
#
#            try:
#                key, secret = share_config.get_access_token(share_base)
#                sess = session.DropboxSession(APP_KEY,
#                                              APP_SECRET,
#                                              ACCESS_TYPE)
#                sess.set_token(key, secret)
#                info = self._get_info(share_base, share_abs_path, sess)
#                sync_dict[share_base] = self.SYNC_CLASS(
#                   info,
#                   service_err_list=self.SERVICE_ERR_LIST,
#                   no_root_entry_err_log="",
#                   no_space_on_local_err_log=Msg.ERR_DROPBOX_NO_SPACE_ON_LOCAL,
#                   no_space_on_server_err_log=Msg.ERR_DROPBOX_NO_SPACE_ON_SERVER,
#                   invalid_identifier_err_log="",
#                   invalid_password_err_log="",
#                   invalid_fs_err_log=Msg.ERR_DROPBOX_INVALID_FORMAT,
#                   file_too_large_err_log=Msg.ERR_DROPBOX_FILE_TOO_LARGE,
#                   skip_err_log=Msg.ERR_DROPBOX_SYNC_SKIP,
#                   timeout_err_log=Msg.ERR_DROPBOX_TIMEOUT,
#                   misc_err_log=Msg.ERR_DROPBOX_SYNC,
#                   nascmd_for_logging=self.NASCMD_FOR_LOGGING,
#                   _set_trees=_set_trees,
#                   _init_check=_init_check,
#                   _is_timeout_err=_is_timeout_err,
#                   _log_event=_log_event,
#                   _clear_announce=_clear_announce)
#
#            except:
#                debug(do_tb=True)
#                pass

        return sync_dict

    def _reset_proxy_info(self):
        self.proxy_manager.clear()

    def _log_event(self, event, args={}, err_info=""):
        _log_event(self, event, args={}, err_info="")

    def _clear_announce(self):
        _clear_announce(self)
