#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""
Sync module for the OneDrive cloud service.

Copyright (C) I-O DATA DEVICE, INC.  All rights reserved.
"""
from __future__ import with_statement

import os
import time
import shutil
import socket
import errno
import pwd
import grp
import urllib2
import logging
import traceback
import requests
from sys import (exc_info,)
from datetime import (datetime,)

from onedrivesdk.error import (OneDriveError, ErrorCode)

from product_conf import (CONF_DIR, TEMP_DIR, LOG_DIR,)
from iobase.exception import (NotFoundError,)
from ionas.exception import (BaseAccessException,
                             NoEntryError, NoRootEntryError,
                             InvalidFileSystemError, NoSpaceError,
                             ExistsError, SizeTooLargeError,
                             InternalError, NameTooLongError,
                             PermissionError, TimedOutError,)
from ionas.fileaccess import (LocalAccess,)
from ionas.filesync import (SKIP_ERROR_LOG_LV, FILE_SIZE_OVER_LOG,
                            LocalFileTree,
                            get_sync_dup_tree_distrust_new, get_same_tree,
                            FILES_NAME, DIRS_NAME, MTIME_NAME, CHILDREN_NAME,
                            STATE_NAME, STATE_NEW, STATE_DEL, STATE_SAME,)
from ionas.filesync.sync import (EconomySyncInfo, EconomySyncEx, SyncMgr,
                                    debug, DOWNLOAD_TIME_INFO)

from fal_msginfo import (Msg, Cat,)
from fal_info import (ReqTypes, Results, FAL_SHARES_PATH,)
from fal_util import (save_event_log, clear_announce,)
from cloudsync.util import (lf2crlf, recreate_http_error,)
from cloudsync.logger import (Logger, InternalLog)
from cloudsync.onedrivelocalfiletree import (OneDriveLocalFileTree,
                                                EQUIVALENT_SKIP_LOG)
from cloudsync.onedrivefiletree import (OneDriveFileTree, ForbiddenNameError)
from nasdsync_util import (get_sync_log_dir_path, is_share_available,
                                get_tmp_path, get_proxy, send_request,)
from pkgaddon.onedrive_config import (ONEDRIVE_CONF_DIR,
                                        ONEDRIVE_BUSINESS_CONF_DIR,
                                        BUSINESS_AUTH_SERVER_URL,
                                        DOWNLOAD_TARGET_CONF_PATH,)

LOGGER_NAME = "OneDrive"
BUSINESS_LOGGER_NAME = "OneDrive for Business"
LOG_FILE_PATH = os.path.join(LOG_DIR, "onedrive.log")

LOG_DIR_NAME = "OneDrive"
BUSINESS_LOG_DIR_NAME = "OneDrive for Business"
SYNC_LOG_DIR = os.path.join(get_sync_log_dir_path(), LOG_DIR_NAME)
BUSINESS_SYNC_LOG_DIR = os.path.join(get_sync_log_dir_path(),
                                        BUSINESS_LOG_DIR_NAME)

UPLOADING = "uploading"
DOWNLOADING = "downloading"

RETRY_PERIOD = 600
RECOVER_PERIOD = 10

NO_ROOT_ENTRY_ERR_LOG = 1
NO_SPACE_ON_LOCAL_ERR_LOG = 2
NO_SPACE_ON_SERVER_ERR_LOG = 3
INVALID_IDENTIFIER_ERR_LOG = 4
INVALID_PASSWORD_ERR_LOG = 5
INVALID_FS_ERR_LOG = 6
FILE_TOO_LARGE_ERR_LOG = 7
SKIP_ERR_LOG = 8
TIMEOUT_ERR_LOG = 9
MISC_ERR_LOG = 10
AUTH_ERR_LOG = 11
FORBIDDEN_FILE_NAME_ERR_LOG = 12
NO_PERMISSION_ERR_LOG = 13
LOCKED_ERR_LOG = 14
CONNECT_ERR_LOG = 15
SAME_FILE_NAME_EXISTS_ERR_LOG = 16
TOO_LONG_FILE_NAME_ERR_LOG = 17

ERR_LOG_MAP = {
        NO_SPACE_ON_LOCAL_ERR_LOG: Msg.ERR_ONEDRIVE_NO_SPACE_ON_LOCAL,
        NO_SPACE_ON_SERVER_ERR_LOG: Msg.ERR_ONEDRIVE_NO_SPACE_ON_SERVER,
        INVALID_FS_ERR_LOG: Msg.ERR_ONEDRIVE_INVALID_FORMAT,
        FILE_TOO_LARGE_ERR_LOG: Msg.ERR_ONEDRIVE_TOO_LARGE_FILE,
        SKIP_ERR_LOG: Msg.ERR_ONEDRIVE_SYNC_SKIP,
        TIMEOUT_ERR_LOG: Msg.ERR_ONEDRIVE_TIMEOUT,
        MISC_ERR_LOG: Msg.ERR_ONEDRIVE_SYNC,
        AUTH_ERR_LOG: Msg.ERR_ONEDRIVE_AUTH,
        FORBIDDEN_FILE_NAME_ERR_LOG: Msg.ERR_ONEDRIVE_FORBIDDEN_FILE_NAME,
        NO_PERMISSION_ERR_LOG: Msg.ERR_ONEDRIVE_NO_PERMISSION,
        CONNECT_ERR_LOG: Msg.ERR_ONEDRIVE_CONNECT,
        SAME_FILE_NAME_EXISTS_ERR_LOG: Msg.ERR_ONEDRIVE_SAME_FILE_NAME_EXISTS,
        TOO_LONG_FILE_NAME_ERR_LOG: Msg.ERR_ONEDRIVE_TOO_LONG_FILE_NAME,
    }

BUSINESS_ERR_LOG_MAP = {
        NO_SPACE_ON_LOCAL_ERR_LOG: Msg.ERR_ONEDRIVEB_NO_SPACE_ON_LOCAL,
        NO_SPACE_ON_SERVER_ERR_LOG: Msg.ERR_ONEDRIVEB_NO_SPACE_ON_SERVER,
        INVALID_FS_ERR_LOG: Msg.ERR_ONEDRIVEB_INVALID_FORMAT,
        FILE_TOO_LARGE_ERR_LOG: Msg.ERR_ONEDRIVEB_TOO_LARGE_FILE,
        SKIP_ERR_LOG: Msg.ERR_ONEDRIVEB_SYNC_SKIP,
        TIMEOUT_ERR_LOG: Msg.ERR_ONEDRIVEB_TIMEOUT,
        MISC_ERR_LOG: Msg.ERR_ONEDRIVEB_SYNC,
        AUTH_ERR_LOG: Msg.ERR_ONEDRIVEB_AUTH,
        FORBIDDEN_FILE_NAME_ERR_LOG: Msg.ERR_ONEDRIVEB_FORBIDDEN_FILE_NAME,
        NO_PERMISSION_ERR_LOG: Msg.ERR_ONEDRIVEB_NO_PERMISSION,
        CONNECT_ERR_LOG: Msg.ERR_ONEDRIVEB_CONNECT,
        SAME_FILE_NAME_EXISTS_ERR_LOG: Msg.ERR_ONEDRIVEB_SAME_FILE_NAME_EXISTS,
        TOO_LONG_FILE_NAME_ERR_LOG: Msg.ERR_ONEDRIVEB_TOO_LONG_FILE_NAME,
    }


ERROR_INVALID_GRANT = "invalid_grant"

JSON_ERR_CODE_KEY = "code"

err_code_except_dict = {
    "item_name_in_use": ExistsError,
    "forbidden": PermissionError,
    "access_denied_insufficient_permissions": PermissionError,
    "not_found": NoEntryError,
#    "access_denied_item_locked": OperationLocked,
    "storage_limit_exceeded": NoSpaceError,
    "file_size_limit_exceeded": SizeTooLargeError,
#    "item_name_invalid": ForbiddenNameError,
    }

NOBODY_NAME = "nobody"
NOGROUP_NAME = "nogroup"

SYNC_DIR_FMT_STR = "%04d%02d%02d_%02d%02d%02d"

SHARE_CONF_SERVICE_ENABLE_SHARE_PATH = \
        "sharelist/share/shareservice/onedrive[@enable='true']/../.."
BUSINESS_SHARE_CONF_SERVICE_ENABLE_SHARE_PATH = \
        "sharelist/share/shareservice/onedrivebusiness[@enable='true']/../.."
SHARE_ONEDRIVE_XPATH = "shareservice/onedrive"
SHARE_ONEDRIVE_BUSINESS_XPATH = "shareservice/onedrivebusiness"

NAME_NAME = "name"
PATH_NAME = "path"
DOWNLOADINTERVAL_NAME = "downloadinterval"
UPLOADMAXFILESIZE_NAME = "uploadmaxfilesize"
MAILADDRESS_NAME = "mailaddress"
PASSWD_NAME = "passwd"

ONEDRIVE_INFO_DIR = os.path.join(ONEDRIVE_CONF_DIR, "info")
ONEDRIVE_BUSINESS_INFO_DIR = os.path.join(ONEDRIVE_BUSINESS_CONF_DIR,
                                                "info")
SESSION_NAME = "session"
SESSION_SUFFIX = ".session"


UID_NAME = "uid"
GID_NAME = "gid"

TARGET_ROOT_INFO = "target_root"
ROOT_PATH = "/"

SUCCESS = "success"
FAILED = "failed"
STOPPED = "stopped"

LOG_TITLE = \
"****************************************\r\n" + \
"   OneDrive Sync Log\r\n" + \
"****************************************\r\n"

LOG_LOCAL_SHARE_TMPL = "Local Share    : '%s'\r\n"
LOG_START_DAYTIME_TMPL = "Start Day Time : %s\r\n"

LOG_START_UPLOAD = \
    "\r\n///// Start uploading files from local ==> cloud /////\r\n"

LOG_END_UPLOAD = \
    "///// End uploading /////\r\n"

LOG_START_DOWNLOAD = \
    "\r\n///// Start downloading files from cloud ==> local /////\r\n"

LOG_END_DOWNLOAD = \
    "///// End downloading /////\r\n"

LOG_CANCEL_SYNC = "///// Cancel Sync /////\r\n"

LOG_EXACT_FIRST_UPLOAD = "///// Exact First Upload /////\r\n"
LOG_FIRST_UPLOAD = "///// First Upload /////\r\n"
LOG_NORMAL_UPLOAD = "///// Normal Upload /////\r\n"
LOG_ABORT_UPLOAD = "///// Abort Upload ///// \r\n"

LOG_EXACT_FIRST_DOWNLOAD = "///// Exact First Download ///// \r\n"
LOG_FIRST_DOWNLOAD = "///// First Download ///// \r\n"
LOG_NORMAL_DOWNLOAD = "///// Normal Download ///// \r\n"
LOG_ABORT_DOWNLOAD = "///// Abort Download ///// \r\n"

RESULT_DIR_NAME = "result"

TOO_LONG_MSG_PART = "exceeds maximum length"


class ServerCommunicationError(BaseAccessException):
    pass


class AuthError(BaseAccessException):
    pass


class OneDriveCallBack():
    """Call back class for a OneDriveSync class."""

    def __init__(self, onedrive_sync):
        """Constructor."""

        self._onedrive_sync = onedrive_sync

    def low_level_except_handler(self, err, target_path=None):
        """Low level exception handler."""

        if isinstance(err, requests.exceptions.Timeout):
            raise TimedOutError(detail=err)

        if isinstance(err, requests.exceptions.ConnectionError):
            raise ServerCommunicationError(detail=err)

        elif isinstance(err, OneDriveError):
            if err.code == ErrorCode.AccessDenied:
                raise PermissionError(detail=err)
            if err.code == ErrorCode.ActivityLimitReached:
                raise NoSpaceError(detail=err)
            if err.code == ErrorCode.QuotaLimitReached:
                raise NoSpaceError(detail=err)
            if err.code == ErrorCode.Unauthenticated:
                raise AuthError(detail=err)
            if err.code == ErrorCode.Malformed:
                raise ServerCommunicationError(detail=err)
            if err.code == ErrorCode.InvalidRequest:
                if (len(err.args) > 0 and
                    isinstance(err.args[0], (str, unicode)) and
                    err.args[0].lower().find(TOO_LONG_MSG_PART) >= 0):

                    raise NameTooLongError(detail=err)

                raise ForbiddenNameError(detail=err)

        elif isinstance(err, ValueError):
            raise ServerCommunicationError(detail=err)

        elif isinstance(err, LookupError):
            raise NoEntryError(detail=err)

        elif isinstance(err, Exception):
            if type(err) == type(Exception()):
                if err.args[0] == ERROR_INVALID_GRANT:
                    raise AuthError(detail=err)

        raise err

    def log_callback(self, log):
        """Call back method for log a "log" Log class instance."""

        self._onedrive_sync.write_log(log)


class OneDriveSync(EconomySyncEx):
    """Sync class for the OneDrive cloud service."""

    SERVICE_ERR_LIST = []

    def _get_sync_start_daytime_str(self, now):
        """Return sync start day and time string."""

        return SYNC_DIR_FMT_STR % (
            now.year, now.month, now.day, now.hour, now.minute, now.second)

    def _is_timeout_exception(self, exc):
        """Check if exception is for timeout."""

        try:
            if isinstance(exc, socket.gaierror):
                timeout_args_list = (
                    socket.EAI_NONAME,
                    socket.EAI_NODATA)

                if exc.args[0] in timeout_args_list:
                    return True

            elif isinstance(exc, socket.timeout):
                return True

            elif isinstance(exc, socket.sslerror):
                if exc.args[0] == "The read operation timed out":
                    return True

            elif isinstance(exc, socket.error):
                timeout_args_list = (
                    errno.ETIMEDOUT,
                    errno.EHOSTUNREACH)

                if exc.args[0] in timeout_args_list:
                    return True

                try:
                    if exc.args[1].endswith("(Gateway Time-out)"):
                        return True
                except:
                    pass

            elif isinstance(exc, TimedOutError):
                return True

            elif isinstance(exc, IOError):
                if isinstance(exc.args[1], socket.timeout):
                    return True

        except:
            return False
        return False

    def _is_misc_recover_err(self, exc):
        """Check if "exc" is recover error."""

        # Assume a recover errro.
        return True

    def _get_ev_log(self, exc):
        """Return evnet log messege code."""

        if self._is_timeout_exception(exc):
            debug(["!!! timeout err !!!"])
            return self.err_log_map[TIMEOUT_ERR_LOG]

        if isinstance(exc, ServerCommunicationError):
            return self.err_log_map[CONNECT_ERR_LOG]

        if self._is_misc_recover_err(exc):
            debug(["!!! misc recover err !!!"])
            return self.err_log_map[MISC_ERR_LOG]

        return None

    def _log_event(self, event, args={}, err_info=""):
        """Log "event" event."""

        ROOT_UID = 0
        ROOT_GID = 0

        if not event:
            return

        if err_info:
            self.internal_log.error(str(err_info))

        try:
            save_event_log(event, args)
        except:
            try:
                pre_uid = os.geteuid()
                pre_gid = os.getegid()
                os.setegid(ROOT_GID)
                os.seteuid(ROOT_UID)
                save_event_log(event, args)
            finally:
                os.setegid(pre_gid)
                os.seteuid(pre_uid)

    def _init_check(self):
        """Check if target share is available."""

        is_share_available(self.info.share_name, self.info.share_abs_path)

    def _set_trees(self):
        """Set local and server tree class instance."""

        self._local_tree = \
            OneDriveLocalFileTree(self.info.share_abs_path,
                               tmp_path=get_tmp_path(self.info.share_abs_path),
                               log_callback=self._callback.log_callback)
        self._server_tree = \
            OneDriveFileTree(session_path=self.info.session_path,
                           for_business=self.info.for_business,
                           root=self.info.root,
                           uploadmaxfilesize=self.info.uploadmaxfilesize,
                           proxy_host=self.info.proxy_host,
                           proxy_port=self.info.proxy_port,
                           misc_callback=self._callback)

        self._local_tree.set_cloudsync_onedriveaccess(
                            self._server_tree.onedriveaccess)

    def _recover_err_handle(self, exc, do_recover):
        share_name = self.info.share_name

        ev_log = self._get_ev_log(exc)

        if not ev_log:
            return False

        if do_recover:
            debug(["!!! do recover !!!"])
            time.sleep(RECOVER_PERIOD)
            return True

        self._log_event(ev_log, {"name": share_name})

        self.is_need_retry_sleep = True

        raise exc, None, exc_info()[2]

    def __no_retry_err_handle(self, exc):
        share_name = self.info.share_name

        if isinstance(exc, InvalidFileSystemError):
            self._log_event(self.err_log_map[INVALID_FS_ERR_LOG],
                            {"name": share_name})
        elif isinstance(exc, NoRootEntryError):
            if exc.detail != None:
                return

            # Don't retry.  Maybe volume is not mount.
        else:
            return

        self._set_disable_share(share_name)

        debug(["!!! no retry err !!!"])

        raise exc, None, exc_info()[2]

    def _retry_err_handle(self, exc):
        share_name = self.info.share_name
        if isinstance(exc, NoSpaceError) and \
                 self.current_direction == DOWNLOADING:
            self._log_event(self.err_log_map[NO_SPACE_ON_LOCAL_ERR_LOG],
                     {"name": share_name})

        elif isinstance(exc, NoSpaceError):
            self._log_event(self.err_log_map[NO_SPACE_ON_SERVER_ERR_LOG],
                     {"name": self.info.root})

        elif isinstance(exc, AuthError):
            self._log_event(self.err_log_map[AUTH_ERR_LOG],
                     {"name": share_name})

#        elif isinstance(exc, NoRootEntryError):
#            self._log_event(self.err_log_map[NO_ROOT_ENTRY_ERR_LOG],
#                    {"name": self.info.mailaddress + ":" + self.info.root})
        else:
            return

        debug(["!!! retry err !!!"])
        self.is_need_retry_sleep = True
        raise exc, None, exc_info()[2]

    def _misc_err_handle(self, exc):
        self._log_event(self.err_log_map[MISC_ERR_LOG],
                        {"name": self.info.share_name},
                        traceback.format_exc())
        raise exc, None, exc_info()[2]

    def _init_except_handler(self, exc):

        self.__no_retry_err_handle(exc)
        self._misc_err_handle(exc)

    def _access_except_handler(self, exc, is_do_recover=False):

        self.__no_retry_err_handle(exc)
        self._retry_err_handle(exc)
        if self._recover_err_handle(exc, is_do_recover):
            return True
        self._misc_err_handle(exc)
        return False

    def _delete_info_dir(self, share_name):
        pass

    def _clear_info_dir(self, share_name):
        try:
            shutil.rmtree(
                os.path.join(self.info.info_dir, share_name))
        except:
            pass

    def _get_download_target_fname(self):

        return os.path.join(TEMP_DIR, "onedrive_download_target")

    def _check_last_target(self):

        target_root = self._get_info(TARGET_ROOT_INFO)
        if not target_root:
            self._set_info(TARGET_ROOT_INFO, self.info.root)
        if target_root != self.info.root:
            self._clear_info_dir(self.info.share_name)
            self._set_info(TARGET_ROOT_INFO, self.info.root)

    def _write_log_header(self, daytime):

        self._logger.write(LOG_TITLE)
        self._logger.write(LOG_LOCAL_SHARE_TMPL % self.info.share_name)
        self._logger.write(LOG_START_DAYTIME_TMPL % \
                               daytime.strftime("%Y/%m/%d %H:%M:%S"))

    def _try_transfer(self, src, dst, last_time=None,
        do_check_update_fix=True, is_last_error=False):
        debug()

        is_updated, delta = src.get_delta_tree()

        if not is_updated and not is_last_error:
            return

        if do_check_update_fix:
            delta = self._check_update_fix(src)

        debug(["--- Before delta ---:", delta], False, fmt="%s%s\n")
        debug(["--- Before src.tree ---:", src.tree], False, fmt="%s%s\n")
        debug(["--- Before dst.tree ---:", dst.tree], False, fmt="%s%s\n")

        if last_time != None:
            delta = get_new_tree_last_time_later(delta, last_time)
            debug(["--- LastTimeLater delta ---:", delta], False, fmt="%s%s\n")

        dst.update(src, delta)

        debug(["--- After src.tree ---:", src.tree], False, fmt="%s%s\n")
        debug(["--- After dst.tree ---:", dst.tree], False, fmt="%s%s\n")

    def _try_dst_tree_access_transfer(
        self, src, dst, do_check_update_fix=True, last_time=None):
        debug()

        is_updated, delta = src.get_delta_tree()

        if not is_updated:
            return

        if do_check_update_fix:
            delta = self._check_update_fix(src)

        debug(["--- Before delta ---:", delta], False, fmt="%s%s\n")
        debug(["--- Before src.tree ---:", src.tree], False, fmt="%s%s\n")
        debug(["--- Before dst.tree ---:", dst.tree], False, fmt="%s%s\n")

        if last_time != None:
            delta = get_new_tree_last_time_later(delta, last_time)
            debug(["--- LastTimeLater delta ---:", delta], False, fmt="%s%s\n")

        src_delta, dst_delta = get_sync_dup_tree_distrust_new(
            delta, get_same_tree(dst.get_tree()))

        debug(["--- DistrustNew src_delta ---:", src_delta],
              False, fmt="%s%s\n")
        debug(["--- DistrustNew dst_delta ---:", dst_delta],
              False, fmt="%s%s\n")

        # COMMENT 削除反映前のサービス再起動への対処
        dst.update_delta_tree(
            lambda dst_tree: self._update_delta_tree(dst_tree, dst_delta))

        debug(["--- PreUpdate dst.tree ---:", dst.tree], False, fmt="%s%s\n")

        dst.update(src, src_delta)

        debug(["--- After src.tree ---:", src.tree], False, fmt="%s%s\n")
        debug(["--- After dst.tree ---:", dst.tree], False, fmt="%s%s\n")

    def _is_exec_download(self, last_download_time):
        if (self.is_force_download or
            (self.info.download_interval and
            self._compare_time(last_download_time,
                               self.info.download_interval))):
            return True
        return False

    def __init_sync_log(self):
        """Initialize sync log."""

        now = datetime.now()
        sync_log_dir = SYNC_LOG_DIR
        if self.info.for_business:
            sync_log_dir = BUSINESS_SYNC_LOG_DIR

        log_file_path = os.path.join(sync_log_dir,
                                     self.info.share_name,
                                     self._get_sync_start_daytime_str(now))
        self._logger.init_log_param(log_file_path)
        self._logger.set_sync_start_time(now)
        self._write_log_header(now)

    def __init_download_log(self):

        now = datetime.now()
        sync_log_dir = SYNC_LOG_DIR
        if self.info.for_business:
            sync_log_dir = BUSINESS_SYNC_LOG_DIR

        log_file_path = os.path.join(sync_log_dir,
                                     self.info.share_name,
                                     self._get_sync_start_daytime_str(now))
        self._logger.init_log_param(log_file_path)
        self._logger.set_sync_start_time(now)
        self._write_log_header(now)

    def __end_log(self, result):

        try:
            self._logger.set_sync_end_time(datetime.today())
            self._logger.end_log(result)
        except:
            pass

    def __init__(self, info):
        """Constructor."""

        self.info = info
        self.current_direction = None
        self.is_need_retry_sleep = False
        now = datetime.now()

        if self.info.for_business:
            self.err_log_map = BUSINESS_ERR_LOG_MAP
            sync_log_dir = BUSINESS_SYNC_LOG_DIR
            self.internal_log = InternalLog(BUSINESS_LOGGER_NAME,
                                        LOG_FILE_PATH, logging.DEBUG)
        else:
            self.err_log_map = ERR_LOG_MAP
            sync_log_dir = SYNC_LOG_DIR
            self.internal_log = InternalLog(LOGGER_NAME, LOG_FILE_PATH,
                                        logging.DEBUG)

        log_file_path = os.path.join(sync_log_dir,
                                     self.info.share_name,
                                     self._get_sync_start_daytime_str(now))
        self._logger = \
            Logger(self.info.share_name, log_file_path, LocalAccess(),
                    OneDriveFileTree)
        self._callback = OneDriveCallBack(self)

        super(OneDriveSync, self).__init__(info)

    def _save_corresponding_event_log(self, log):
        """Arg log is Log class instance."""

        if isinstance(log.exc_arg, NoSpaceError):
            if isinstance(log.exc_arg.detail, OneDriveError):
                self._log_event(self.err_log_map[NO_SPACE_ON_SERVER_ERR_LOG],
                        {"name": self.info.root})

        # skip log
        if log.lv != SKIP_ERROR_LOG_LV:
            return

        if self.current_direction == UPLOADING:
            # upload時はコピー元
            log_arg_prefix = self.info.share_name + "/"
        else:
            log_arg_prefix = ""

        if log.log == FILE_SIZE_OVER_LOG:
            self._log_event(self.err_log_map[FILE_TOO_LARGE_ERR_LOG],
                {"name": self.info.share_name + "/" + log.relpath})

        elif log.log == EQUIVALENT_SKIP_LOG:
            self._log_event(self.err_log_map[SAME_FILE_NAME_EXISTS_ERR_LOG],
                {"name": self.info.share_name + "/" + log.relpath})

        elif isinstance(log.exc_arg, ForbiddenNameError):
            self._log_event(self.err_log_map[FORBIDDEN_FILE_NAME_ERR_LOG],
                {"name": log_arg_prefix + log.relpath})

        elif isinstance(log.exc_arg, NameTooLongError):
            self._log_event(self.err_log_map[TOO_LONG_FILE_NAME_ERR_LOG],
                {"name": log_arg_prefix + log.relpath})

        elif isinstance(log.exc_arg, PermissionError):
            self._log_event(self.err_log_map[NO_PERMISSION_ERR_LOG],
                {"name": log.relpath})

        elif self._is_timeout_exception(log.exc_arg):
            self._log_event(self.err_log_map[TIMEOUT_ERR_LOG],
                {"name": log.relpath})

        else:
            self._log_event(self.err_log_map[SKIP_ERR_LOG],
                {"name": log_arg_prefix + log.relpath})

    def write_log(self, log):
        """Arg log is Log class instance."""

        self._save_corresponding_event_log(log)
        self._logger.write_log(log)

    def _save_result(self, result):

        result_path = os.path.join(self.info.info_dir,
                                   self.info.share_name,
                                   RESULT_DIR_NAME)

        dir_name = os.path.dirname(result_path)
        if not os.path.isdir(dir_name):
            os.makedirs(dir_name)

        with open(result_path, "w") as fh:
            fh.write(str(result))

    def _server_session_open(self):

        self._server_tree.session_open()

    def _server_session_close(self):

        try:
            self._server_tree.session_close()
        except:
            pass

    def _set_force_download(self, enable=True):

        try:
            with open(DOWNLOAD_TARGET_CONF_PATH) as conf:
                if self.info.share_name == conf.read().strip():
                    self.is_force_download = True
                    os.unlink(os.path.join(DOWNLOAD_TARGET_CONF_PATH))
        except:
            pass

    def _is_enable_download(self):

        return self.is_force_download or self.info.download_interval

    def __end_handling(self, result):

        try:
            self._server_session_close()
        except:
            pass

        try:
            self._save_result(result)
            self.__end_log(result)
            self.current_direction = None
        except:
            pass

        if self.is_need_retry_sleep:
            time.sleep(RETRY_PERIOD)
            self.is_need_retry_sleep = False

    def _try_exact_first_upload(self):

        self._logger.write(LOG_EXACT_FIRST_UPLOAD)

        try:
            self._try_dst_tree_access_transfer(self._local_tree,
                                               self._server_tree,
                                               do_check_update_fix=True)
        except:
            raise
        else:
            self._post_exact_first_upload()

    def _try_exact_first_download(self):

        self._logger.write(LOG_EXACT_FIRST_DOWNLOAD)

        is_executed = False

        before_uid = os.geteuid()
        before_gid = os.getegid()
        os.setegid(self.info.gid)
        os.seteuid(self.info.uid)

        try:
            last_download_time = self._get_last_time(DOWNLOAD_TIME_INFO)
            if not self._is_exec_download(last_download_time):
                debug(["not exec download"])
                return

            is_executed = True

            self._try_dst_tree_access_transfer(self._server_tree,
                self._local_tree, do_check_update_fix=False)

        except:
            os.seteuid(before_uid)
            os.setegid(before_gid)
            raise
        else:
            os.seteuid(before_uid)
            os.setegid(before_gid)
            if is_executed:
                self._post_exact_first_download()

    def _try_first_download(self):

        self._logger.write(LOG_FIRST_DOWNLOAD)
        super(OneDriveSync, self)._try_first_download()

    def _try_normal_upload(self):

        try:
            if self.is_first_upload:
                self._logger.write(LOG_FIRST_UPLOAD)
            else:
                self._logger.write(LOG_NORMAL_UPLOAD)
        except:
            pass

        super(OneDriveSync, self)._try_normal_upload()

    def _try_normal_download(self):

        self._logger.write(LOG_NORMAL_DOWNLOAD)
        super(OneDriveSync, self)._try_normal_download()

    def try_sync(self):

        result = SUCCESS
        try:
            self.is_need_retry_sleep = False
            self.current_direction = UPLOADING

            self.__init_sync_log()

            # upload
            self._logger.write(LOG_START_UPLOAD)
            self._try_upload()
            self._logger.write(LOG_END_UPLOAD)

            # download
            self._set_force_download()
            if self._is_enable_download():
                self.current_direction = DOWNLOADING
                self._logger.write(LOG_START_DOWNLOAD)
                self._try_download()
                self._logger.write(LOG_END_DOWNLOAD)

        except:
            result = FAILED
            try:
                self._logger.write(LOG_ABORT_UPLOAD + \
                               lf2crlf(traceback.format_exc()),
                               is_need_save=True)
            except:
                pass
            raise
        finally:
            self.__end_handling(result)

    def try_download_immediate(self):

        result = SUCCESS
        try:
            self.is_need_retry_sleep = False
            self.current_direction = DOWNLOADING
            self.is_force_download = True
            self.__init_download_log()
            self._logger.write(LOG_START_DOWNLOAD)
            self._try_download()
            self._logger.write(LOG_END_DOWNLOAD)
        except:
            result = FAILED
            try:
                self._logger.write(LOG_ABORT_DOWNLOAD + \
                                   lf2crlf(traceback.format_exc()),
                                   is_need_save=True)
            except:
                pass
            raise
        finally:
            self.__end_handling(result)

    def cancel_sync(self):

        try:
            self._logger.write(LOG_CANCEL_SYNC, is_need_save=True)
            self.__end_log(STOPPED)
        except:
            pass
        finally:
            self._server_session_close()


class OneDriveSyncMgr(SyncMgr):
    """Sync Manager for the OneDrive cloud service."""

    INFO_DIR = ONEDRIVE_INFO_DIR
    ANNOUNCE_CAT = Cat.ONEDRIVE

    class OneDriveSyncInfo(EconomySyncInfo):
        """Sync information calss for the OneDrive cloud service."""

        def __init__(self, share_name, share_abs_path,
                     uid, gid, download_interval,
                     session_path, for_business,
                     root, uploadmaxfilesize,
                     proxy_host, proxy_port,
                     info_dir):
            """Constructor."""

            super(self.__class__, self).__init__(share_name,
                                                 share_abs_path,
                                                 uid,
                                                 gid,
                                                 download_interval)
            self.session_path = session_path
            self.for_business = for_business
            self.root = root
            self.uploadmaxfilesize = uploadmaxfilesize * 1024 * 1024  # MB -> B
            self.proxy_host = proxy_host
            self.proxy_port = proxy_port
            self.info_dir = info_dir

    def _load_share_list(self, target_xpath):
        share_config = send_request(FAL_SHARES_PATH, ReqTypes.LOAD)
        try:
            return share_config.get_node_list(target_xpath)
        except NotFoundError:
            return []

    def _get_share_user_group_id(self, share):

        nobody_uid = pwd.getpwnam(NOBODY_NAME)[2]
        nogroup_gid = grp.getgrnam(NOGROUP_NAME)[2]

        uid = share.get_value(UID_NAME, nobody_uid)
        gid = share.get_value(GID_NAME, nogroup_gid)

        return (int(uid), int(gid))

    def _make_session_path(self, share_name):

        return os.path.join(ONEDRIVE_CONF_DIR, SESSION_NAME,
                            share_name + SESSION_SUFFIX)

    def _get_sync_info(self, share, for_business):

        onedrive_node = share.get_node(SHARE_ONEDRIVE_XPATH)
        info_dir = ONEDRIVE_INFO_DIR
        if for_business:
            onedrive_node = share.get_node(SHARE_ONEDRIVE_BUSINESS_XPATH)
            info_dir = ONEDRIVE_BUSINESS_INFO_DIR

        share_name = share.get_node(NAME_NAME).get_text()
        uid, gid = self._get_share_user_group_id(share)

        try:
            target_path = onedrive_node.get_node(PATH_NAME).get_text()
            if not target_path:
                target_path = ROOT_PATH
        except:
            target_path = ROOT_PATH

        proxy_host, proxy_port = self._get_proxy(BUSINESS_AUTH_SERVER_URL)

        return self.OneDriveSyncInfo(
            share_name,
            share.get_node(PATH_NAME).get_text(),
            uid,
            gid,
            int(onedrive_node.get_node(DOWNLOADINTERVAL_NAME).get_text()),
            self._make_session_path(share_name),
            for_business, target_path,
            int(onedrive_node.get_node(
                    UPLOADMAXFILESIZE_NAME).get_text(do_strip=False)),
            proxy_host,
            proxy_port,
            info_dir=info_dir)

    def _make_sync_dict(self, share, for_business):

        info = self._get_sync_info(share, for_business)
        return {info.share_name: OneDriveSync(info)}

    def _save_result(self, result, share_name, for_business):

        info_dir = ONEDRIVE_INFO_DIR
        if for_business:
            info_dir = ONEDRIVE_BUSINESS_INFO_DIR

        result_path = os.path.join(info_dir, share_name, RESULT_DIR_NAME)
        dir_name = os.path.dirname(result_path)
        if not os.path.isdir(dir_name):
            os.makedirs(dir_name)
        with open(result_path, "w") as fh:
            fh.write(str(result))

    def _save_fail_result(self, fail_share_list, for_business):

        for share in fail_share_list:
            try:
                self._save_result(FAILED, share, for_business)
            except:
                pass

    def _get_sync_dict_for_target_shares(self, target_xpath, for_business):

        fail_list = []
        sync_dict = {}
        enable_sync_dict = {}

        for share in self._load_share_list(target_xpath):
            try:
                enable_sync_dict[share.get_node(NAME_NAME).get_text()] = None
                sync_dict.update(self._make_sync_dict(share, for_business))
            except:
                debug(do_tb=True)
                try:
                    fail_list.append(share.get_node(NAME_NAME).get_text())
                except:
                    pass
                continue

        # Note: "sync_dict" don't include a "info" for share which is in
        #       a volume not mounted.
        self._delete_disable_info_dir(enable_sync_dict)
        self._save_fail_result(fail_list, for_business)
        return sync_dict

    def _get_sync_dict(self):

        return self._get_sync_dict_for_target_shares(
                        SHARE_CONF_SERVICE_ENABLE_SHARE_PATH, False)

    def _get_proxy(self, url):
        """Return the tuple of a proxy host and port for "url"."""

        proxy_info = get_proxy(url)

        if proxy_info:
            return tuple(proxy_info.split(":"))
        else:
            return (None, None)

    def _reset_proxy_info(self):
        for share in self.sync_dict:
            sync = self.sync_dict[share]
            host, port = self._get_proxy(BUSINESS_AUTH_SERVER_URL)
            sync.reset_proxy_info(host, port)

    def _clear_announce(self):
        clear_announce("all", self.ANNOUNCE_CAT)

    def download_immediate(self, tgt_share_name_list):

        try:
            debug([self.__class__.__name__, datetime.now()],
                  False, fmt="/// %s CloudSyncStart /// %s\n")

            is_cur_err = False

            for share_name, sync in self.sync_dict.iteritems():

                if share_name not in tgt_share_name_list:
                    continue

                try:
                    sync.try_download_immediate()

                except Exception:
                    debug(do_tb=True)

                    if not sync.is_disable:
                        is_cur_err = True

                    self._reset_proxy_info()

                finally:
                    debug([share_name, datetime.now()],
                          False, fmt="+++ %s ShareSyncEnd +++ %s\n")

            if self.is_last_error and not is_cur_err:
                self._clear_announce()

            self.is_last_error = is_cur_err

        finally:
            debug([self.__class__.__name__, datetime.now()],
                  False, fmt="/// %s CloudSyncEnd /// %s\n")

    def sigterm_handler(self, signum, frame):

        for share, sync in self.sync_dict.iteritems():
            try:
                sync.cancel_sync()
            except:
                pass

        os._exit(0)


class OneDriveBusinessSyncMgr(OneDriveSyncMgr):
    """Sync Manager for the OneDrive cloud service."""

    INFO_DIR = ONEDRIVE_BUSINESS_INFO_DIR
    ANNOUNCE_CAT = Cat.ONEDRIVEB

    def _make_session_path(self, share_name):

        return os.path.join(ONEDRIVE_BUSINESS_CONF_DIR, SESSION_NAME,
                            share_name + SESSION_SUFFIX)

    def _get_sync_dict(self):

        return self._get_sync_dict_for_target_shares(
                        BUSINESS_SHARE_CONF_SERVICE_ENABLE_SHARE_PATH, True)


def get_new_tree_last_time_later(tree, last_time, is_under_new_dir=False):

    for file in tree[FILES_NAME].keys():
        if not is_under_new_dir:
            if tree[FILES_NAME][file][STATE_NAME] != STATE_DEL:
                if tree[FILES_NAME][file][MTIME_NAME] > last_time:
                    tree[FILES_NAME][file][STATE_NAME] = STATE_NEW
                else:
                    tree[FILES_NAME][file][STATE_NAME] = STATE_SAME
        else:
            tree[FILES_NAME][file][STATE_NAME] = STATE_NEW

    for dir in tree[DIRS_NAME].keys():
        if not is_under_new_dir:
            if tree[DIRS_NAME][dir][STATE_NAME] != STATE_DEL:
                if tree[DIRS_NAME][dir][MTIME_NAME] > last_time:
                    tree[DIRS_NAME][dir][STATE_NAME] = STATE_NEW
                else:
                    tree[DIRS_NAME][dir][STATE_NAME] = STATE_SAME
        else:
            tree[DIRS_NAME][dir][STATE_NAME] = STATE_NEW
        get_new_tree_last_time_later(tree[DIRS_NAME][dir][CHILDREN_NAME],
                            last_time, is_under_new_dir)
    return tree


