#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""
Module for file tree on the BOX server.

Copyright (C) I-O DATA DEVICE, INC.  All rights reserved.
"""

import os
import sys
import errno
import socket
import time
import calendar
import re
import stat

import locale
import traceback
import email.utils

from ionas.exception import *
from ionas.fileaccess import (MODE_WRITE, DEFAULT_LOCALE)
from ionas.filesync.basefilesync import (IgnoreTargetFile)
from iobase.exception import (NotFoundError)
from cloudsync.boxaccess import (BoxAccess, PRIORITY_NONE, PRIORITY_FILE,
                                    PRIORITY_DIR)

from cloudsync.boxlocalfiletree import (LOCAL_NAME, MultipleEquivarentExist)
from cloudsync.boxinfo import (BoxInfo,)

from ionas.filesync.basefilesync import(
    DEBUG,
    SrcFull, DstFull,
    DIRS_NAME,
    FILES_NAME,
    MTIME_NAME,
    SIZE_NAME,
    CHILDREN_NAME,
    STATE_NAME,
    ERRINFO_NAME,
    STATE_NEW,
    STATE_DEL,
    STATE_SAME,
    ERR_STATE_PREFIX,
    NG_LOG,
    GET_TREE_LOG,
    GET_DELTA_TREE_LOG,
    ERROR_LOG_LV,
    get_new_tree,
    MODE_CH_FILE,
    MODE_CH_DIR)

DEFAULT_MAX_FILE_SIZE = 256 * 1024 * 1024

LINE_FMT = "%s,%s,%s,%s,%s\n"

NOTHING = None

FNAME_IDX = -5
ID_IDX = -4
MTIME_IDX = -3
CTIME_IDX = -2
SIZE_IDX = -1

INT_NONE = -1

GET_TREE_RETRY_NUM = 2


class OperationLocked(BaseAccessException):
    pass


class ForbiddenNameError(BaseAccessException):
    pass


def box_except_handler(err):
    raise err


##
# DstFull クラス内のメソッドをオーバライドし，フォルダ内のファイル数制限に
# 対応する
#
# フォルダ内のファイル数制限に引っかかった場合，そこで同期を中断せずに
# 次のフォルダの同期に移るように変更する．
#
class BoxDstFull(DstFull):

    def _update_exception_handler(self, exc, pathlist):
        pass

    def _DstFull__update(self, relpath, tree):
        """継承している DstFull クラスの __updateメソッドを
           オーバロードしている．
           フォルダあたりのファイル数が制限を越えた場合に転送を
           中断するのではなく次のフォルダの転送に移るように変更するため
        """

        skip_create = False

        local_dir_path = self.src.sep.join([self.src.root] + relpath)
        local_eq_map = self.src.get_equivarent_name_map(local_dir_path)

        # 削除すべきファイル
        for file in tree[FILES_NAME].keys():
            file_info = tree[FILES_NAME][file]

            real_file_name = file

            # 同等ファイルはスキップ
            if file_info[STATE_NAME] == STATE_DEL:
                try:
                    real_file_name = \
                        self.src.handle_equivarent_file(local_eq_map, relpath,
                                                        file, file_info)
                except MultipleEquivarentExist:
                    del tree[FILES_NAME][file]
                    continue

            try:
                if self._DstFull__try_operation(
                                    self.OPERATION[self.OPE_FILE_DELETE],
                                    relpath, real_file_name, file_info):
                    del tree[FILES_NAME][file]
            except:
                raise

        # 削除すべきディレクトリ
        for dir in tree[DIRS_NAME].keys():
            dir_info = tree[DIRS_NAME][dir]

            real_dir_name = dir

            # 同等ディレクトリはスキップ
            if dir_info[STATE_NAME] == STATE_DEL:
                try:
                    real_dir_name = \
                        self.src.handle_equivarent_dir(local_eq_map, relpath,
                                                        dir, dir_info)
                except MultipleEquivarentExist:
                    del tree[DIRS_NAME][dir]
                    continue

            try:
                if self._DstFull__try_operation(
                                        self.OPERATION[self.OPE_DIR_DELETE],
                                        relpath, real_dir_name, dir_info):
                    del tree[DIRS_NAME][dir]
            except:
                raise

        # 作成・更新すべきファイル
        for file in tree[FILES_NAME].keys():
            file_info = tree[FILES_NAME][file]

            real_file_name = file

            # 同等ファイルはスキップ
            if file_info[STATE_NAME] == STATE_NEW:
                try:
                    real_file_name = \
                         self.src.handle_equivarent_file(local_eq_map, relpath,
                                                            file, file_info)
                except MultipleEquivarentExist:
                    self.post_file_equivarent(relpath + [file], file_info)
                    continue

            try:
                self._DstFull__try_operation(
                                    self.OPERATION[self.OPE_FILE_TRANSFER],
                                    relpath, file, file_info)
            except NoSpaceError, e:
                self._update_exception_handler(e, relpath + [real_file_name])
                skip_create = True
                # 既存のファイルを更新する場合はファイル数は増えないので
                # できれば更新すべきだが，エラーが出まくることになる．
                # ここでは抜けておくことにする．
                break

        # 作成・更新すべきディレクトリ
        for dir in tree[DIRS_NAME].keys():
            dir_info = tree[DIRS_NAME][dir]

            real_dir_name = dir

            # 同等ディレクトリはスキップ
            if dir_info[STATE_NAME] == STATE_NEW:
                try:
                    real_dir_name = self.src.handle_equivarent_dir(
                                        local_eq_map, relpath, dir, dir_info)
                except MultipleEquivarentExist:
                    self.post_dir_equivarent(relpath + [dir], file_info)
                    continue

            try:
                if not skip_create:
                    self._DstFull__try_operation(
                                        self.OPERATION[self.OPE_DIR_CREATE],
                                        relpath, real_dir_name, dir_info)
            except NoSpaceError, e:
                self._update_exception_handler(e, relpath + [real_dir_name])
                skip_create = True
                continue

            if not (skip_create and dir_info[STATE_NAME] == STATE_NEW):
                # ディレクトリ階層を下る
                self._DstFull__update(relpath + [real_dir_name],
                                      tree[DIRS_NAME][dir][CHILDREN_NAME])

    def is_skip_err(self, exc_type, exc_ins):
        if (exc_type == NoEntryError or
            exc_type == PermissionError or
            exc_type == NameTooLongError or
            exc_type == ExistsError or
            exc_type == SizeTooLargeError or
            exc_type == ModifiedInTransferring or
            exc_type == FileLockError or
            exc_type == NonSupportedFileError or
            exc_type == IgnoreTargetFile or
            exc_type == OperationLocked or
            exc_type == ForbiddenNameError or
            self.is_misc_skip_err(exc_type, exc_ins)):
            return True

        return False

    def post_file_equivarent(self, file_relpath, file_info):
        pass

    def post_dir_equivarent(self, dir_relpath, dir_info):
        pass


class BoxFileTree(SrcFull, BoxDstFull):

    def __init__(self,
                 token_id,
                 mailaddress,
                 passwd,
                 root,
                 uploadmaxfilesize=DEFAULT_MAX_FILE_SIZE,
                 ignored_list=[],
                 encoding=None,
                 for_sync=True,
                 proxy_host=None,
                 proxy_port=None,
                 misc_callback=None):

        if encoding:
            self.encoding = encoding
        else:
            self.encoding = locale.getdefaultlocale()[1]
            if not self.encoding:
                self.encoding = DEFAULT_LOCALE

        self.misc_callback = misc_callback

        SrcFull.__init__(self, root, ignored_list,
                         log_callback=None,
                         misc_callback=misc_callback)

        BoxDstFull.__init__(self, root,
                               log_callback=None,
                               for_sync=for_sync,
                               misc_callback=misc_callback)

        # 最大ファイルサイズ + 1
        self.MAX_FILE_SIZE = uploadmaxfilesize

        self.token_id = token_id
        self.mailaddress = mailaddress
        self.passwd = passwd
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port

        self.low_level_except_handler = None
        if self.misc_callback:
            self.low_level_except_handler = \
                    self.misc_callback.low_level_except_handler

        self.box_access = BoxAccess(
             token_id=self.token_id,
             mailaddress=self.mailaddress,
             passwd=self.passwd,
             encoding=self.encoding,
             proxy_host=self.proxy_host,
             proxy_port=self.proxy_port,
             except_handler=self.low_level_except_handler)

        self.sep = self.box_access.sep
        self.same_name_priority = PRIORITY_DIR  # Force priority

        # To use convert_path() and compare_path() method
        self.boxinfo = BoxInfo("", "", "", "", "", None)
        self.conv_name_f = self.boxinfo.convert_path
        self.compare_path = self.boxinfo.compare_path

    def session_open(self, is_real_session_open=True):
        """Open session."""

        # Do nothing.
        pass

    def session_close(self):
        """Close session."""

        # Do nothing.
        pass

    def listdir(self, path):
        return self.box_access.listdir(path)

    def open(self, path, mode):
        return self.box_access.open(path, mode)

    def mkdir(self, path):
        self.box_access.mkdir(path)

    def stat(self, path):
        return self.box_access.stat(path)

    def remove(self, path):
        try:
            self.box_access.remove(path)
        except ForbiddenNameError, err:
            raise NameTooLongError(err)

    def rmdir(self, path):
        try:
            self.box_access.rmdir(path)
        except ForbiddenNameError, err:
            raise NameTooLongError(err)

    def __add_file(self, tree, name, size, mtime):
        if self.same_name_priority == PRIORITY_DIR:
            try:
                tree[DIRS_NAME][name]
            except KeyError:
                tree[FILES_NAME][name] = {
                    MTIME_NAME: mtime,
                    SIZE_NAME: size}
            return
        if self.same_name_priority == PRIORITY_FILE:
            try:
                del tree[DIRS_NAME][name]
            except:
                pass
            finally:
                tree[FILES_NAME][name] = {
                    MTIME_NAME: mtime,
                    SIZE_NAME: size}
            return
        if self.same_name_priority == PRIORITY_NONE:
            tree[FILES_NAME][name] = {
                    MTIME_NAME: mtime,
                    SIZE_NAME: size}
            return

    def __add_dir(self, tree, name, mtime):
        if name in tree[DIRS_NAME]:
            return

        if self.same_name_priority == PRIORITY_DIR:
            try:
                del tree[FILES_NAME][name]
            except:
                pass
            finally:
                tree[DIRS_NAME][name] = {
                    MTIME_NAME: mtime,
                    CHILDREN_NAME: {DIRS_NAME: {}, FILES_NAME: {}}}
            return
        if self.same_name_priority == PRIORITY_FILE:
            try:
                tree[FILES_NAME][name]
            except KeyError:
                tree[DIRS_NAME][name] = {
                    MTIME_NAME: mtime,
                    CHILDREN_NAME: {DIRS_NAME: {}, FILES_NAME: {}}}
            return
        if self.same_name_priority == PRIORITY_NONE:
            tree[DIRS_NAME][name] = {
                MTIME_NAME: mtime,
                CHILDREN_NAME: {DIRS_NAME: {}, FILES_NAME: {}}}

    def __get_tree(self, key, tree, prefix=""):

        #suffix = key.name[len(prefix):].lstrip(self.sep)
        tmp_path = key.name.lstrip(self.sep)
        suffix = tmp_path[len(prefix):].lstrip(self.sep)
        path_list = suffix.rstrip(self.sep).split(self.sep)

        mtime = key.last_modified

        if len(path_list) == 1:
            if not suffix.endswith(self.sep):
                """ This is File. """
                self.__add_file(tree, path_list[0], int(key.size), mtime)
            else:
                """ This is Directory. """
                self.__add_dir(tree, path_list[0], mtime)
        else:
            try:
                children = tree[DIRS_NAME][path_list[0]][CHILDREN_NAME]
            except KeyError:
                self.__add_dir(tree, path_list[0], mtime)
                try:
                    children = tree[DIRS_NAME][path_list[0]][CHILDREN_NAME]
                except KeyError:
                    return
            self.__get_tree(
                key,
                children,
                str(prefix + self.sep + path_list[0]).lstrip(self.sep))

    def __get_equivarent_file_name(self, tgt_name, file_name_list):

        for file_name in file_name_list:
            if self.compare_path(self.conv_name_f(tgt_name),
                                 self.conv_name_f(file_name)):
                return file_name
        return None

    def __get_equivarent_dir_name(self, tgt_name, dir_name_list):
        for dir_name in dir_name_list:
            if self.compare_path(self.conv_name_f(tgt_name),
                                 self.conv_name_f(dir_name)):
                return dir_name
        return None

    def __set_update(self, root_tree, update_tree):
        is_updated = False
        for update_name in update_tree[FILES_NAME]:
            # if update_name in root_tree[FILES_NAME].keys():
            real_file_name = self.__get_equivarent_file_name(
                                 update_name, root_tree[FILES_NAME].keys())
            if real_file_name:
                update_info = update_tree[FILES_NAME][update_name]
                root_info = root_tree[FILES_NAME][real_file_name]
                if root_info[STATE_NAME].startswith(ERR_STATE_PREFIX):
                    root_info[STATE_NAME] = re.sub(
                        "^" + ERR_STATE_PREFIX, "", root_info[STATE_NAME])
                    is_updated = True

                elif root_info[STATE_NAME] == STATE_DEL:
                    root_info[STATE_NAME] = STATE_SAME

                elif (update_info[MTIME_NAME] != root_info[MTIME_NAME] or
                    update_info[SIZE_NAME] != root_info[SIZE_NAME]):
                    root_info[MTIME_NAME] = update_info[MTIME_NAME]
                    root_info[SIZE_NAME] = update_info[SIZE_NAME]
                    root_info[STATE_NAME] = STATE_NEW
                    is_updated = True

            else:
                is_updated = True
                root_tree[FILES_NAME][update_name] = {
                    MTIME_NAME: update_tree[
                        FILES_NAME][update_name][MTIME_NAME],
                    SIZE_NAME: update_tree[
                        FILES_NAME][update_name][SIZE_NAME],
                    STATE_NAME: STATE_NEW}

        for update_dir in update_tree[DIRS_NAME]:
            # if update_dir in root_tree[DIRS_NAME].keys():
            real_dir_name = self.__get_equivarent_dir_name(
                                 update_dir, root_tree[DIRS_NAME].keys())
            if real_dir_name:
                update_info = update_tree[DIRS_NAME][update_dir]
                root_info = root_tree[DIRS_NAME][real_dir_name]

                if root_info[STATE_NAME] == STATE_DEL:
                    root_info[STATE_NAME] = STATE_SAME

                elif root_info[
                    STATE_NAME].startswith(ERR_STATE_PREFIX):
                    root_info[STATE_NAME] = re.sub(
                        "^" + ERR_STATE_PREFIX, "", root_info[STATE_NAME])
                    is_updated = True

                else:
                    # COMMENT exact_first時に
                    # 両方にほぼ同時刻のdir"A"があり、
                    # その下のfile,dirに差異がある場合に、
                    # get_sync_dup_tree_distrust_new()では
                    # dir"A"はsameになるが、
                    # その下はnewになり、dst.update()内では
                    # dir"A"はsameなので
                    # update_delta_infoされず、
                    # その下のfile,dirの転送の際に
                    # update_delta_infoされるので、そのとき
                    # update_delta_infoの"mtimeのない
                    # dirエントリができる"ルーチンを
                    # 通ることへの対処。
                    # mtimeがないと、
                    # 後のget_sync_dup_tree_distrust_newで
                    # KeyErrorになる
                    try:
                        root_info[MTIME_NAME]
                    except KeyError:
                        root_info[MTIME_NAME] = update_info[MTIME_NAME]
                        root_info[STATE_NAME] = STATE_SAME

            else:
                is_updated = True
                root_tree[DIRS_NAME][update_dir] = {
                    MTIME_NAME: update_tree[DIRS_NAME][update_dir][MTIME_NAME],
                    STATE_NAME: STATE_NEW,
                    CHILDREN_NAME: {DIRS_NAME: {}, FILES_NAME: {}}}

            if real_dir_name:
                target_dir_name = real_dir_name
            else:
                target_dir_name = update_dir

            if self.__set_update(
                root_tree[DIRS_NAME][target_dir_name][CHILDREN_NAME],
                update_tree[DIRS_NAME][update_dir][CHILDREN_NAME]):
                is_updated = True
        return is_updated

    def __set_delete(self, root_tree, update_tree):
        is_updated = False
        for file in root_tree[FILES_NAME]:

            #if (not file in update_tree[FILES_NAME].keys() and
            #    root_tree[FILES_NAME][file][STATE_NAME] != STATE_DEL):
            eq_file = self.__get_equivarent_file_name(
                          file, update_tree[FILES_NAME].keys())
            if (not eq_file and
                root_tree[FILES_NAME][file][STATE_NAME] != STATE_DEL):

                is_updated = True
                root_tree[FILES_NAME][file][STATE_NAME] = STATE_DEL

        for dir in root_tree[DIRS_NAME]:
            # if not dir in update_tree[DIRS_NAME].keys():
            eq_dir = self.__get_equivarent_dir_name(
                         dir, update_tree[DIRS_NAME].keys())
            if not eq_dir:
                if root_tree[DIRS_NAME][dir][STATE_NAME] != STATE_DEL:
                    is_updated = True
                    root_tree[DIRS_NAME][dir][STATE_NAME] = STATE_DEL
            else:
                if self.__set_delete(
                    root_tree[DIRS_NAME][dir][CHILDREN_NAME],
                    update_tree[DIRS_NAME][eq_dir][CHILDREN_NAME]):
                    is_updated = True
        return is_updated

    def __get_delta_tree(self, root_tree, update_tree):
        is_updated = False
        if self.__set_update(root_tree, update_tree):
            is_updated = True
        if self.__set_delete(root_tree, update_tree):
            is_updated = True
        return (is_updated, root_tree)

    def get_tree(self):
        """ Over ride. """
        try:
            def _get_tree():
                tree = {DIRS_NAME: {}, FILES_NAME: {}}
                for key in self.box_access.get_list(self.root):
                    self.__get_tree(key, tree)
                return tree

            def _is_retry_exc(ins):
                if not isinstance(ins, socket.error):
                    return False
                if ins.args[0] != errno.ECONNRESET:
                    return False
                return True

            for num in reversed(range(GET_TREE_RETRY_NUM)):
                try:
                    return _get_tree()
                except Exception, ins:
                    tb_info = sys.exc_info()[2]
                    if not _is_retry_exc(ins) or num == 0:
                        raise ins, None, tb_info

        except Exception, ins:
            self.logging(ERROR_LOG_LV, GET_TREE_LOG + " " + NG_LOG,
                src=self.root, trace=True)

            # get_listは戻り値をイテレーションしたときに
            # 例外発生するのでここでhandle
            box_except_handler(ins)

    def get_delta_tree(self):
        """ Over ride. """
        # COMMENT 戻り値:(更新有無, 差分情報ツリー)
        try:
            if not self.tree:
                self.tree = get_new_tree(self.get_tree())
                return (True, self.tree)
            else:
                update_tree = self.get_tree()
                is_updated, self.tree = self.__get_delta_tree(
                    self.tree, update_tree)
                return (is_updated, self.tree)
        except Exception, ins:
            self.logging(ERROR_LOG_LV, GET_DELTA_TREE_LOG + " " + NG_LOG,
                src=self.root, trace=True)

            # get_listは戻り値をイテレーションしたときに
            # 例外発生するのでここでhandle
            box_except_handler(ins)

    def reset_proxy_info(self, proxy_host, proxy_port):
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self.box_access = BoxAccess(
             token_id=self.token_id,
             mailaddress=self.mailaddress,
             passwd=self.passwd,
             encoding=self.encoding,
             proxy_host=self.proxy_host,
             proxy_port=self.proxy_port,
             except_handler=self.low_level_except_handler)

    def pre_file_transfer(self, file_relpath, file_info):
        if self.MAX_FILE_SIZE:
            if file_info[SIZE_NAME] >= self.MAX_FILE_SIZE:
                raise SizeTooLargeError

    def pre_dir_create(self, dir_relpath, dir_info):

        if self.for_sync:
            self.src.chmod(self.src.sep.join(
                [self.src.root] + dir_relpath), MODE_CH_DIR)
            src_dir_stat = self.src.stat(
                self.src.sep.join([self.src.root] + dir_relpath))
            src_mtime = src_dir_stat[stat.ST_MTIME]
            self.src.update_delta_info(dir_relpath, mtime=src_mtime)

    def post_file_transfer(self, file_relpath, file_info):

        if self.for_sync:
            dst_file_stat = self.stat(self.sep.join(
                [self.root] + file_relpath))
            dst_mtime = dst_file_stat[stat.ST_MTIME]
            dst_size = dst_file_stat[stat.ST_SIZE]
            self.update_delta_info(file_relpath, mtime=dst_mtime,
                                   size=dst_size, state=STATE_SAME,
                                   delete_keys=[ERRINFO_NAME])

            src = self.src
            src_abspath = self.sep.join([src.root] + file_relpath)
            src.chmod(src_abspath, MODE_CH_FILE)
            src_file_stat = src.stat(src_abspath)
            src_mtime = src_file_stat[stat.ST_MTIME]
            src_size = src_file_stat[stat.ST_SIZE]
            self.src.update_delta_info(file_relpath,
                                       mtime=src_mtime,
                                       size=src_size,
                                       state=STATE_SAME,
                                       delete_keys=[ERRINFO_NAME])
        else:
            self.src.update_delta_info(
                file_relpath, state=STATE_SAME)

    def post_dir_create(self, dir_relpath, dir_info):

        if self.for_sync:
            dst_dir_stat = self.stat(self.sep.join([self.root] + dir_relpath))
            dst_mtime = dst_dir_stat[stat.ST_MTIME]
            self.update_delta_info(dir_relpath, mtime=dst_mtime,
                                   state=STATE_SAME, is_dir=True,
                                   delete_keys=[ERRINFO_NAME])

            src = self.src
            src_abspath = self.sep.join([src.root] + dir_relpath)
            src.chmod(src_abspath, MODE_CH_DIR)
            src_dir_stat = src.stat(src_abspath)
            src_mtime = src_dir_stat[stat.ST_MTIME]
            self.src.update_delta_info(dir_relpath,
                                       mtime=src_mtime,
                                       state=STATE_SAME,
                                       delete_keys=[ERRINFO_NAME])
        else:
            self.src.update_delta_info(dir_relpath, state=STATE_SAME)

    def post_file_delete(self, file_relpath, file_info):
        if self.for_sync:
            self.update_delta_info(file_relpath, delete=True)

    def post_dir_delete(self, dir_relpath, dir_info):
        if self.for_sync:
            self.update_delta_info(dir_relpath, delete=True)

    def _SrcFull__update_delta_info(self, tree, path_list,
                            mtime=None, size=None,
                            state=None, delete=False, is_dir=False,
                            update_dict={}, delete_keys=[]):

        path = path_list[0]
        # if path in tree[DIRS_NAME].keys():
        real_dir = self.__get_equivarent_dir_name(
                                 path, tree[DIRS_NAME].keys())
        if real_dir:
            if len(path_list) == 1:
                if mtime != None:
                    tree[DIRS_NAME][real_dir][MTIME_NAME] = mtime
                if size != None:
                    tree[DIRS_NAME][real_dir][SIZE_NAME] = size
                if state != None:
                    tree[DIRS_NAME][real_dir][STATE_NAME] = state

                tree[DIRS_NAME][real_dir].update(update_dict)

                for key in delete_keys:
                    try:
                        del tree[DIRS_NAME][real_dir][key]
                    except KeyError:
                        pass

                if delete:
                    del tree[DIRS_NAME][real_dir]
            else:
                self._SrcFull__update_delta_info(
                            tree[DIRS_NAME][real_dir][CHILDREN_NAME],
                            path_list[1:], mtime, size, state,
                            delete, is_dir, update_dict, delete_keys)
            return

        # if path in tree[FILES_NAME].keys():
        real_file = self.__get_equivarent_file_name(
                                 path, tree[FILES_NAME].keys())
        if real_file:
            if mtime != None:
                tree[FILES_NAME][real_file][MTIME_NAME] = mtime
            if size != None:
                tree[FILES_NAME][real_file][SIZE_NAME] = size
            if state != None:
                tree[FILES_NAME][real_file][STATE_NAME] = state

            tree[FILES_NAME][real_file].update(update_dict)

            for key in delete_keys:
                try:
                    del tree[FILES_NAME][real_file][key]
                except KeyError:
                    pass

            if delete:
                del tree[FILES_NAME][real_file]

            return

        if not delete:
            if len(path_list) > 1:
                # COMMENT ここにくると、
                # dirのmtimeがないツリーができる。
                tree[DIRS_NAME][path] = {
                    CHILDREN_NAME: {DIRS_NAME: {}, FILES_NAME: {}},
                    STATE_NAME: STATE_SAME}
                self._SrcFull__update_delta_info(
                    tree[DIRS_NAME][path][CHILDREN_NAME],
                    path_list[1:], mtime, size, state,
                    delete, is_dir, update_dict, delete_keys)

            if len(path_list) == 1:
                if is_dir:
                    tree[DIRS_NAME][path] = {
                        MTIME_NAME: mtime,
                        CHILDREN_NAME: {DIRS_NAME: {}, FILES_NAME: {}},
                        STATE_NAME: state}
                    tree[DIRS_NAME][path].update(update_dict)
                else:
                    tree[FILES_NAME][path] = {
                        MTIME_NAME: mtime,
                        SIZE_NAME: size,
                        STATE_NAME: state}
                    tree[FILES_NAME][path].update(update_dict)
