# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""Module to access the BOX server with REST API."""

from __future__ import with_statement

try:
    import json
except:
    import simplejson as json

import os
import ssl
try:
    import cloudsync.box.urllib2 as urllib2
except:
    import urllib2
import cookielib
from urllib import (urlencode,)
from urlparse import (urlparse, parse_qs)
from HTMLParser import (HTMLParser,)

import time
import calendar
import re
try:
    import cloudsync.box.httplib as httplib
except:
    import httplib
from ionas.restapi.accessinfo import (AccessInfo, RequestInfo, NeedToRetry)
from ionas.restapi.pathid import (PathIdMap)
from pkgaddon.box_config import (BOX_CONF_DIR, BOX_CONF_FILE,
                                    HTTP_USER_AGENT_TMPL, ROOT_PATH,
                                    ROOT_FOLDER_ID)
from util import (to_str, to_utf8, get_stream, get_stat_info,
                    call_as_root, UploadFileData)


SSL_CREATE_UNVERIFIED_CONTEXT = "_create_unverified_context"

# For parsing a HTML page
FORM_TAG_NAME = "form"
INPUT_TAG_NAME = "input"
BUTTON_TAG_NAME = "button"
ID_ATTR = "id"
NAME_ATTR = "name"
VALUE_ATTR = "value"
TYPE_ATTR = "type"
ACTION_ATTR = "action"
SUBMIT_TYPE = "submit"

LOGIN_NAME = "login"
PASSWORD_NAME = "password"
REQUEST_TOKEN_NAME = "request_token"
CODE_NAME = "code"
ERROR_NAME = "error"

LOCATION_KEY = "Location"

MTIME_UTC_RE = "(.*)Z"
MTIME_TZ_RE = "(.*)([+-])(\d\d):*(\d\d)"
MTIME_FORMAT = "%Y-%m-%dT%H:%M:%S"

MAX_RETRY_COUNT = 1
SERVER_ERR_RETRY_COUNT = 3
RETRY_SLEEP_TIME = 10

ACCESS_TOKEN_NAME = "access_token"
REFRESH_TOKEN_NAME = "refresh_token"
TOTAL_COUNT_NAME = "total_count"
ENTRIES_NAME = "entries"
STAT_NAME = "stat"
TYPE_NAME = "type"
ID_NAME = "id"
NAME_NAME = "name"
SIZE_NAME = "size"
MODIFIED_AT_NAME = "modified_at"
PARENT_NAME = "parent"
CONTEXT_INFO_NAME = "context_info"
CONFLICTS_NAME = "conflicts"

FOLDER_TYPE = "folder"
FILE_TYPE = "file"

FOLDERS_ITEMS_PARAM_TMPL = \
    "fields=type,id,name,size,modified_at&limit=%d&offset=%d"
RECURSIVE_PARAM = "recursive=true"
ITEM_LIMIT = 1000

FOLDERS_URL_TMPL = "/folders"
FOLDERS_ID_URL_TMPL = "/folders/%s"
FOLDERS_ITEMS_URL_TMPL = FOLDERS_ID_URL_TMPL + "/items"
MAKE_DIRECTORY_URL_TMPL = FOLDERS_URL_TMPL
FILES_URL_TMPL = "/files"
FILES_ID_URL_TMPL = "/files/%s"
FILE_READ_URL_TMPL = FILES_ID_URL_TMPL + "/content"
FILE_WRITE_URL_TMPL = FILES_URL_TMPL + "/content"
FILE_UPDATE_URL_TMPL = FILES_ID_URL_TMPL + "/content"

POST_METHOD = "POST"
PUT_METHOD = "PUT"
DELETE_METHOD = "DELETE"

BOUNDARY_MAIN = "BOX_FILE_UPLOAD_BOUDARY"

UPLOAD_ATTRIBUTES_TMPL = '{"name": "%s", "parent": {"id": "%s"}}'

INTERNAL_ERROR_FILE = "intarnal"

ERR_INVALID_GRANT = "invalid_grant"


def _get_error_stream(status, err_code):
    """Return a stream include the JSON format error information."""

    err_data = {
            "type": "error",
            "status": status,
            "code": err_code,
        }

    return get_stream(json.dumps(err_data))


def to_json(response):
    """Load JSON data from "response" and return it."""

    json_response = {}
    try:
        json_response = json.loads(response)
    except:
        pass
    return json_response


def get_code_in_response(response):
    """Return a value of "code" in a JSON data "response"."""

    return to_json(response)[CODE_NAME]


def recreate_http_error(http_err, err_data):
    """Recreate urllib2.HTTPError instance and return it."""

    return urllib2.HTTPError(http_err.filename, http_err.code, http_err.msg,
                              http_err.hdrs, get_stream(err_data))


class NoHTTPRedirectHandler(urllib2.HTTPRedirectHandler):
    """Class for no redirection."""

    def http_error_302(self, req, fp, code, msg, headers):
        """Return fp for no redirection."""

        return fp

    http_error_301 = http_error_303 = http_error_307 = http_error_302


class HtmlPageError(Exception):
    """Error class for a invalid HTML page."""

    pass


class LoginError(Exception):
    """Error class for a login error."""

    def __init__(self, name):
        """Constructor."""

        Exception.__init__(self, "Login Error: %s" % (name,))


class ErrCode:
    """Class for defining the error code."""

    NOT_FOUND = "not_found"
    ITEM_NAME_IN_USE = "item_name_in_use"


class NotFoundHttpError(urllib2.HTTPError):
    """Class for the dummy "HTTPError" representing the internal error."""

    def __init__(self):
        """Constructor."""

        urllib2.HTTPError.__init__(self, INTERNAL_ERROR_FILE,
                            httplib.NOT_FOUND,
                            httplib.responses.get(httplib.NOT_FOUND, ""), {},
                            _get_error_stream(httplib.NOT_FOUND,
                                                ErrCode.NOT_FOUND))


class BoxHtmlParser(HTMLParser):
    """Class for parse BOX HTML page."""

    _LOGIN_FORM_NAME = "login_form"
    _CONSENT_FORM_NAME = "consent_form"
    _IGNORE_INPUT_LIST = ["consent_reject"]

    TYPE_UNKNOWN = 0
    TYPE_LOGIN = 1
    TYPE_CONSENT = 2

    def _to_attrs_dict(self, attrs):
        """Return a dictionary of attributes in "attrs"."""

        attrs_dict = {}
        for (key, value) in attrs:
            attrs_dict[key] = value
        return attrs_dict

    def _handle_input_tag(self, id_name, attrs_dict):
        """Handle input tag."""

        if NAME_ATTR in attrs_dict:
            if attrs_dict[NAME_ATTR] in self._IGNORE_INPUT_LIST:
                return

            if VALUE_ATTR in attrs_dict:
                self.form_data[attrs_dict[NAME_ATTR]] = attrs_dict[VALUE_ATTR]
            else:
                self.form_data[attrs_dict[NAME_ATTR]] = ""

    def __init__(self):
        """Constructor."""

        HTMLParser.__init__(self)
        self.form_data = {}
        self.page_type = self.TYPE_UNKNOWN
        self.action = ""

    def handle_starttag(self, tag, attrs):
        """Handle start tag."""

        attrs_dict = self._to_attrs_dict(attrs)
        id_name = attrs_dict.get(ID_ATTR, "")
        if tag == FORM_TAG_NAME:
            if attrs_dict[NAME_ATTR] == self._LOGIN_FORM_NAME:
                self.page_type = self.TYPE_LOGIN
                self.action = attrs_dict[ACTION_ATTR]
            elif attrs_dict[NAME_ATTR] == self._CONSENT_FORM_NAME:
                self.page_type = self.TYPE_CONSENT
                self.action = attrs_dict[ACTION_ATTR]

        elif tag == INPUT_TAG_NAME:
            self._handle_input_tag(id_name, attrs_dict)
        elif tag == BUTTON_TAG_NAME:
            if attrs_dict.get(TYPE_ATTR) == SUBMIT_TYPE:
                self._handle_input_tag(id_name, attrs_dict)

    def handle_endtag(self, tag):
        """Handle end tag."""

        pass

    def get_page_type(self):
        """Return the target page type."""

        return self.page_type

    def get_action(self):
        """Return form action."""

        return self.action

    def get_form_data(self):
        """Return form data."""

        return self.form_data

    def set_form_data(self, name, value):
        """Add "name" and "value" pair to a form data."""

        self.form_data[name] = value


class BoxUploadFileData(UploadFileData):
    """Class for upload file data for BOX."""

    def _get_prefix_string(self, file_name):
        """Return the prefix string data for uploading."""

        prefix = self._get_form_data_string(
                            "attributes",
                            UPLOAD_ATTRIBUTES_TMPL %
                                (file_name, self.parent_id))
        prefix += self._get_form_data_string(self.FILE_NAME,
                                type="application/octet-stream",
                                file_name=file_name)
        prefix += self.LINE_END
        return prefix

    def __init__(self, file_name, parent_id, data, length, boundary,
            overwrite=False):
        """Constructor."""

        self.parent_id = to_utf8(parent_id)
        self.overwrite = overwrite
        UploadFileData.__init__(self, to_utf8(file_name), data, length,
                                boundary)

    def do_overwrite(self):
        """Return if overwrite data."""

        return self.overwrite


class BoxTokenMgr:
    """Class to manage access token for BOX."""

    TOKEN_SEP = "\t"

    class TokenInfo:
        def __init__(self, mailaddress, password, access_token, refresh_token):
            """Constructor."""

            self.mailaddress = to_utf8(mailaddress)
            self.password = to_utf8(password)
            self.access_token = to_utf8(access_token)
            self.refresh_token = to_utf8(refresh_token)

        def is_user_info_same(self, mailaddress, password):
            """Check if mailaddress and password are same."""

            if self.mailaddress != mailaddress:
                return False
            if self.password != password:
                return False
            return True

    def _load_token_info(self):
        """Load saved token information."""

        if not os.path.exists(self.conf_path):
            return

        with open(self.conf_path) as conf:
            token_id = ""
            mailaddress = ""
            access_token = ""
            refresh_token = ""
            for line in conf:
                if token_id:
                    self.token_info_map[token_id] = \
                        self.TokenInfo(mailaddress, line.rstrip("\n"),
                                access_token, refresh_token)
                    token_id = ""
                else:
                    try:
                        token_id, mailaddress, access_token, refresh_token = \
                                    line.rstrip("\n").split(self.TOKEN_SEP)
                    except:
                        token_id = ""

    def _save_token_info(self):
        """Save token information."""

        temp_path = self.conf_path + "." + str(os.getpid())
        if not os.path.exists(os.path.dirname(self.conf_path)):
            os.mkdir(os.path.dirname(self.conf_path))

        try:
            with open(temp_path, "w") as conf:
                for (token_id, token_info) in self.token_info_map.items():
                    conf.write(self.TOKEN_SEP.join(
                            [
                                token_id,
                                token_info.mailaddress,
                                token_info.access_token,
                                token_info.refresh_token
                            ]) + "\n")
                    conf.write(token_info.password + "\n")

            os.rename(temp_path, self.conf_path)
        except:
            if os.path.exists(temp_path):
                os.remove(temp_path)
            raise

    def _request(self, url, param=None):
        """Send the HTTP request to "url" with "param"."""

        data = param
        if param != None:
            data = urlencode(param)

        request = urllib2.Request(url, data)
        return self.opener.open(request)

    def _request_and_html_parse(self, url, param=None):
        """Send the HTTP request to "url" with "param" and return
           the HTML parser for result.
        """

        response = self._request(url, param)
        parser = self.html_parser_cls()
        parser.feed(response.read())
        return parser

    def _authorize(self):
        """Request the authorize page to the BOX OAuth server and return
           the HTML parser for result.
        """

        param = {
            "response_type": "code",
            "client_id": self.client_id,
            "redirect_url": self.redirect_url,
        }

        return self._request_and_html_parse(
                        self.base_url + "/authorize" + "?" + urlencode(param))

    def _login(self, mailaddress, password, parser):
        """Login the BOX OAuth server and return a the HTML parser
           for result.
        """

        param = parser.get_form_data()
        request_token = ""
        if REQUEST_TOKEN_NAME in param:
            request_token = param[REQUEST_TOKEN_NAME]

        param[LOGIN_NAME] = mailaddress
        param[PASSWORD_NAME] = password
        login_parser = self._request_and_html_parse(parser.get_action(), param)
        login_parser.set_form_data(REQUEST_TOKEN_NAME, request_token)
        return login_parser

    def _get_auth_code(self, mailaddress, password):
        """Get authorization code from the BOX OAuth server and return it."""

        authorized_parser = self._authorize()
        if authorized_parser.get_page_type() != BoxHtmlParser.TYPE_LOGIN:
            raise HtmlPageError("page is not for login!!")

        logined_parser = self._login(mailaddress, password,
                                        authorized_parser)
        if logined_parser.get_page_type() != BoxHtmlParser.TYPE_CONSENT:
            raise LoginError(mailaddress)

        response = self._request(logined_parser.get_action(),
                                logined_parser.get_form_data())
        query_dict = parse_qs(
                        urlparse(response.info()[LOCATION_KEY]).query)

        return query_dict[CODE_NAME][0]

    def _update_token_info(self, token_id, mailaddress, password, param):
        """Update token information for "mailaddress"."""

        response = self._request(self.base_url + "/token", param)
        response_data = json.loads(response.read())
        self.token_info_map[token_id] = \
                        self.TokenInfo(mailaddress, password,
                                response_data[ACCESS_TOKEN_NAME],
                                response_data[REFRESH_TOKEN_NAME])
        call_as_root(self._save_token_info)

    def _revoke_token(self, token):
        """Revoke token "mailaddress"."""

        param = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "token": token,
        }

        self._request(self.base_url + "/revoke", param)

    def _delete_token_info(self, token_id):
        """Delete token information for "token_id"."""

        try:
            self._revoke_token(self.token_info_map[token_id].refresh_token)
        except:
            # Ignore error
            pass

        del self.token_info_map[token_id]
        call_as_root(self._save_token_info)

    def __init__(self, base_url, client_id, client_secret,
            redirect_url, proxies=None, conf_dir=BOX_CONF_DIR,
            html_parser_cls=BoxHtmlParser):
        """Constructor."""

        self.base_url = base_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_url = redirect_url
        self.conf_path = os.path.join(conf_dir, BOX_CONF_FILE)
        self.html_parser_cls = html_parser_cls
        self.opener = None
        self.set_proxies(proxies)
        self.token_info_map = {}
        call_as_root(self._load_token_info)

    def set_proxies(self, proxies):
        """Build a opener to access to the server with "proxies"."""

        handlers = []
        if hasattr(ssl, SSL_CREATE_UNVERIFIED_CONTEXT):
            handlers += [urllib2.HTTPSHandler(
                                    context=ssl._create_unverified_context())]
        if proxies != None:
            handlers += [urllib2.ProxyHandler(proxies)]

        cookie_jar = cookielib.CookieJar()
        handlers += [NoHTTPRedirectHandler(),
                        urllib2.HTTPCookieProcessor(cookie_jar)]
        self.opener = urllib2.build_opener(*handlers)
        return proxies

    def get_access_token(self, token_id, mailaddress, password):
        """Return a access token for the content API of BOX server."""

        if (token_id in self.token_info_map and
            not self.token_info_map[token_id].is_user_info_same(
                                                mailaddress,
                                                password)):

            self._delete_token_info(token_id)

        if not token_id in self.token_info_map:
            auth_code = self._get_auth_code(mailaddress, password)

            param = {
                "grant_type": "authorization_code",
                "code": auth_code,
                "client_id": self.client_id,
                "client_secret": self.client_secret,
                "redirect_url": self.redirect_url,
            }

            self._update_token_info(token_id, mailaddress, password, param)

        return self.token_info_map[token_id].access_token

    def refresh_access_token(self, token_id):
        """Refresh a access token and return it."""

        token_info = self.token_info_map[token_id]
        param = {
            "grant_type": "refresh_token",
            "refresh_token": token_info.refresh_token,
            "client_id": self.client_id,
            "client_secret": self.client_secret,
        }

        try:
            self._update_token_info(token_id,
                                token_info.mailaddress,
                                token_info.password,
                                param)
        except urllib2.HTTPError, err:
            try:
                err_data = err.read()
                json_response = to_json(err_data)
                if json_response[ERROR_NAME] == ERR_INVALID_GRANT:
                    # Token has been expired. Get access token again.
                    del self.token_info_map[token_id]
                    return self.get_access_token(token_id,
                                                    token_info.mailaddress,
                                                    token_info.password)
            except:
                # Ignore a new exception.
                pass
            raise recreate_http_error(err, err_data)

        return token_info.access_token


class BoxInfo(AccessInfo):
    """Class to access to the BOX server."""

    def _normalize_path(self, path):
        """Delete a first "/" if path starts with "//"."""

        if path.startswith("//"):
            return path[1:]
        return path

    def _set_authrization_header(self, headers):
        """Set "Authrization" header for "headers"."""

        headers["Authorization"] = "Bearer " + \
                self.token_mgr.get_access_token(self.token_id,
                                                self.mailaddress,
                                                self.password)

    def _get_headers(self):
        """Return headers used in common requests."""

        headers = {"User-Agent": HTTP_USER_AGENT_TMPL}

        self._set_authrization_header(headers)
        return headers

    def _to_mtime(self, mtime_str):
        """Transform "mtime_str" to a UTC epoch time and return it."""

        if not mtime_str:
            return 0

        match = re.match(MTIME_UTC_RE, mtime_str)
        tz_adjust = 0
        if match == None:
            match = re.match(MTIME_TZ_RE, mtime_str)
            tz_adjust = int(match.group(3)) * 3600 + int(match.group(4)) * 60
            if match.group(2) == "+":
                tz_adjust *= -1

        time_tuple = time.strptime(match.group(1), MTIME_FORMAT)
        return int(calendar.timegm(time_tuple)) + tz_adjust

    def _get_id_dict(self, contents):
        """Return dictionary of the id in "contents"."""

        id_dict = {}
        for content in contents.get(ENTRIES_NAME, []):
            name = content.get(NAME_NAME, "")
            id = content[ID_NAME]
            if name:
                id_dict[name] = id

        return id_dict

    def _init_ids(self):
        """Initialize each IDs."""

        self.path_id_map = PathIdMap()
        self.path_id_map.append_path(ROOT_PATH, ROOT_FOLDER_ID)

    def _handle_unauthorized(self, json_response, request_info, retry_count):
        """Handle HTTP UNAUTHORIZED error."""

        if retry_count < MAX_RETRY_COUNT:
            self.token_mgr.refresh_access_token(self.token_id)
            self._set_authrization_header(request_info.get_headers())
            raise NeedToRetry(request_info)

    def _handle_conflict(self, json_response, request_info, retry_count):
        """Handle HTTP CONFLICT error."""

        try:
            code = json_response[CODE_NAME]
            id = json_response[CONTEXT_INFO_NAME][CONFLICTS_NAME][ID_NAME]
        except:
            return

        if not id or code != ErrCode.ITEM_NAME_IN_USE:
            return

        data = request_info.get_data()
        file_write_url = self.upload_base_url + FILE_WRITE_URL_TMPL

        if (not request_info.get_url().startswith(file_write_url) or
            retry_count != 0 or
            not hasattr(data, "do_overwrite") or
            not data.do_overwrite()):

            # Can't retry other request
            return

        url = self.upload_base_url + FILE_UPDATE_URL_TMPL % (id,)
        raise NeedToRetry(RequestInfo(url, data, request_info.get_headers()))

    def __init__(self, base_url, upload_base_url, token_id, mailaddress,
            password, token_mgr):
        """Constructor."""

        self.base_url = base_url
        self.upload_base_url = upload_base_url
        self.token_id = token_id
        self.mailaddress = mailaddress
        self.password = password
        self.token_mgr = token_mgr

        self.target_name = ""
        self.path_info_base_path = ""
        self.next_offset = 0
        self.path_id_map = None
        self._init_ids()

    def set_path_info(self, response):
        """Set the path information which is in "response"."""

        contents = json.loads(to_str(response))
        id_dict = self._get_id_dict(contents)

        for name in id_dict:
            self.path_id_map.append_path(
                                os.path.join(self.path_info_base_path, name),
                                id_dict[name])
        
        if self.next_offset >= contents[TOTAL_COUNT_NAME]:
            self.next_offset = 0
            try:
                if (self.target_name and
                    self.path_id_map.get_id(
                                        os.path.join(self.path_info_base_path,
                                        self.target_name)) == None):

                    raise NotFoundHttpError()
            except:
                raise NotFoundHttpError()

    def set_create_path_info(self, path, response):
        """Set the path information which is in "response" of request
        for creating the "path".
        """

        path = self._normalize_path(path)
        contents = json.loads(to_str(response))
        id = ""
        if ENTRIES_NAME in contents:
            name = os.path.basename(path)
            id_dict = self._get_id_dict(contents)
            if name in id_dict:
                id = id_dict[name]
        else:
            id = contents.get(ID_NAME, "")

        if id:
            self.path_id_map.append_path(path, id)

    def to_content_list(self, response):
        """Convert "response" to the content list."""

        contents = json.loads(to_str(response))
        content_list = []
        for content in contents.get(ENTRIES_NAME, []):
            content_info = {NAME_NAME: content[NAME_NAME],
                            STAT_NAME: self.to_stat_info(json.dumps(content))}

            content_list.append(content_info)

        return content_list

    def to_stat_info(self, response):
        """Convert "response" to the stat information."""

        json_stat_info = json.loads(to_str(response))
        is_dir = json_stat_info[TYPE_NAME] == FOLDER_TYPE
        if is_dir:
            size = 0
        else:
            size = json_stat_info.get(SIZE_NAME, 0)

        mtime = self._to_mtime(json_stat_info[MODIFIED_AT_NAME])

        return get_stat_info(is_dir, size, mtime)

    def handle_exception(self, exc, request_info, retry_count):
        """Handle "exc" exception."""

        try:
            raise exc
        except urllib2.HTTPError, err:
            # Reset data stream for retry.
            data = request_info.get_data()
            if data != None and hasattr(data, "reset_stream"):
                data.reset_stream()

            if (err.code == httplib.INTERNAL_SERVER_ERROR or
                err.code == httplib.BAD_GATEWAY or
                err.code == httplib.SERVICE_UNAVAILABLE or
                err.code == httplib.GATEWAY_TIMEOUT):

                if retry_count < SERVER_ERR_RETRY_COUNT:
                    time.sleep(RETRY_SLEEP_TIME)
                    raise NeedToRetry(request_info)
                raise

            err_data = err.read()
            json_response = to_json(err_data)

            if err.code == httplib.UNAUTHORIZED:
                self._handle_unauthorized(json_response, request_info,
                                            retry_count)
            elif err.code == httplib.CONFLICT:
                self._handle_conflict(json_response, request_info, retry_count)
            raise recreate_http_error(err, err_data)

    def get_prepare_path_req_info(self, path, purpose):
        """Return the "RequestInfo" to prepare accesing a "path"."""

        DIRECT_DIRECTORY_PURPOSE_LIST = [
                AccessInfo.Purpose.ACCESS_DIRECTORY,
                AccessInfo.Purpose.DELETE_DIRECTORY,
                AccessInfo.Purpose.RENAME_DIRECTORY,
                AccessInfo.Purpose.MOVE_DIRECTORY,
                AccessInfo.Purpose.MOVE_TO,
            ]

        DIRECT_FILE_PURPOSE_LIST = [
                AccessInfo.Purpose.ACCESS_FILE,
                AccessInfo.Purpose.DELETE_FILE,
                AccessInfo.Purpose.RENAME_FILE,
                AccessInfo.Purpose.MOVE_FILE,
            ]

        PARENT_DIRECTORY_PURPOSE_LIST = [
                AccessInfo.Purpose.MAKE_DIRECTORY,
                AccessInfo.Purpose.WRITE_FILE,
            ]

        path = self._normalize_path(path)
        target_path = ""
        if purpose in DIRECT_DIRECTORY_PURPOSE_LIST:
            target_path = path
            self.path_id_map.append_directory(target_path)
        elif purpose in DIRECT_FILE_PURPOSE_LIST:
            target_path = path
            self.path_id_map.append_path(target_path)
        elif purpose in PARENT_DIRECTORY_PURPOSE_LIST:
            target_path = os.path.dirname(path)
            self.path_id_map.append_directory(target_path)
        else:
            self.next_offset = 0
            return None

        if (not self.next_offset and
            self.path_id_map.get_id(target_path) != None):

            self.next_offset = 0
            return None

        while True:
            target_name = os.path.basename(target_path)
            target_path = os.path.dirname(target_path)
            parent_id = self.path_id_map.get_id(target_path)
            if parent_id != None:
                break

        if target_name == ROOT_PATH:
            target_name = ""
        self.target_name = target_name
        self.path_info_base_path = target_path
        url = self.base_url + FOLDERS_ITEMS_URL_TMPL % (parent_id,)
        url += "?" + FOLDERS_ITEMS_PARAM_TMPL % (ITEM_LIMIT, self.next_offset)
        self.next_offset += ITEM_LIMIT

        return RequestInfo(url, headers=self._get_headers())

    def get_list_contents_req_info(self, path, offset=0, limit=ITEM_LIMIT):
        """Return the "RequestInfo" to list contents in "path" directory."""

        path = self._normalize_path(path)
        url = self.base_url + FOLDERS_ITEMS_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        url += "?" + FOLDERS_ITEMS_PARAM_TMPL % (limit, offset)
        return RequestInfo(url, headers=self._get_headers())

    def get_make_directory_req_info(self, path):
        """Return the "RequestInfo" to make "path" directory."""

        path = self._normalize_path(path)
        url = self.base_url + MAKE_DIRECTORY_URL_TMPL
        param = {
                    NAME_NAME: to_utf8(os.path.basename(path)),
                    PARENT_NAME: {ID_NAME: \
                            self.path_id_map.get_id(os.path.dirname(path))},
            }
        return RequestInfo(url, json.dumps(param),
                            headers=self._get_headers())

    def get_delete_directory_req_info(self, path):
        """Return the "RequestInfo" to delete "path" directory."""

        path = self._normalize_path(path)
        url = self.base_url + FOLDERS_ID_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        url += "?" + RECURSIVE_PARAM
        return RequestInfo(url, headers=self._get_headers(),
                            method=DELETE_METHOD)

    def get_stat_directory_req_info(self, path):
        """Return the "RequestInfo" to get status of "path" directory."""

        path = self._normalize_path(path)
        url = self.base_url + FOLDERS_ID_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        return RequestInfo(url, headers=self._get_headers())

    def get_rename_directory_req_info(self, path, new_name):
        """Return the "RequestInfo" to rename "path" directory."""

        path = self._normalize_path(path)
        url = self.base_url + FOLDERS_ID_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        param = {NAME_NAME: to_utf8(new_name)}
        return RequestInfo(url, json.dumps(param),
                            headers=self._get_headers(),
                            method=PUT_METHOD)

    def get_move_directory_req_info(self, path, parent_path):
        """Return the "RequestInfo" to move "path" directory."""

        path = self._normalize_path(path)
        url = self.base_url + FOLDERS_ID_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        param = {PARENT_NAME: {ID_NAME: \
                            self.path_id_map.get_id(parent_path)}}
        return RequestInfo(url, json.dumps(param),
                            headers=self._get_headers(),
                            method=PUT_METHOD)

    def get_delete_file_req_info(self, path):
        """Return the "RequestInfo" to delete "path" file."""

        path = self._normalize_path(path)
        url = self.base_url + FILES_ID_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        return RequestInfo(url, headers=self._get_headers(),
                            method=DELETE_METHOD)

    def get_stat_file_req_info(self, path):
        """Return the "RequestInfo" to get status of "path" file."""

        path = self._normalize_path(path)
        url = self.base_url + FILES_ID_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        return RequestInfo(url, headers=self._get_headers())

    def get_rename_file_req_info(self, path, new_name):
        """Return the "RequestInfo" to rename "path" file."""

        path = self._normalize_path(path)
        url = self.base_url + FILES_ID_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        param = {NAME_NAME: to_utf8(new_name)}
        return RequestInfo(url, json.dumps(param),
                            headers=self._get_headers(),
                            method=PUT_METHOD)

    def get_move_file_req_info(self, path, parent_path):
        """Return the "RequestInfo" to move "path" file."""

        path = self._normalize_path(path)
        url = self.base_url + FILES_ID_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        param = {PARENT_NAME: {ID_NAME: \
                            self.path_id_map.get_id(parent_path)}}
        return RequestInfo(url, json.dumps(param),
                            headers=self._get_headers(),
                            method=PUT_METHOD)

    def get_read_file_req_info(self, path):
        """Return the "RequestInfo" to read from "path" file."""

        path = self._normalize_path(path)
        url = self.base_url + FILE_READ_URL_TMPL % \
                                        (self.path_id_map.get_id(path),)
        return RequestInfo(url, headers=self._get_headers())

    def get_write_file_req_info(self, path, data, length=0, overwrite=False):
        """Return the "RequestInfo" to write to "path" file."""

        path = self._normalize_path(path)
        url = self.upload_base_url + FILE_WRITE_URL_TMPL
        upload_file_data = BoxUploadFileData(
                                to_utf8(os.path.basename(path)),
                                self.path_id_map.get_id(
                                                    os.path.dirname(path)),
                                data, length,
                                BOUNDARY_MAIN, overwrite)

        req_headers = self._get_headers()
        req_headers.update(upload_file_data.get_headers())
        return RequestInfo(url, upload_file_data,
                        headers=req_headers)
