#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

from buffer_object import BufferObject, BufferObjectW, BufferObjectR, DEFAULT_BUFFER_SIZE
import Queue
import os


class ChunkedBuffer2(BufferObject):
    """Chunked-BufferObject2
    Constituted by individual closed possible plurality of buffers
    """

    def __init__(
            self,
            size=-1,
            buffer_size=DEFAULT_BUFFER_SIZE,
            queue_maxsize=1,
    ):
        """Constructor
        Args:
        size (int): size to be written
        buffer_size (int): buffer(chunk) size
        """
        self._buffer_size = buffer_size
        super(ChunkedBuffer2, self).__init__()
        self._size = size
        self._buffers = Queue.Queue(maxsize=queue_maxsize)
        self._r_buf = None
        self._w_buf = BufferObject(size=min(self._buffer_size, self._size))
        self._kill_flag = False
        self._do_read = False

        if size < 0:
            raise ValueError("Invalid size")

    @property
    def lack(self):
        if self._w_buf:
            return self._w_buf.lack
        else:
            return 0

    def seek(self, pos, whence=os.SEEK_SET):
        if self._do_read:
            raise RuntimeError("No rewinding")
        super(ChunkedBuffer2, self).seek(pos, whence)

    def close(self):
        while True:
            try:
                buf = self._buffers.get(block=False)
                buf.close()
                del buf
            except Queue.Empty:
                break

    def read(self, size=-1):
        if self._r_pos == self._size:
            return ""

        try:
            if size < 0:
                size = self._size
            size = min(size, self._size - self._r_pos)

            w_buf = BufferObjectW(size=size)
            while w_buf.lack > 0:
                if self._kill_flag:
                    break

                if self._r_buf is None:
                    try:
                        self._r_buf = self._buffers.get(timeout=10)
                    except Queue.Empty:
                        raise

                w_buf.write(self._r_buf.read(w_buf.lack))

                if self._r_buf.rest == 0:
                    self._r_buf.close()
                    del self._r_buf
                    self._r_buf = None

            data = w_buf.read()
            self._r_pos += w_buf.length
            w_buf.close()
            self._do_read = True
            return data
        except Exception:
            self._kill_flag = True
            raise

    def write(self, data):
        if len(data) == 0 and self._size == 0:
            self._buffers.put(BufferObject(size=0), timeout=10)
        else:
            try:
                r_buf = BufferObjectR(data=data)
                while r_buf.rest > 0:
                    # Force Quit
                    if self._kill_flag:
                        break

                    if self._w_buf is None:
                        size = min(self._buffer_size, self._size - self._w_pos)
                        self._w_buf = BufferObject(size=size)

                    data = r_buf.read(self._w_buf.lack)
                    self._w_buf.write(data)
                    self._w_pos += len(data)

                    if self._w_buf.lack == 0:
                        self._buffers.put(self._w_buf, timeout=10)
                        self._w_buf = None

                r_buf.close()
                self._len = self._w_pos
            except Exception:
                self._kill_flag = True
                raise

    def clear(self):
        raise NotImplementedError
