#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta
"""Module for each logging information."""

import os
import re
import urllib2
import locale
import logging
import inspect
from logging.handlers import (RotatingFileHandler,)
from StringIO import StringIO
from ionas.fileaccess import (DEFAULT_LOCALE, MODE_WRITE, MODE_APPEND,)
from ionas.filesync import (SKIP_ERROR_LOG_LV, FILE_SIZE_OVER_LOG)

from ionas.exception import (BaseAccessException,
                             NoEntryError,
                             NoRootEntryError,
                             InvalidFileSystemError,
                             NoSpaceError,
                             ExistsError,
                             SizeTooLargeError,
                             InternalError,
                             NameTooLongError,
                             PermissionError)

from ionas.filesync.basefilesync import (IgnoreTargetFile)
from iobase.exception import (NotFoundError, IoBaseException)
from ionas.filesync import (TYPE_FILE_TRANSFER_OK,
                            TYPE_FILE_TRANSFER_NG,
                            TYPE_SYNC_SKIP,
                            OK_LOG, NG_LOG,
                            FILE_DELETE_LOG,
                            DIR_CREATE_LOG,
                            DIR_DELETE_LOG)
from ionas.filesync import (MODE_CH_FILE, MODE_CH_DIR)

from cloudsync.util import (lf2crlf,)


HEADER = "header"
BODY = "body"

RESULT = "result"
SHARE = "share"
JOB = "job"
START = "start"
END = "end"
ELAPSE = "elapse"
NG_FILES = "ngfiles"
NG_DIRS = "ngdirs"
TOTAL_FILES = "totfiles"
TOTAL_DIRS = "totdirs"

RESULT_TMPL = "%(" + RESULT + ")s"
SHARE_TMPL = "%(" + SHARE + ")s"
JOB_TMPL = "%(" + JOB + ")s"
START_TMPL = "%(" + START + ")s"
END_TMPL = "%(" + END + ")s"
ELAPSE_TMPL = "%(" + ELAPSE + ")s"

SUMMARY_OK_KEY = "OK"
SUMMARY_NG_KEY = "NG"
SUMMARY_FILE_DEL_OK_KEY = "FDELOK"
SUMMARY_FILE_DEL_NG_KEY = "FDELNG"
SUMMARY_DIR_OK_KEY = "DIROK"
SUMMARY_DIR_NG_KEY = "DIRNG"
SUMMARY_DIR_DEL_OK_KEY = "DDELOK"
SUMMARY_DIR_DEL_NG_KEY = "DDELNG"

TO_REVERSE_KEY_MAP = {
        SUMMARY_OK_KEY: SUMMARY_NG_KEY,
        SUMMARY_NG_KEY: SUMMARY_OK_KEY,
        SUMMARY_FILE_DEL_OK_KEY: SUMMARY_FILE_DEL_NG_KEY,
        SUMMARY_FILE_DEL_NG_KEY: SUMMARY_FILE_DEL_OK_KEY,
        SUMMARY_DIR_OK_KEY: SUMMARY_DIR_NG_KEY,
        SUMMARY_DIR_NG_KEY: SUMMARY_DIR_OK_KEY,
        SUMMARY_DIR_DEL_OK_KEY: SUMMARY_DIR_DEL_NG_KEY,
        SUMMARY_DIR_DEL_NG_KEY: SUMMARY_DIR_DEL_OK_KEY,
    }

FILE_UPLOAD_OK_TMPL = "%(" + SUMMARY_OK_KEY + ")d"
FILE_UPLOAD_NG_TMPL = "%(" + SUMMARY_NG_KEY + ")d"
DIR_UPLOAD_OK_TMPL = "%(" + SUMMARY_DIR_OK_KEY + ")d"
DIR_UPLOAD_NG_TMPL = "%(" + SUMMARY_DIR_NG_KEY + ")d"

FILE_DOWNLOAD_OK_TMPL = "%(" + SUMMARY_OK_KEY + ")d"
FILE_DOWNLOAD_NG_TMPL = "%(" + SUMMARY_NG_KEY + ")d"
DIR_DOWNLOAD_OK_TMPL = "%(" + SUMMARY_DIR_OK_KEY + ")d"
DIR_DOWNLOAD_NG_TMPL = "%(" + SUMMARY_DIR_NG_KEY + ")d"

FILE_DEL_OK_TMPL = "%(" + SUMMARY_FILE_DEL_OK_KEY + ")d"
FILE_DEL_NG_TMPL = "%(" + SUMMARY_FILE_DEL_NG_KEY + ")d"

DIR_DEL_OK_TMPL = "%(" + SUMMARY_DIR_DEL_OK_KEY + ")d"
DIR_DEL_NG_TMPL = "%(" + SUMMARY_DIR_DEL_NG_KEY + ")d"

TOTAL_FILES_TMPL = "%(" + TOTAL_FILES + ")d"
TOTAL_DIRS_TMPL = "%(" + TOTAL_DIRS + ")d"

SUMMARY_HEADER_TMPL = u"""============================
    クラウド同期結果
============================
結果: %s
共有名: %s
開始日時: %s
終了日時: %s
経過時間: %s
""" % (RESULT_TMPL, SHARE_TMPL, START_TMPL, END_TMPL, ELAPSE_TMPL)

SUMMARY_UPLOAD_BODY_TMPL = u"""
---------------------------------------------------------
ファイル:
            アップロード成功 : %s
            アップロード失敗 : %s
  クラウドのファイル削除成功 : %s
  クラウドのファイル削除失敗 : %s
                        合計 : %s

フォルダー:
  クラウドのフォルダー作成成功 : %s
  クラウドのフォルダー作成失敗 : %s
  クラウドのフォルダー削除成功 : %s
  クラウドのフォルダー削除失敗 : %s
                        合計   : %s
""" % (FILE_UPLOAD_OK_TMPL,
       FILE_UPLOAD_NG_TMPL,
       FILE_DEL_OK_TMPL,
       FILE_DEL_NG_TMPL,
       TOTAL_FILES_TMPL,
       DIR_UPLOAD_OK_TMPL,
       DIR_UPLOAD_NG_TMPL,
       DIR_DEL_OK_TMPL,
       DIR_DEL_NG_TMPL,
       TOTAL_DIRS_TMPL)

SUMMARY_DOWNLOAD_BODY_TMPL = u"""
---------------------------------------------------------
ファイル:
            ダウンロード成功 : %s
            ダウンロード失敗 : %s
       NASのファイル削除成功 : %s
       NASのファイル削除失敗 : %s
                        合計 : %s

フォルダー:
     NASのフォルダー作成成功 : %s
     NASのフォルダー作成失敗 : %s
     NASのフォルダー削除成功 : %s
     NASのフォルダー削除失敗 : %s
                      合計   : %s
""" % (FILE_DOWNLOAD_OK_TMPL,
       FILE_DOWNLOAD_NG_TMPL,
       FILE_DEL_OK_TMPL,
       FILE_DEL_NG_TMPL,
       TOTAL_FILES_TMPL,
       DIR_DOWNLOAD_OK_TMPL,
       DIR_DOWNLOAD_NG_TMPL,
       DIR_DEL_OK_TMPL,
       DIR_DEL_NG_TMPL,
       TOTAL_DIRS_TMPL)

SUMMARY_UPLOAD_NG_FILES = u"[アップロードに失敗したファイル一覧]\r\n"
SUMMARY_UPLOAD_NG_DIRS = u"[アップロードに失敗したフォルダ一覧]\r\n"
SUMMARY_DOWNLOAD_NG_FILES = u"[ダウンロードに失敗したファイル一覧]\r\n"
SUMMARY_DOWNLOAD_NG_DIRS = u"[ダウンロードに失敗したフォルダ一覧]\r\n"
SUMMARY_DEL_NG_FILES = u"[削除に失敗したファイル一覧]\r\n"
SUMMARY_DEL_NG_DIRS = u"[削除に失敗したフォルダ一覧]\r\n"



ERR_MARK = "+++ Error +++"
ERR_TEXT_FMT = "%s " + ERR_MARK + "\r\n%s"
LOG_FMT_STR = "%s -> %s\r\n"

SUCCESS = "success"
FAILED = "failed"
STOPPED = "stopped"

MAX_FAIL_FILE_COUNT = 1000
LIMIT_MEM_LOG_LENGTH = 10 * 1024 * 1024  # 10Mbyte

UPLOAD_COUNT = "upload"
DOWNLOAD_COUNT = "download"

LOG_SUFFIX = ".log"
FAIL_LOG_SUFFIX = "_failed.log"
STOP_LOG_SUFFIX = "_stopped.log"
SUMMARY_LOG_SUFFIX = "_summary.log"
MAX_LOG_COUNT = 20

_log_map = {}


class CreateLogFileError(IoBaseException):
    pass


class Summary(object):

    def __init__(self, share_name, remote_filetree_class):
        self.__share_name = share_name
        self.__share_map = {}
        self.__ng_files_map = {}
        self.__ng_dirs_map = {}
        self.__ng_del_files_map = {}
        self.__ng_del_dirs_map = {}

        self.__share_map[self.__share_name] = \
            {UPLOAD_COUNT: {SUMMARY_OK_KEY: 0,
                            SUMMARY_NG_KEY: 0,
                            SUMMARY_FILE_DEL_OK_KEY: 0,
                            SUMMARY_FILE_DEL_NG_KEY: 0,
                            SUMMARY_DIR_OK_KEY: 0,
                            SUMMARY_DIR_NG_KEY: 0,
                            SUMMARY_DIR_DEL_OK_KEY: 0,
                            SUMMARY_DIR_DEL_NG_KEY: 0,
                            },
             DOWNLOAD_COUNT: {SUMMARY_OK_KEY: 0,
                              SUMMARY_NG_KEY: 0,
                              SUMMARY_FILE_DEL_OK_KEY: 0,
                              SUMMARY_FILE_DEL_NG_KEY: 0,
                              SUMMARY_DIR_OK_KEY: 0,
                              SUMMARY_DIR_NG_KEY: 0,
                              SUMMARY_DIR_DEL_OK_KEY: 0,
                              SUMMARY_DIR_DEL_NG_KEY: 0, }}

        self.__ng_files_map[self.__share_name] = {UPLOAD_COUNT: [],
                                                  DOWNLOAD_COUNT: []}
        self.__ng_dirs_map[self.__share_name] = {UPLOAD_COUNT: [],
                                                 DOWNLOAD_COUNT: []}
        self.__ng_del_files_map[self.__share_name] = {UPLOAD_COUNT: [],
                                                      DOWNLOAD_COUNT: []}
        self.__ng_del_dirs_map[self.__share_name] = {UPLOAD_COUNT: [],
                                                     DOWNLOAD_COUNT: []}

        try:
            self.default_encode = locale.getdefaultlocale()[1]
            if not self.default_encode:
                self.default_encode = DEFAULT_LOCALE
        except:
            self.default_encode = DEFAULT_LOCALE

        self.__remote_filetree_class = remote_filetree_class

    def init_summay_param(self):

        self.__share_map = {}
        self.__ng_files_map = {}
        self.__share_map[self.__share_name] = \
            {UPLOAD_COUNT: {SUMMARY_OK_KEY: 0,
                            SUMMARY_NG_KEY: 0,
                            SUMMARY_FILE_DEL_OK_KEY: 0,
                            SUMMARY_FILE_DEL_NG_KEY: 0,
                            SUMMARY_DIR_OK_KEY: 0,
                            SUMMARY_DIR_NG_KEY: 0,
                            SUMMARY_DIR_DEL_OK_KEY: 0,
                            SUMMARY_DIR_DEL_NG_KEY: 0,
                            },
             DOWNLOAD_COUNT: {SUMMARY_OK_KEY: 0,
                              SUMMARY_NG_KEY: 0,
                              SUMMARY_FILE_DEL_OK_KEY: 0,
                              SUMMARY_FILE_DEL_NG_KEY: 0,
                              SUMMARY_DIR_OK_KEY: 0,
                              SUMMARY_DIR_NG_KEY: 0,
                              SUMMARY_DIR_DEL_OK_KEY: 0,
                              SUMMARY_DIR_DEL_NG_KEY: 0, }}

        self.__ng_files_map[self.__share_name] = {UPLOAD_COUNT: [],
                                                  DOWNLOAD_COUNT: []}
        self.__ng_dirs_map[self.__share_name] = {UPLOAD_COUNT: [],
                                                 DOWNLOAD_COUNT: []}
        self.__ng_del_files_map[self.__share_name] = {UPLOAD_COUNT: [],
                                                      DOWNLOAD_COUNT: []}
        self.__ng_del_dirs_map[self.__share_name] = {UPLOAD_COUNT: [],
                                                     DOWNLOAD_COUNT: []}

    def set_sync_start_time(self, start_daytime):
        self.__start_daytime = start_daytime

    def set_sync_end_time(self, end_daytime):
        self.__end_daytime = end_daytime

    def __get_count_target_key(self, log_info):

        try:
            # ファイル作成
            if log_info.type == TYPE_FILE_TRANSFER_OK:
                return SUMMARY_OK_KEY

            if log_info.type == TYPE_FILE_TRANSFER_NG or \
               log_info.exc_type == SizeTooLargeError or \
               log_info.exc_type == IgnoreTargetFile:
                return SUMMARY_NG_KEY

            # ファイル削除
            if log_info.log.startswith(FILE_DELETE_LOG + " " + OK_LOG):
                return SUMMARY_FILE_DEL_OK_KEY

            if log_info.log.startswith(FILE_DELETE_LOG + " " + NG_LOG):
                return SUMMARY_FILE_DEL_NG_KEY

            # ディレクトリ作成
            if log_info.log.startswith(DIR_CREATE_LOG + " " + OK_LOG):
                return SUMMARY_DIR_OK_KEY

            if log_info.log.startswith(DIR_CREATE_LOG + " " + NG_LOG):
                return SUMMARY_DIR_NG_KEY

            # ディレクトリ削除
            if log_info.log.startswith(DIR_DELETE_LOG + " " + OK_LOG):
                return SUMMARY_DIR_DEL_OK_KEY

            if log_info.log.startswith(DIR_DELETE_LOG + " " + NG_LOG):
                return SUMMARY_DIR_DEL_NG_KEY
        except:
            pass

        return None

    def __get_ng_list(self, tgt_key, share_name, direction):

        if tgt_key == SUMMARY_NG_KEY:
            return self.__ng_files_map[share_name][direction]

        elif tgt_key == SUMMARY_FILE_DEL_NG_KEY:
            return self.__ng_del_files_map[share_name][direction]

        elif tgt_key == SUMMARY_DIR_NG_KEY:
            return self.__ng_dirs_map[share_name][direction]

        elif tgt_key == SUMMARY_DIR_DEL_NG_KEY:
            return self.__ng_del_dirs_map[share_name][direction]

        return None

    def __get_ng_list_from_ok_key(self, tgt_key, share_name, direction):

        if tgt_key == SUMMARY_OK_KEY:
            return self.__ng_files_map[share_name][direction]

        elif tgt_key == SUMMARY_FILE_DEL_OK_KEY:
            return self.__ng_del_files_map[share_name][direction]

        elif tgt_key == SUMMARY_DIR_OK_KEY:
            return self.__ng_dirs_map[share_name][direction]

        elif tgt_key == SUMMARY_DIR_DEL_OK_KEY:
            return self.__ng_del_dirs_map[share_name][direction]

        return None

    def update(self, log_info):

        #if isinstance(log_info.dst, ServerFileTree):
        if isinstance(log_info.dst, self.__remote_filetree_class):
            direction = UPLOAD_COUNT
        else:
            direction = DOWNLOAD_COUNT

        tgt_key = self.__get_count_target_key(log_info)
        if not tgt_key:
            return

        file_path = log_info.relpath.replace("/", "\\")

        ng_list = self.__get_ng_list(tgt_key, self.__share_name, direction)

        if ng_list is not None:
            if file_path not in ng_list:
                if len(ng_list) < MAX_FAIL_FILE_COUNT:
                    ng_list.append(file_path)

                self.__share_map[self.__share_name][direction][tgt_key] += 1

        else:
            self.__share_map[self.__share_name][direction][tgt_key] += 1

            ng_list = self.__get_ng_list_from_ok_key(
                          tgt_key, self.__share_name, direction)

            if ng_list is None:
                return

            # リライで成功した場合
            if file_path in ng_list:
                ng_list.remove(file_path)

                self.__share_map[self.__share_name][
                    direction][TO_REVERSE_KEY_MAP[tgt_key]] -= 1

    def get_summary_body(self, is_download=False):

        share = self.__share_name

        NG_FILES_IDX = 0
        NG_DIRS_IDX = 1
        NG_DEL_FILES_IDX = 2
        NG_DEL_DIRS_IDX = 3

        if is_download:
            count_type = DOWNLOAD_COUNT
            TITLE_LST = [SUMMARY_DOWNLOAD_NG_FILES,
                         SUMMARY_DOWNLOAD_NG_DIRS,
                         SUMMARY_DEL_NG_FILES,
                         SUMMARY_DEL_NG_DIRS]
        else:
            count_type = UPLOAD_COUNT
            TITLE_LST = [SUMMARY_UPLOAD_NG_FILES,
                         SUMMARY_UPLOAD_NG_DIRS,
                         SUMMARY_DEL_NG_FILES,
                         SUMMARY_DEL_NG_DIRS]

        cnt_map = self.__share_map[share][count_type]

        body_map = cnt_map

        total_files = cnt_map[SUMMARY_OK_KEY] + \
                      cnt_map[SUMMARY_NG_KEY] + \
                      cnt_map[SUMMARY_FILE_DEL_OK_KEY] + \
                      cnt_map[SUMMARY_FILE_DEL_NG_KEY]

        total_dirs = cnt_map[SUMMARY_DIR_OK_KEY] + \
                     cnt_map[SUMMARY_DIR_NG_KEY] + \
                     cnt_map[SUMMARY_DIR_DEL_OK_KEY] + \
                     cnt_map[SUMMARY_DIR_DEL_NG_KEY]

        body_map.update({TOTAL_FILES: total_files,
                         TOTAL_DIRS: total_dirs})

        if is_download:
            body_str = SUMMARY_DOWNLOAD_BODY_TMPL % body_map
        else:
            body_str = SUMMARY_UPLOAD_BODY_TMPL % body_map

        if len(self.__ng_files_map[share][count_type]) > 0:
            body_str += TITLE_LST[NG_FILES_IDX]
            body_str += \
                unicode("\r\n".join(self.__ng_files_map[share][count_type]),
                        self.default_encode)
            body_str += u"\r\n\r\n"

        if len(self.__ng_dirs_map[share][count_type]) > 0:
            body_str += TITLE_LST[NG_DIRS_IDX]
            body_str += \
                unicode("\r\n".join(self.__ng_dirs_map[share][count_type]),
                        self.default_encode)
            body_str += u"\r\n\r\n"

        if len(self.__ng_del_files_map[share][count_type]) > 0:
            body_str += TITLE_LST[NG_DEL_FILES_IDX]
            body_str += \
                unicode("\r\n".join(
                            self.__ng_del_files_map[share][count_type]),
                        self.default_encode)
            body_str += u"\r\n\r\n"

        if len(self.__ng_del_dirs_map[share][count_type]) > 0:
            body_str += TITLE_LST[NG_DEL_DIRS_IDX]
            body_str += \
                unicode("\r\n".join(self.__ng_del_dirs_map[share][count_type]),
                        self.default_encode)
            body_str += u"\r\n\r\n"

        body_str = body_str.replace("\r\n", "\n").replace("\n", "\r\n")
        return body_str.encode(self.default_encode)

    def get_summary_header(self, result):

        delta_time = self.__end_daytime - self.__start_daytime
        elapse_seconds = delta_time.days * 24 * 3600 + delta_time.seconds
        elapse_time_str = "%02d:%02d:%02d" % \
                              (elapse_seconds / 3600,
                               (elapse_seconds % 3600) / 60,
                               elapse_seconds % 60)
        start_time = self.__start_daytime.strftime('%Y-%m-%d %H:%M:%S')
        end_time = self.__end_daytime.strftime('%Y-%m-%d %H:%M:%S')

        header = {RESULT: result.decode(self.default_encode),
                  SHARE: self.__share_name.decode(
                                       self.default_encode),
                  START: start_time.decode(self.default_encode),
                  END: end_time.decode(self.default_encode),
                  ELAPSE: elapse_time_str.decode(
                                        self.default_encode)}

        header_str = SUMMARY_HEADER_TMPL % header
        header_str = header_str.replace("\r\n", "\n").replace("\n", "\r\n")
        return header_str.encode(self.default_encode)

    def get_file_access_count(self):

        return \
            sum(self.__share_map[self.__share_name][UPLOAD_COUNT].values()) + \
            sum(self.__share_map[self.__share_name][DOWNLOAD_COUNT].values())

    def get_ng_file_count(self):

        cnt_map = self.__share_map[self.__share_name]
        return cnt_map[UPLOAD_COUNT][SUMMARY_NG_KEY] + \
               cnt_map[DOWNLOAD_COUNT][SUMMARY_NG_KEY]


class Logger():

    def __init__(self,
                 share_name,
                 log_file_path,
                 log_accesser,
                 remote_filetree_class,
                 buf_limit_len=LIMIT_MEM_LOG_LENGTH):

        self.__summary = Summary(share_name, remote_filetree_class)
        self.__result = ""
        self.__share_name = share_name
        self.__is_need_save = False

        def _mkdir(path):
            log_accesser.mkdir(path)
            log_accesser.chmod(path, MODE_CH_DIR)

        self._log_open = lambda path, mode:\
            log_accesser.open(path, mode)
        self._log_mkdir = lambda path:\
            _mkdir(path)
        self._log_mv = lambda orig, dst:\
            log_accesser.move(orig, dst)
        self._log_chmod = lambda path, mode:\
            log_accesser.chmod(path, mode)
        self._log_chmod = lambda path, mode:\
            log_accesser.chmod(path, mode)
        self._log_listdir = lambda path:\
            log_accesser.listdir(path)
        self._log_rmdir = lambda path:\
            log_accesser.rmdir(path)
        self._log_stat = lambda path:\
            log_accesser.stat(path)
        self._log_remove = lambda path:\
            log_accesser.remove(path)

        self._log_io = None
        self._log_file_path = log_file_path + LOG_SUFFIX
        self._summary_log_file_path = log_file_path + SUMMARY_LOG_SUFFIX
        self._buf_limit_len = buf_limit_len
        self.make_log_dir()

    def init_log_param(self, log_file_path):

        self.__is_need_save = False
        self._log_file_path = log_file_path + LOG_SUFFIX
        self._summary_log_file_path = log_file_path + SUMMARY_LOG_SUFFIX
        self.__summary.init_summay_param()

    def get_ng_file_count(self):

        try:
            return self.__summary.get_ng_file_count()
        except:
            return 0

    def __write(self, string):

        if not self._log_io:
            self._log_io = StringIO()

        self._log_io.write(string)

        if self._log_io.len > self._buf_limit_len:
            self._flush_and_renew_log_buf()

    def make_log_dir(self):

        sep = "/"
        dir_path = os.path.dirname(self._log_file_path)
        dir_parts = dir_path.split(sep)

        current_path = sep
        for part in dir_parts:
            current_path += sep + part

            try:
                self._log_stat(current_path)
            except NoEntryError:
                try:
                    self._log_mkdir(current_path)
                except ExistsError:
                    pass

    def __flush(self):

        log_file = \
            self._log_open(self._log_file_path, MODE_APPEND)

        self._log_io.seek(0)
        log_file.write(self._log_io)
        log_file.close()

    def _renew_log_buf(self):

        if self._log_io:
            self._log_io.close()
            del self._log_io

        self._log_io = StringIO()

    def _flush_and_renew_log_buf(self):
        """
           Flush log data into log file.
           And free buf and realloc new buf to avoid huge memory consumption.
        """

        self.__is_need_save = False

        try:
            self.__flush()
        except:
            raise CreateLogFileError

        self._log_chmod(self._log_file_path, MODE_CH_FILE)
        self._renew_log_buf()

    def end_log(self, result):

        log_suffix_map = {SUCCESS: LOG_SUFFIX,
                          FAILED: FAIL_LOG_SUFFIX,
                          STOPPED: STOP_LOG_SUFFIX}

        self.__result = result
        suffix = log_suffix_map[result]

        if not self.__is_need_save:
            self._renew_log_buf()
            return

        self._flush_and_renew_log_buf()

        if result in log_suffix_map.keys():
            real_log_path = \
                re.sub(LOG_SUFFIX + "$", suffix, self._log_file_path)
            self._log_mv(self._log_file_path, real_log_path)

        self._save_summary()

        self._delete_oldest_log()

    def set_share(self, share):
        self.__summary.set_share(share)

    def set_sync_start_time(self, start_daytime):
        self.__summary.set_sync_start_time(start_daytime)

    def set_sync_end_time(self, end_daytime):
        self.__summary.set_sync_end_time(end_daytime)

    def __get_detail_info_str(self, log_info):

        DETAIL_STR_TMPL = \
            "HTTPError detail info:\r\n" + \
            "  code : %(code)s\r\n" + \
            "  msg : %(msg)s\r\n" + \
            "  data : %(data)s\r\n"

        try:
            if hasattr(log_info.exc_arg, "detail"):
                err = log_info.exc_arg.detail
                if isinstance(err, urllib2.HTTPError):

                    http_data = ""
                    if hasattr(err, "read"):
                        http_data = err.read()

                    return DETAIL_STR_TMPL % {"code": str(err.code),
                                              "msg": str(err.msg),
                                              "data": http_data}

            if isinstance(log_info.exc_arg, urllib2.HTTPError):
                err = log_info.exc_arg

                http_data = ""
                if hasattr(err, "read"):
                    http_data = err.read()

                return DETAIL_STR_TMPL % {"code": str(err.code),
                                          "msg": str(err.msg),
                                          "data": http_data}
        except:
            pass

        return  ""

    def write_log(self, log_info):
        result = log_info.log
        if result and log_info.type != TYPE_SYNC_SKIP:
            if log_info.lv >= SKIP_ERROR_LOG_LV:
                trace_info = log_info.trace
                if str(trace_info).strip() == "":
                    trace_info = ""
                result = ERR_TEXT_FMT % (result, trace_info)

            src = (self.__share_name + '\\' + \
                      log_info.relpath).replace("/", "\\")

            self.__is_need_save = True
            self.__write(lf2crlf(LOG_FMT_STR % (src, result)))

            if log_info.lv >= SKIP_ERROR_LOG_LV:
                detail_str = self.__get_detail_info_str(log_info)
                if detail_str:
                    self.__write(detail_str + "\r\n")

        self.__summary.update(log_info)

    def write(self, data, is_need_save=False):

        self.__write(data)
        if is_need_save:
            self.__is_need_save = True

    def _save_summary(self):

        try:
            self._log_mkdir(os.path.dirname(self._summary_log_file_path))
        except ExistsError:
            pass

        log_file = self._log_open(self._summary_log_file_path, MODE_WRITE)
        log_file.write(StringIO(
            self.__summary.get_summary_header(self.__result)))
        log_file.write(StringIO(self.__summary.get_summary_body(False)))
        log_file.write(StringIO(self.__summary.get_summary_body(True)))
        log_file.close()
        return

    def _delete_oldest_log(self):

        log_dir = os.path.dirname(self._log_file_path)

        try:
            log_list = [log for log in self._log_listdir(log_dir)
                            if not log.endswith(FAIL_LOG_SUFFIX)]
        except:
            return

        for log in sorted(log_list, reverse=True)[MAX_LOG_COUNT:]:
            try:
                self._log_remove(os.path.join(log_dir, log))
            except:
                pass

    def get_file_access_count(self):

        return self.__summary.get_file_access_count()


class InternalLog():
    """Class for the internal log."""

    LOG_FORMAT = "%(asctime).19s: %(filename)s: %(levelname)-8s: %(message)s"
    LOG_MAX_SIZE = 256 * 1024
    LOG_COUNT = 9

    def _add_info_to_message(self, message):
        """Add caller information to message."""

        stack = inspect.stack()
        message += " : %s(%d)" % (os.path.basename(stack[2][1]), stack[2][2])
        del stack
        return message

    def __init__(self, logger_name, log, log_level=None):
        """Constructor."""

        self.logger = None
        if not logger_name or not log:
            return

        if log_level:
            log_level = logging.WARNING

        if not logger_name in _log_map:
            _log_map[logger_name] = []

        self.logger = logging.getLogger(logger_name)
        if log not in _log_map[logger_name]:
            _log_map[logger_name].append(log)
            handler = RotatingFileHandler(log, "a", self.LOG_MAX_SIZE,
                                            self.LOG_COUNT)
            handler.setFormatter(logging.Formatter(self.LOG_FORMAT))
            handler.setLevel(log_level)
            self.logger.addHandler(handler)

    def debug(self, message, *args):
        """Log "message" for level DEBUG."""

        if self.logger != None:
            self.logger.debug(self._add_info_to_message(message), *args)

    def info(self, message, *args):
        """Log "message" for level INFO."""

        if self.logger != None:
            self.logger.info(self._add_info_to_message(message), *args)

    def warning(self, message, *args):
        """Log "message" for level WARNING."""

        if self.logger != None:
            self.logger.warning(self._add_info_to_message(message), *args)

    def error(self, message, *args):
        """Log "message" for level ERROR."""

        if self.logger != None:
            self.logger.error(self._add_info_to_message(message), *args)

    def critical(self, message, *args):
        """Log "message" for level CRITICAL."""

        if self.logger != None:
            self.logger.critical(self._add_info_to_message(message), *args)

    def exception(self, message, *args):
        """Log "message" for exception."""

        if self.logger != None:
            self.logger.error(self._add_info_to_message(message),
                                exc_info=True, *args)

    def debug_exception(self, message, *args):
        """Log "message" for level DEBUG on exception."""

        if self.logger != None:
            self.logger.debug(self._add_info_to_message(message),
                                exc_info=True, *args)
