#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""Module for accessing to the OneDrive server."""

import os
import sys
import stat
import locale
import errno
import re
import time

from iobase.exception import (InvalidArgError)
from ionas.exception import (NoEntryError, ExistsError,
                             InvalidFileSystemError, NoSpaceError,
                             SizeTooLargeError,
                             InternalError, NameTooLongError,
                             PermissionError,)
from ionas.fileaccess import (MODE_READ, MODE_WRITE, DEFAULT_LOCALE)
from ionas.fileaccess.filelikeobj import FileLikeObj

from pkgaddon.onedrive_config import (APP_ID, CLIENT_SECRET, REDIRECT_URL,
                                        BUSINESS_APP_ID,
                                        BUSINESS_CLIENT_SECRET,
                                        BUSINESS_REDIRECT_URL,
                                        BUSINESS_AUTH_SERVER_URL,
                                        BUSINESS_AUTH_TOKEN_URL,
                                        BUSINESS_SERVICE_NAME,
                                        SCOPES,)
from cloudsync.util import (get_proxies_dict,)
from onedrive.onedriveaccess import (FILE_SEPARATOR,)
import onedrive.onedriveaccess as onedrive



RETRY_MAX = 1  # 現行はリトライしない
RETRY_PERIOD = 5

PROTOCOL_LIST = ("http", "https")

BUF_SIZE = 64 * 1024
CHUNK_SIZE = 64 * 1024 * 1024


##
# コンテナオブジェクト
#
# データを保持するだけのオブジェクト．何もしない．
#
class Container:
    pass

PRIORITY_NONE = "none"
PRIORITY_FILE = "file"
PRIORITY_DIR = "dir"

LOCAL_CANCEL_ERRS = (errno.EACCES, errno.EPERM,
                     errno.ENOENT, errno.EEXIST,
                     errno.ENAMETOOLONG)


def default_except_handler(err):
    raise err


class OneDriveFileObj(FileLikeObj):
    """Class for a file on OneDrive server."""

    def __init__(self, path, mode, access):
        self.mode = mode
        self.access = access
        self.fileobj = None

        if self.mode == MODE_WRITE:
            self.filesize = 0
        elif self.mode == MODE_READ:
            self.filesize = -1  # < 0 で未取得を表す
        else:
            raise InvalidArgError("invalid access mode %s" % str(mode))

        FileLikeObj.__init__(self, None, path)

    def __len__(self):
        if self.filesize >= 0:
            return self.filesize

        err = None
        for r in range(0, RETRY_MAX):
            try:
                self.filesize = self.access.stat(self.path)[stat.ST_SIZE]
                return self.filesize
            except:
                cl, err = sys.exc_info()[:2]
                self.access.except_handler(err)
        else:
            self.access.except_handler(err)

    def tell(self):
        raise NotImplementedError

    def seek(self, offset, whence=os.SEEK_SET):
        raise NotImplementedError

    def get_path(self):

        raise NotImplementedError

    def read(self, size=None):

        err = None
        for r in range(0, RETRY_MAX):
            try:
                if self.fileobj == None:
                    self.fileobj = self.access.raw_open(self.path, self.mode)
                break
            except:
                cl, err = sys.exc_info()[:2]
                self.access.except_handler(err, self.path)

        if size == None:
            size = -1

        for r in range(0, RETRY_MAX):
            try:
                return self.fileobj.read(size)
            except:
                cl, err = sys.exc_info()[:2]
                self.access.except_handler(err, self.path)

        self.access.except_handler(err, self.path)

    def __get_file_size(self, file_like):

        try:
            return len(file_like)
        except:
            # Ignore error
            pass

        try:
            return file_like.get_size()
        except:
            # Ignore error
            pass

        try:
            return os.fstat(file_like.fileno()).st_size
        except:
            # Ignore error
            pass

        return -1

    def __update_file_size(self, file_like):

        self.filesize = self.__get_file_size(file_like)

    def __write(self, file_like):

        try:
            target_file = self.access.raw_open(self.path, self.mode,
                                            self.__get_file_size(file_like))
        except:
            cl, err = sys.exc_info()[:2]
            self.access.except_handler(err, self.path)

        try:
           while True:
               data = file_like.read(BUF_SIZE)
               if len(data) <= 0:
                   break
               target_file.write(data)

           target_file.close()
        except:
            cl, err = sys.exc_info()[:2]
            self.access.except_handler(err, self.path)

    def __write_by_path_with_mkdir(self, dir_path, file_like):

        try:
            self.access._mkdir_recursive(dir_path)
            self.__write(file_like)
        except:
            cl, err = sys.exc_info()[:2]
            self.access.except_handler(err, self.path)
        else:
            self.__update_file_size(file_like)
        return

    def write_by_path(self, file_like):

        path_dir, file_name = \
            self.path.rstrip(self.access.sep).rsplit(self.access.sep, 1)

        err = None
        for r in range(0, RETRY_MAX):
            try:
                try:
                    self.__write(file_like)
                except (InvalidFileSystemError, NoSpaceError,
                        SizeTooLargeError, InternalError, NameTooLongError,
                        PermissionError):
                    raise
                except:
                    time.sleep(RETRY_PERIOD)
                    if hasattr(file_like, "seek"):
                        file_like.seek(0)
                    self.__write(file_like)
            except:
                cl, err = sys.exc_info()[:2]
                try:
                    self.access.except_handler(err, self.path)
                except (NoEntryError, ExistsError):
                    if self.access.isdir(self.path, is_need_conv_path=False):
                        self.access.remove(self.path, maybe_dir=True,
                                           is_need_conv_path=False)
                    file_like.seek(0)
                    self.__write_by_path_with_mkdir(path_dir, file_like)
                    return
            else:
                self.__update_file_size(file_like)
                return

        self.access.except_handler(err, self.path)

    def close(self):
        pass


class OneDriveAccess():
    """Class for accessing the OneDrive server."""

    def _pack_multi_separator(self, path):

        return re.sub("[%s]+" % (self.sep,), self.sep, path)

    def _get_server_path(self, path):

        try:
            _path = path.decode(self.encoding)
            return self.onedrive_access.convert_name(_path)
        except:
            cl, err = sys.exc_info()[:2]
            self.except_handler(err)

    def __init__(self,
                 session_path,
                 for_business,
                 encoding=None,
                 proxy_host=None,
                 proxy_port=None,
                 except_handler=None):

        self.session_path = session_path
        self.for_business = for_business
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self.sep = str(FILE_SEPARATOR) # パス区切り文字

        self.invalid_list = []

        if except_handler:
            self.except_handler = except_handler
        else:
            self.except_handler = default_except_handler

        if encoding:
            self.encoding = encoding
        else:
            self.encoding = locale.getdefaultlocale()[1]
            if not self.encoding:
                self.encoding = DEFAULT_LOCALE

        self.file_client = None
        proxies_dict = get_proxies_dict(proxy_host, proxy_port,
                                            PROTOCOL_LIST)

        if self.for_business:
            self.onedrive_access = onedrive.OneDriveAccess(BUSINESS_APP_ID,
                                    BUSINESS_CLIENT_SECRET,
                                    scopes=SCOPES,
                                    redirect_url=BUSINESS_REDIRECT_URL,
                                    session_path=self.session_path,
                                    auth_server_url=BUSINESS_AUTH_SERVER_URL,
                                    auth_token_url=BUSINESS_AUTH_TOKEN_URL,
                                    service_name=BUSINESS_SERVICE_NAME,
                                    proxies=proxies_dict)
        else:
            self.onedrive_access = onedrive.OneDriveAccess(APP_ID,
                                    CLIENT_SECRET,
                                    scopes=SCOPES,
                                    redirect_url=REDIRECT_URL,
                                    session_path=self.session_path,
                                    proxies=proxies_dict)

        self.onedrive_access.authenticate()

    def get_invalid_list(self):

        return self.invalid_list

    def append_invalid_path(self, path, rel_path):

        invalid_path = \
                path.lstrip(self.sep).encode(self.encoding)[len(rel_path):]
        if invalid_path not in self.invalid_list:
            self.invalid_list.append(invalid_path)

    def remove_invalid_path(self, path, rel_path):

        try:
            invalid_path = \
                path.lstrip(self.sep).encode(self.encoding)[len(rel_path):]
            if invalid_path in self.invalid_list:
                self.invalid_list.remove(invalid_path)
        except:
            # Ignore error
            pass

    def listdir(self, path):

        _path = self.__conv_to_internal_path(path)

        try:
            return self.onedrive_access.get_contents(_path).keys()
        except:
            cl, err = sys.exc_info()[:2]
            self.except_handler(err, path)

    def isdir(self, path, is_need_conv_path=True):

        path = self._pack_multi_separator(path)
        if is_need_conv_path:
            _path = self._get_server_path(path)
        else:
            _path = path

        try:
            stat_info = self.stat(_path, is_need_conv_path=False)
            return stat.S_ISDIR(stat_info[stat.ST_MODE])
        except:
            return False

    def isfile(self, path, is_need_conv_path=True):

        path = self._pack_multi_separator(path)
        if is_need_conv_path:
            _path = self._get_server_path(path)
        else:
            _path = path

        try:
            stat_info = self.stat(_path, is_need_conv_path=False)
            return stat.S_ISREG(stat_info[stat.ST_MODE])
        except:
            return False

    def open(self, path, mode):

        path = self._pack_multi_separator(path)
        return OneDriveFileObj(self._get_server_path(path), mode, self)

    def raw_open(self, path, mode, length=-1):

        path = self._pack_multi_separator(path)
        raw_mode = "r"
        if mode == MODE_WRITE:
            raw_mode = "w"
        return self.onedrive_access.open(path, raw_mode, length, CHUNK_SIZE)

    def stat(self, path, maybe_dir=False, is_need_conv_path=True):

        path = self._pack_multi_separator(path)
        if is_need_conv_path:
            _path = self._get_server_path(path)
        else:
            _path = path

        stat_info = {}
        err = None
        for r in range(0, RETRY_MAX):
            try:
                if maybe_dir:
                    try:
                        stat_info = self.onedrive_access.stat_directory(_path)
                    except:
                        stat_info = self.onedrive_access.stat_file(_path)
                else:
                    try:
                        stat_info = self.onedrive_access.stat_file(_path)
                    except:
                        stat_info = self.onedrive_access.stat_directory(_path)
            except:
                cl, err = sys.exc_info()[:2]
                self.except_handler(err, path)
            else:
                # 正常終了
                break
        else:
            # リトライ回数オーバー
            self.except_handler(err, path)

        return stat_info

    def remove(self, path, maybe_dir=False, is_need_conv_path=True):

        path = self._pack_multi_separator(path)
        if is_need_conv_path:
            _path = self._get_server_path(path)
        else:
            _path = path

        err = None
        for r in range(0, RETRY_MAX):
            try:
                if maybe_dir:
                    try:
                        self.onedrive_access.delete_directory(_path)
                    except:
                        self.onedrive_access.delete_file(_path)
                else:
                    try:
                        self.onedrive_access.delete_file(_path)
                    except:
                        self.onedrive_access.delete_directory(_path)

            except:
                cl, err = sys.exc_info()[:2]
                self.except_handler(err, path)
            else:
                # 正常終了
                break
        else:
            # リトライ回数オーバー
            self.except_handler(err, path)

    def rmdir(self, path):
        self.remove(path, maybe_dir=True)

    def __is_exist_dir(self, path):

        try:
            stat_info = self.file_client.stat_directory(path)
            if stat.ST_MODE not in stat_info:
                return False
            if stat_info[stat.ST_MODE] & stat.S_IFDIR:
                return True
            return False
                 
        except:
            cl, err = sys.exc_info()[:2]
            try:
                self.except_handler(err, path)
            except NoEntryError:
                return False
        return True

    def _mkdir_recursive(self, path):

        path_list = path.split(self.sep)

        # 上から順番に作る
        for idx in range(len(path_list) + 1):

            _path = self.sep.join(path_list[:idx])
            try:
                if self.sep == _path or not _path:
                    continue
                if not self.__is_exist_dir(_path):
                    self.onedrive_access.make_directory(_path)
            except:
                cl, err = sys.exc_info()[:2]
                try:
                    self.except_handler(err, path)
                except ExistsError:
                    continue

    def mkdir(self, path):

        path = self._pack_multi_separator(path)
        _path = self._get_server_path(path)

        err = None
        for r in range(0, RETRY_MAX):
            try:
                self.onedrive_access.make_directory(_path)
            except:
                cl, err = sys.exc_info()[:2]
                try:
                    self.except_handler(err, path)
                except (NoEntryError, ExistsError):
                    if self.isfile(_path, is_need_conv_path=False):
                        self.remove(_path, maybe_dir=False,
                                    is_need_conv_path=False)
                    self._mkdir_recursive(_path)
                    return
            else:
                # 正常終了
                break
        else:
            self.except_handler(err, path)

    def _make_dir_item(self, base_path, name, contents):
        c = Container()
        c.name = (base_path + name + self.sep).encode(self.encoding)
        c.last_modified = contents[stat.ST_MTIME]
        c.size = contents[stat.ST_SIZE]
        return c

    def _make_file_item(self, base_path, name, contents):
        c = Container()
        c.name = (base_path + name).encode(self.encoding)
        c.last_modified = contents[stat.ST_MTIME]
        c.size = contents[stat.ST_SIZE]
        return c

    def _get_list(self, path, rel_path):

        err = None
        contents_dict = {}
        for r in range(0, RETRY_MAX):
            try:
                try:
                    contents_dict = self.onedrive_access.get_contents(path,
                                                                        True)
                    self.remove_invalid_path(path, rel_path)
                except (InvalidFileSystemError, NoSpaceError,
                        SizeTooLargeError, InternalError, NameTooLongError,
                        PermissionError):
                    raise
                except LookupError:
                    self.append_invalid_path(path, rel_path)
                    contents_dict = {}

                except:
                    time.sleep(RETRY_PERIOD)
                    try:
                        contents_dict = self.onedrive_access.get_contents(path,
                                                                        True)
                        self.remove_invalid_path(path, rel_path)
                    except LookupError:
                        self.append_invalid_path(path, rel_path)
                        contents_dict = {}
            except:
                cl, err = sys.exc_info()[:2]
                self.except_handler(err, path)
            else:
                break
        else:
            # リトライ回数オーバー
            self.except_handler(err, path)

        for name in contents_dict:
            base_path = path + self.sep
            if path == self.sep:
                base_path = path
            if stat.S_ISDIR(contents_dict[name][stat.ST_MODE]):
                item = self._make_dir_item(base_path, name,
                                            contents_dict[name])
                yield item
                for item in self._get_list(base_path + name, rel_path):
                    yield item

            else:
                item = self._make_file_item(base_path, name,
                                                contents_dict[name])
                yield item

    def get_list(self, path):

        path = self._pack_multi_separator(path)
        rel_path = path

        _path = self._get_server_path(path)

        retlist = []

        for i in self._get_list(_path, rel_path):
            i.name = i.name[len(rel_path):]
            retlist.append(i)

        return retlist
