#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""
Module for file tree on local for syncing to the OneDrive server.

Copyright (C) I-O DATA DEVICE, INC.  All rights reserved.
"""

import locale
import time

from collections import defaultdict
from ionas.fileaccess import (DEFAULT_LOCALE)

from ionas.filesync import (LocalFileTree)

from ionas.filesync import (FILES_NAME, DIRS_NAME, CHILDREN_NAME,
                            STATE_NEW, STATE_NAME, STATE_DEL,)
from ionas.filesync.basefilesync import(SKIP_ERROR_LOG_LV,)
from ionas.exception import (BaseAccessException)

from onedrive.onedriveaccess import (OneDriveAccess,)


DUMMY_SESSION_PATH = "/vat/tmp/onedrive_dummy_session"
LOCAL_NAME = "local"
EQUIVALENT_SKIP_LOG = "Skip Equivarent File/Folder"
INVALID_NAME_FOLDER_SKIP_LOG = "Skip Invalid Name Folder"

NOTICE_LOG_LV = 10
OK_LOG = "OK"


class MultipleEquivarentExist(BaseAccessException):
    pass


class OneDriveLocalFileTree(LocalFileTree):

    def _is_invalid_path(self, target_path):

        if self.cloudsync_onedriveaccess == None:
            return False

        for invalid_path in self.cloudsync_onedriveaccess.get_invalid_list():
            if target_path == invalid_path:
                return True

            if invalid_path[-1] != "/":
                invalid_path += "/"

            if target_path.startswith(invalid_path):
                return True
        return False

    def _DstFull__try_operation(self, ope, relpath, file, file_info):

        from cloudsync.onedrivefiletree import (ForbiddenNameError,)

        is_executed = False
        try:
            target_path = self.sep.join(relpath + [file])
            if self._is_invalid_path(target_path):
                raise ForbiddenNameError("Invalid path", target_path)

            if ope[self.METHOD_NAME](relpath + [file], file_info):
                is_executed = True
                self.logging(NOTICE_LOG_LV,
                    ope[self.LOG_NAME] + " " + OK_LOG, "",
                    self.sep.join(relpath + [file]), self.src, self,
                    log_type=self._DstFull__get_log_type(ope[self.LOG_NAME],
                                                            True))
        except:
            self.post_operation_except(ope, relpath, file, file_info)

        return is_executed

    def __init__(self, path, ignored_list=[],
                 log_callback=None, tmp_path="", encoding="",
                 ctime=True, for_sync=True, is_skip_file_callback=None):

        LocalFileTree.__init__(self, path, ignored_list,
             log_callback, tmp_path, encoding, ctime, for_sync,
             is_skip_file_callback)

        self.encoding = locale.getdefaultlocale()[1]
        if not self.encoding:
            self.encoding = DEFAULT_LOCALE

        # To use normalize_name() and compare_name() method
        self.onedriveaccess = OneDriveAccess(session_path=DUMMY_SESSION_PATH)
        self.conv_name_f = self.onedriveaccess.normalize_name
        self.compare_path = self.onedriveaccess.compare_name

        self.cloudsync_onedriveaccess = None

    def set_cloudsync_onedriveaccess(self, cloudsync_onedriveaccess):

        self.cloudsync_onedriveaccess = cloudsync_onedriveaccess

    def get_equivarent_name_map(self, local_dir_path):

        try:
            name_list = self.listdir(local_dir_path)
        except:
            name_list = []

        equivalent_map = defaultdict(list)

        for base in name_list:
            for target in name_list:
                conv_base = self.conv_name_f(base)
                if self.compare_path(conv_base, target):
                    equivalent_map[conv_base] += [target]
        return equivalent_map

    def handle_equivarent_file(self, local_eq_map, relpath, file_name,
            file_info):

        conv_file = self.conv_name_f(file_name)
        eq_list = local_eq_map.get(conv_file)
        if not eq_list:
            return file_name

        if len(eq_list) > 1:
            self.logging(SKIP_ERROR_LOG_LV, EQUIVALENT_SKIP_LOG,
                            relpath=self.sep.join(relpath + [file_name]),
                            src=self, trace=False)
            raise MultipleEquivarentExist

        if len(eq_list) == 1:
            if self.compare_path(file_name.decode(self.encoding),
                                 eq_list[0].decode(self.encoding)):
                return eq_list[0]

        return file_name

    def handle_equivarent_dir(self, local_eq_map, relpath, dir_name, dir_info):

        conv_dir = self.conv_name_f(dir_name)
        eq_list = local_eq_map.get(conv_dir)

        if not eq_list:
            return dir_name

        if len(eq_list) > 1:
            self.logging(SKIP_ERROR_LOG_LV, EQUIVALENT_SKIP_LOG,
                            relpath=self.sep.join(relpath + [dir_name]),
                            src=self, trace=False)
            raise MultipleEquivarentExist

        if len(eq_list) == 1:
            if self.compare_path(dir_name.decode(self.encoding),
                                    eq_list[0].decode(self.encoding)):
                return eq_list[0]

        return dir_name

    def __update(self, relpath, tree):
        """This method is used at WF downloading"""

        local_dir_path = self.sep.join([self.root] + relpath)
        local_eq_map = self.get_equivarent_name_map(local_dir_path)

        for file in tree[FILES_NAME].keys():

            file_info = tree[FILES_NAME][file]

            real_file_name = file

            # 同等ファイルはスキップ
            if (file_info[STATE_NAME] == STATE_DEL or
                file_info[STATE_NAME] == STATE_NEW):

                try:
                    real_file_name = \
                        self.handle_equivarent_file(local_eq_map, relpath,
                                                    file, file_info)
                except MultipleEquivarentExist:
                    if file_info[STATE_NAME] == STATE_NEW:
                        self.post_file_equivarent(relpath + [file], file_info)
                    elif file_info[STATE_NAME] == STATE_DEL:
                        pass
                    continue

            try:
                self._DstFull__try_operation(
                            self.OPERATION[self.OPE_FILE_TRANSFER],
                            relpath, real_file_name, file_info)
            except:
                time.sleep(5)
                self._DstFull__try_operation(
                            self.OPERATION[self.OPE_FILE_TRANSFER],
                            relpath, real_file_name, file_info)

            if self._DstFull__try_operation(
                        self.OPERATION[self.OPE_FILE_DELETE],
                        relpath, real_file_name, file_info):
                del tree[FILES_NAME][file]

        for dir in tree[DIRS_NAME].keys():
            dir_info = tree[DIRS_NAME][dir]

            real_dir_name = dir

            # 同等ディレクトリはスキップ
            if (dir_info[STATE_NAME] == STATE_DEL or
                dir_info[STATE_NAME] == STATE_NEW):

                try:
                    real_dir_name = \
                        self.handle_equivarent_dir(
                                        local_eq_map, relpath, dir, dir_info)
                except MultipleEquivarentExist:
                    if dir_info[STATE_NAME] == STATE_NEW:
                        self.post_dir_equivarent(relpath + [dir], file_info)
                    elif dir_info[STATE_NAME] == STATE_DEL:
                        pass
                    continue

            from cloudsync.onedrivefiletree import (ForbiddenNameError,)

            try:
                self._DstFull__try_operation(
                                self.OPERATION[self.OPE_DIR_CREATE],
                                relpath, real_dir_name, dir_info)
            except ForbiddenNameError:
                self.logging(SKIP_ERROR_LOG_LV, INVALID_NAME_FOLDER_SKIP_LOG,
                            relpath=self.sep.join(relpath + [dir]),
                            src=self.src, trace=False)
                continue

            if self._DstFull__try_operation(
                                self.OPERATION[self.OPE_DIR_DELETE],
                                relpath, real_dir_name, dir_info):
                del tree[DIRS_NAME][dir]

            else:
                self.__update(relpath + [real_dir_name],
                    tree[DIRS_NAME][dir][CHILDREN_NAME])

    def update(self, src, tree):
        self.src = src
        self.__update([], tree)

    def post_file_equivarent(self, file_relpath, file_info):
        pass

    def post_dir_equivarent(self, dir_relpath, dir_info):
        pass
