# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

from onedrivesdk.error import OneDriveError
from onedrivesdk.model.upload_session import UploadSession
from onedrivesdk.model.item import Item
from onedrivesdk.options import HeaderOption
from onedrivesdk.request.item_request_builder import ItemRequestBuilder
from onedrivesdk.request_builder_base import RequestBuilderBase
from onedrivesdk.request_base import RequestBase
from onedrivesdk.request.item_content_request import ItemContentRequest
import trollius as asyncio
from trollius import From, Return
import json
import math
import time

__PART_SIZE = 10 * 1024 * 1024  # recommended file size. Should be multiple of 320 * 1024
__MAX_SINGLE_FILE_UPLOAD = 100 * 1024 * 1024


class ItemUploadFragment2(RequestBase):
    def __init__(self, request_url, client, options, data):
        super(ItemUploadFragment2, self).__init__(request_url, client, options)
        self.method = "PUT"
        self.data = data

    def post(self):
        """Sends the POST request

        Returns:
            :class:`UploadSession<onedrivesdk.model.upload_session.UploadSession>`:
                The resulting entity from the operation
        """
        entity = UploadSession(json.loads(self.send(data=self.data).content))
        return entity

    @asyncio.coroutine
    def post_async(self):
        """Sends the POST request using an asyncio coroutine

        Yields:
            :class:`UploadedSession<onedrivesdk.model.upload_session.UploadedSession>`:
                The resulting entity from the operation
        """
        future = self._client._loop.run_in_executor(None,
                                                    self.post)
        # entity = yield from future
        # return entity
        entity = yield From(future)
        raise Return(entity)


class ItemUploadFragmentBuilder2(RequestBuilderBase):
    def __init__(self, request_url, client, fp, total_size):
        super(ItemUploadFragmentBuilder2, self).__init__(request_url, client)
        self._method_options = {}
        self._file_handle = fp
        self._total_length = total_size

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._file_handle.close()

    def request(self, begin, length, options=None):
        """Builds the request for the ItemUploadFragment

        Args:
            begin (int): First byte in range to be uploaded
            length (int): Number of bytes in range to be uploaded
            options (list of :class:`Option<onedrivesdk.options.Option>`):
                Default to None, list of options to include in the request

        Returns:
            :class:`ItemUploadFragment<onedrivesdk.request.item_upload_fragment.ItemUploadFragment>`:
                The request
        """
        if not (options is None or len(options) == 0):
            opts = options.copy()
        else:
            opts = []

        self.content_type = "application/octet-stream"

        opts.append(HeaderOption("Content-Range", "bytes %d-%d/%d" % (begin, begin + length - 1, self._total_length)))
        opts.append(HeaderOption("Content-Length", str(length)))

        # file_slice = FileSlice(self._file_handle, begin, length=length)
        self._file_handle.seek(begin)
        data = self._file_handle.read(length)
        req = ItemUploadFragment2(self._request_url, self._client, opts, data)
        return req

    def post(self, begin, length, options=None):
        """Sends the POST request

        Returns:
            :class:`UploadedFragment<onedrivesdk.model.uploaded_fragment.UploadedFragment>`:
            The resulting UploadSession from the operation
        """
        return self.request(begin, length, options).post()

    @asyncio.coroutine
    def post_async(self, begin, length, options=None):
        """Sends the POST request using an asyncio coroutine

        Yields:
            :class:`UploadedFragment<onedrivesdk.model.uploaded_fragment.UploadedFragment>`:
                The resulting UploadSession from the operation
        """
        # entity = yield from self.request(begin, length, options).post_async()
        # return entity
        entity = yield From(self.request(begin, length, options).post_async())
        raise Return(entity)


def fragment_upload_async(self, buff, file_size, chunk_size, conflict_behavior=None, upload_status=None):
    if chunk_size is None:
        chunk_size = __PART_SIZE

    if file_size <= chunk_size:
        return OneDriveSingleUploader(self, buff, file_size, chunk_size, conflict_behavior, upload_status)
    else:
        return OneDriveFragmentUploader(self, buff, file_size, chunk_size, conflict_behavior, upload_status)


class OneDriveSingleUploader():
    def __init__(self, builder, buff, file_size, chunk_size, conflict_behavior=None, upload_status=None):
        self.builder = builder
        self.buff = buff

    def upload(self):
        self.builder.content.request().upload_data(self.buff.read())


class OneDriveFragmentUploader():
    def __init__(self, builder, buff, file_size, chunk_size, conflict_behavior=None, upload_status=None):
        self.builder = builder
        self.buff = buff
        self.file_size = file_size
        self.chunk_size = chunk_size
        self.part_number = 0
        self.total_parts = math.ceil((self.file_size - 1) / self.chunk_size) + 1
        self.upload_status = upload_status

        if conflict_behavior:
            item = Item({'@name.conflictBehavior': conflict_behavior})
        else:
            item = Item({})

        session = builder.create_session(item).post()
        self.fragment_builder = ItemUploadFragmentBuilder2(session.upload_url, builder._client, buff, file_size)

    def upload(self):
        if self.upload_status:
            self.upload_status(self.part_number, self.total_parts)

        begin = self.part_number * self.chunk_size
        length = min(self.chunk_size, self.file_size - begin)
        tries = 0
        resp = None
        while True:
            try:
                tries += 1
                resp = self.fragment_builder.post(begin, length)
            except OneDriveError as exc:
                if tries >= 5:
                    raise

                if exc.status_code in (408, 500, 502, 503, 504) and tries < 5:
                    time.sleep(5)
                    continue
                elif exc.status_code == 416:
                    # Fragment already received
                    break
                elif exc.status_code == 401:
                    self.builder._client.auth_provider.refresh_token()
                    continue
                else:
                    raise exc
            except ValueError:
                if tries >= 5:
                    raise
                # Swallow value errors (usually JSON error) and try again.
                continue
            self.part_number += 1
            break  # while True
        return resp


def upload_data(self, data):
    self.method = "PUT"
    entity_response = self.send(data=data)
    entity = Item(json.loads(entity_response.content))
    return entity


# Overwrite the standard upload operation to use this one
ItemRequestBuilder.upload_async = fragment_upload_async
ItemContentRequest.upload_data = upload_data
