# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""OneDrive File Access module.

Define classes for OneDrive file access.
"""

# common module
from __future__ import unicode_literals
from cloudsync.fileaccess import (FileObj, Access)
import os
import stat
import arrow
import urllib
import unicodedata
from cloudsync.buffer_object import BufferObject, BufferObjectW, ChunkedBuffer
from cloudsync.fileaccess_helper import *

# feature module
import onedrivesdk
from onedrivesdk.helpers import http_provider_with_proxy
from onedrivesdk.helpers.resource_discovery import ResourceDiscoveryRequest, ResourceDiscoveryRequestWithProxy
import onedrivesdk.error
import requests
import onedrive_fragment_uploader
import onedrivesdk_helper
from onedrive_fragment_uploader import __PART_SIZE as ONEDRIVE_DEFAULT_CHUNK_SIZE

# common constant
FILE_SEPARATOR = "/"
FILE_MODE = 0
DIR_MODE = stat.S_IFDIR
ERR_MSG_RENAME = "Invalid new name"
ERR_MSG_INVALID_FILE_LENGTH = "Invalid file length"
ERR_MSG_INVALID_OPEN_MODE = "Invalid open mode"
OPEN_MODE_READ = "r"
OPEN_MODE_WRITE = "w"

# feature constant
ERR_MSG_NOT_FOUND_SERVICE = "Not found service_name from discovery service"

MS_GRAPH_API_URL = "https://graph.microsoft.com/"

TIMEOUT_SECS = 300


def quote_path(path):
    """Quote path for url."""

    return urllib.quote(path.encode("UTF-8")).decode("UTF-8")


class OneDriveFileObj(FileObj):
    """Class for OneDrive file."""

    def __init__(self, access, path, mode, length, chunk_size):
        """Constructor."""
        super(OneDriveFileObj, self).__init__()
        self._access = access
        self._path = path
        self._mode = (filter((lambda m: m in mode.lower()), (OPEN_MODE_READ, OPEN_MODE_WRITE)) or [None])[0]
        self._buffer = None
        self._length = length
        self._chunk_size = chunk_size
        if self._chunk_size is None:
            self._chunk_size = ONEDRIVE_DEFAULT_CHUNK_SIZE
        self._r_pos = 0
        self._w_pos = 0
        self._downloader = None
        self._uploader = None

        if self._mode == OPEN_MODE_READ:
            self._download()
        elif self._mode == OPEN_MODE_WRITE:
            if self._length < 0:
                raise ValueError(ERR_MSG_INVALID_FILE_LENGTH)
            self._upload()
        else:
            raise ValueError(ERR_MSG_INVALID_OPEN_MODE)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def close(self):
        """Close file."""

        try:
            if self._mode == OPEN_MODE_WRITE:
                if self._buffer is not None and self._buffer.rest > 0:
                    self._upload(force=True)

        finally:
            if self._buffer:
                self._buffer.close()

    def read(self, size=-1):
        """Return a "size" bytes data from current position in file.
        If the size argument is negative or omitted, then return all data.
        """

        if size < 0:
            read_size = self._length
        else:
            read_size = min(size, self._length - self._r_pos)

        ret_buf = BufferObjectW(size=read_size)
        while ret_buf.lack > 0:
            data = self._buffer.read(ret_buf.lack)
            if not len(data):
                break

            ret_buf.write(data)
            if self._buffer.rest == 0:
                self._buffer.close()
                self._download()

        self._r_pos += ret_buf.length
        ret = ret_buf.read()
        ret_buf.close()

        return ret

    def _download(self):
        if not self._downloader:
            item = self._access._get_item(self._path, be_raise=(self._mode == OPEN_MODE_READ))
            if item is not None:
                self._length = self._access._stat(self._path)[stat.ST_SIZE]
                req = self._access._client.item(path=quote_path(self._path)).content.request()
                self._access._client.auth_provider.authenticate_request(req)
                self._buffer = BufferObject(size=self._chunk_size)
                get_kwargs = {
                    'url': req.request_url,
                    'headers': req._headers,
                    'stream': True,
                    'timeout': TIMEOUT_SECS,
                }
                if self._access._proxies is not None:
                    get_kwargs['proxies'] = self._access._proxies
                    get_kwargs['verify'] = True
                r = requests.get(**get_kwargs)
                self._downloader = r.iter_content(chunk_size=self._chunk_size)

        self._buffer.set_w_mode()
        self._buffer.clear()
        while self._buffer.lack > 0:
            try:
                data = self._downloader.next()
                self._buffer.write(data)
            except StopIteration:
                break
        self._buffer.set_r_mode()

    def write(self, data):
        """Write a "data" data to current position in file."""
        self._buffer.set_w_mode()
        self._buffer.write(data)
        self._w_pos += len(data)
        self._upload(force=False)
        self._buffer.close_chunk()

    def _upload(self, force=False):
        if not self._uploader:
            self._buffer = ChunkedBuffer(size=self._length, buffer_size=self._chunk_size, read_at_first=True)
            self._uploader = self._access._client.item(path=quote_path(self._path))\
                .upload_async(buff=self._buffer, file_size=self._length, chunk_size=self._chunk_size)

        if self._buffer.lack == 0 or force:
            self._buffer.set_r_mode()
            self._uploader.upload()


class OneDriveAccess(Access):
    """class accessing to OneDrive file."""

    def _get_file(self, target_path, be_raise=False):
        return self._get_item(target_path, be_raise)

    def _get_directory(self, target_path, be_raise=False):
        return self._get_item(target_path, be_raise)

    def _get_item(self, target_path, be_raise=False):
        """Return item object for "target_path"."""

        target_path = self.convert_name(target_path)

        try:
            item = self._client.item(path=quote_path(target_path)).get()
        except onedrivesdk.error.OneDriveError as err:
            if err.code == "itemNotFound":
                if not be_raise:
                    item = None
                else:
                    raise LookupError(target_path)
            else:
                raise

        return item

    @staticmethod
    def _stat_item(item):
        _mode = None

        if item.file is not None and item.folder is None:
            _mode = FILE_MODE
        if item.file is None and item.folder is not None:
            _mode = DIR_MODE

        _mtime = arrow.get(item.last_modified_date_time).timestamp
        _ctime = None

        if _mode == DIR_MODE:
            _size = None
        else:
            _size = item.size

        return {
                stat.ST_MODE: _mode,
                stat.ST_MTIME: _mtime,
                stat.ST_CTIME: _ctime,
                stat.ST_SIZE: _size,
            }

    def _stat(self, target_path):
        """Return a stat information dictionary for "target_path"."""

        item = self._get_item(target_path=target_path, be_raise=True)
        return self._stat_item(item)

    def __init__(self,
                 client_id=None,
                 client_secret=None,
                 scopes=None,
                 response_type='code',
                 redirect_url='http://localhost:8080',
                 session_path=None,
                 auth_server_url=None,
                 auth_token_url=None,
                 proxies=None,
                 discovery_url='https://api.office.com/discovery/',
                 service_name=None,
                 ):
        """Constructor.

        Args:
            client_id (unicode): Defaults to None, the client id for
                application
            client_secret (unicode): The client secret of application
            scopes (list of unicode): Defaults to None, the scopes
                that are required for application
            response_type (unicode): Response type query param value.
                If not provided, defaults to 'code'. Should be either
                'code' or 'token'.
            redirect_url (unicode): The URL to redirect the callback to
            session_path (unicode):
                File path for session(token, expired, etc...)
            auth_server_url (unicode):
                URL where OAuth authentication can be performed. If
                None, defaults to OAuth for Microsoft Account.
                (https://login.live.com/oauth20_authorize.srf)
            auth_token_url (unicode):
                URL where OAuth token can be redeemed. If None,
                defaults to OAuth for Microsoft Account.
                (https://login.live.com/oauth20_token.srf)
            proxies (dict of unicode:unicode):
                Mapping of protocols to proxy URLs.
                If None, bypass proxies
                e.g.
                proxies = {
                    'http': 'http://127.0.0.1:8888',
                    'https': 'https://127.0.0.1:8888',
                }
            discovery_url (unicode):
                for "OneDrive for business"
                ResourceId for Discovery Service
                If serive_name is None, not in use
             serive_name (unicode):
                for "OneDrive for business"
                Used to find the endpoint.
                If not use a specific serviceName, "Office 365 SharePoint"
        """

        super(OneDriveAccess, self).__init__()

        self._client_id = self.convert_name(client_id)
        self._client_secret = self.convert_name(client_secret)
        self._scopes = self.convert_name(scopes)
        self._response_type = self.convert_name(response_type)
        self._redirect_url = self.convert_name(redirect_url)
        self._session_path = self.convert_name(session_path)
        self._proxies = self.convert_name(proxies)
        self._discovery_url = self.convert_name(discovery_url)
        self._service_name = self.convert_name(service_name)

        self._client = None

        if self._proxies is not None:
            self._http_provider = http_provider_with_proxy.HttpProviderWithProxy(proxies=self._proxies, verify_ssl=True)
        else:
            self._http_provider = onedrivesdk.HttpProvider()

        self._auth_provider = onedrivesdk.AuthProvider(
            http_provider=self._http_provider,
            client_id=self._client_id,
            scopes=self._scopes,
            auth_server_url=auth_server_url,
            auth_token_url=auth_token_url,
        )

        if os.path.exists(self._session_path):
            self._auth_provider.load_session(path=self._session_path)

        if self._auth_provider._session is not None \
            and (
                    (
                        self._auth_provider._session.refresh_token is None and
                            self._auth_provider._session.is_expired() == False
                    )
                    or
                    (
                        self._auth_provider._session.refresh_token is not None
                    )
            ):
            self._auth_url = redirect_url
        else:
            if self._service_name == "":
                self._auth_url = self._auth_provider.get_auth_url(redirect_url, response_type=self._response_type)
            else:
                self._auth_url = self._auth_provider.get_auth_url(redirect_url)

    @property
    def auth_url(self):
        """Get the auth url using the params provided
        If session is already valid, returns redirect_url
        """
        return self._auth_url

    def authenticate(self, params=None):
        """Takes in params, creates a session.

        Args:
            params(dict):
                The GET parameter sent to redirect_url
                If session is already valid, is None.
                e.g.
                1) {'code': '....'}
                2) {'access_token': '...', 'client_id': '...', ...}
        """

        if params is None:
            params = {}

        code = params.get("code")
        if code is not None:
            if self._service_name is None:
                _resource = None
            else:
                _resource = self._discovery_url

            self._auth_provider.authenticate(code,
                                             self._redirect_url,
                                             self._client_secret,
                                             resource=_resource)

        if self._response_type == 'code' \
                and self._auth_provider._session.is_expired() \
                and len(set(self._auth_provider.scopes).
                                intersection(set(['wl.offline_access', 'offline_access']))) > 0:
            self._auth_provider.refresh_token()

        self._auth_provider.save_session(path=self._session_path)

        if self._service_name is None:
            _url = 'https://api.onedrive.com/v1.0/'
        else:
            _url = ''
            resource_discovery_request = None
            if self._proxies is not None:
                resource_discovery_request = ResourceDiscoveryRequestWithProxy(proxies=self._proxies, verify_ssl=True)
            else:
                resource_discovery_request = ResourceDiscoveryRequest()
            self._auth_provider.redeem_refresh_token(MS_GRAPH_API_URL)
            service_infos = resource_discovery_request.get_service_info(self._auth_provider.access_token)
            for service_info in service_infos:
                if service_info.service_name == self._service_name:
                    # _url = service_info.service_endpoint_uri + "/"
                    _url = service_info.service_resource_id.rstrip("/") + "/_api/v2.0/"
                    self._auth_provider.redeem_refresh_token(service_info.service_resource_id)
                    break
            if _url == "":
                raise Exception(ERR_MSG_NOT_FOUND_SERVICE)

        self._client = onedrivesdk.OneDriveClient(self.convert_name(_url), self._auth_provider, self._http_provider)

    def normalize_name(self, name):
        """Normalize charcter in "name" according to the server rules."""

        is_decoded = False
        if isinstance(name, str):
            try:
                is_decoded = True
                name = name.decode("UTF-8")
            except:
                # Do nothing
                return name

        # Normalize "name" according to unicode "NFC" rules."""
        normalized_name = unicodedata.normalize("NFC", name)
        if is_decoded:
            normalized_name = normalized_name.encode("UTF-8")
        return normalized_name

    def compare_name(self, one, other):
        """Compare "one" and "other" according to the server rules."""

        return self.normalize_name(one) == self.normalize_name(other)

    def make_directory(self, dir_path):
        """Create "dir_path" directory."""

        if dir_path == "" or dir_path == FILE_SEPARATOR:
            # if blank, do nothing
            return

        target_id = self._get_item(dir_path)
        if target_id is not None:
            # if exists, do nothing
            return

        dir_path = self.convert_name(dir_path)

        f = onedrivesdk.Folder()
        i = onedrivesdk.Item()
        parent_path = os.path.dirname(dir_path)
        new_dir_name = os.path.basename(dir_path)

        i.name = new_dir_name
        i.folder = f

        self._client.item(path=quote_path(parent_path)).children.add(i)

    def delete_directory(self, dir_path):
        """Delete "dir_path" directory."""

        dir_path = self.convert_name(dir_path)
        self.delete_file(dir_path)

    def stat_directory(self, dir_path):
        """Return a stat information dictionary for "file_path"
           which keys below:
            stat.ST_MODE, stat.ST_MTIME, stat.CTIME, stat.ST_SIZE
        """

        dir_path = self.convert_name(dir_path)
        return self._stat(dir_path)

    def rename_directory(self, dir_path, new_name):
        """Rename a name of "dir_path" to "new_name"."""

        if FILE_SEPARATOR in new_name:
            raise Exception(ERR_MSG_RENAME)

        dir_path = self.convert_name(dir_path)
        new_name = self.convert_name(new_name)

        self.rename_file(dir_path, new_name)

    def move_directory(self, dir_path, new_dir_path):
        """Move "dir_path" to "new_path"."""

        dir_path = self.convert_name(dir_path)
        new_dir_path = self.convert_name(new_dir_path)
        self.move_file(dir_path, new_dir_path)

    def get_contents(self, dir_path, with_stat=False):
        """Return a contents dictionary in "dir_path".
           A stat information must be included, if "with_stat" is True,
        """

        dir = self._get_directory(dir_path, be_raise=True)
        contents = {}
        if dir:
            items = self._client.item(path=quote_path(dir_path)).children.request().get()
            while True:
                for item in iter(items):
                    _name = os_path_bottom_name(item.name)

                    if with_stat:
                        contents[_name] = self._stat_item(item)
                    else:
                        contents[_name] = {}

                if hasattr(items, "_next_page_link"):
                    items = onedrivesdk.ChildrenCollectionRequest.get_next_page_request(items, self._client).get()
                else:
                    break
        return contents

    def get_updated_contents(self, dir_path):
        """Return a updated contents dictionary in "dir_path"."""

        raise NotImplementedError

    def open(self, file_path, mode="r", length=-1, chunk_size=None):
        """Return a class instance of the "FileObj" or sub class
           corresponding to "file_path".
        """

        return OneDriveFileObj(self, file_path, mode, length, chunk_size)

    def delete_file(self, file_path):
        """Delete "file_path" file."""

        file = self._get_file(file_path, be_raise=True)
        if file:
            self._client.item(path=quote_path(file_path)).delete()

    def stat_file(self, file_path):
        """Return a stat information dictionary for "file_path"
           which keys below:
            stat.ST_MODE, stat.ST_MTIME, stat.CTIME, stat.ST_SIZE
        """

        return self._stat(file_path)

    def rename_file(self, file_path, new_name):
        """Rename a name of "file_path" to "new_name"."""

        if FILE_SEPARATOR in new_name:
            raise Exception(ERR_MSG_RENAME)

        new_name = self.convert_name(new_name)
        item = self._get_item(file_path, be_raise=True)

        renamed_item = onedrivesdk.Item()
        renamed_item.name = new_name
        renamed_item.id = item.id

        self._client.item(id=item.id).update(renamed_item)

    def move_file(self, file_path, new_dir_path):
        """Move "file_path" to "new_path"."""

        file = self._get_file(file_path, be_raise=True)
        if file:
            new_dir_path = self.convert_name(new_dir_path)
            new_dir = self._get_item(new_dir_path, be_raise=True)

            ref = onedrivesdk.ItemReference()
            ref.id = new_dir.id

            basename = os.path.basename(file_path)
            copy_operaion = self._client.\
                item(path=quote_path(file_path)).\
                copy(name=basename, parent_reference=ref).\
                post()
            copy_operaion.poll_until_done()

            self.delete_file(file_path)

    def test(self):
        pass


